#include <config.h>
module gpu_block_solver_class
  use iso_c_binding
  use comlib
  use lattice_class
  use error_class
  use field_gauge_class
  use field_fermion_class
  use quark_clover_class
  implicit none
  private

  public :: gpu_block_solver
  public :: new,delete
  public :: assign_inv_mult_eoprec_wd


  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP

  integer, save, allocatable :: GPU_TASK_TABLE(:,:)
  integer, parameter :: D_GAUGE_SIZE = COL*COL*(NTH+1)*(NZ+2)*(NY+2)*(NX+2)*NDIM
  integer, parameter :: D_FERMI_SIZE = COL*SPIN*(NTH+1)*(NZ+2)*(NY+2)*(NX+2)

  integer, parameter :: S_GAUGE_SIZE = COL*(COL-1)*NTH*NZ*NY*NX*NDIM
  integer, parameter :: S_FERMI_SIZE = COL*SPIN*NTH*NZ*NY*NX

#ifdef _CLOVER
  integer, parameter :: D_CLV_SIZE = CLSPH*2*NTH*NZ*NY*NX
  integer, parameter :: S_CLV_SIZE = CLSPH*2*NTH*NZ*NY*NX
#endif


  type gpu_block_solver
    private
    complex(C_FLOAT), allocatable ::  swe(:,:)
    complex(C_FLOAT), allocatable ::  sve(:,:)
    complex(C_FLOAT), allocatable ::   su(:,:)
#ifdef _CLOVER
    complex(C_FLOAT), allocatable :: sclv(:,:)
#endif
    integer, public :: ieo
    integer, public :: num_pf
    integer, public :: num_gpu
    real(DP), public :: tol
    real(DP), public :: err_norm
    integer,  public :: iter
    integer(C_INT),  allocatable :: iter_all(:)
#ifdef _DEBUG
    integer,  public :: print_flag = 1
#else
    integer,  public :: print_flag = 0
#endif
  end type

  type gpu_block_solver_ptr
    type(gpu_block_solver), pointer :: ptr => NULL()
  end type

  interface new
    module procedure new_gpu_blk_slv_f
  end interface

  interface delete
    module procedure delete_gpu_blk_slv_f
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_blk_gpu
  end interface

  interface

    subroutine bind_cpu_to_gpu() BIND(C,name="bind_cpu_to_gpu_")
    end subroutine

    subroutine unbind_cpu_to_gpu() BIND(C,name="unbind_cpu_to_gpu_")
    end subroutine

!=================================================================
! for GPU# 1
!=================================================================
#ifdef _CLOVER
    subroutine new_gpu_blk_solver_g1(fue,fuo,ieo,fclve,fclvo) BIND(C,name="new_gpu_solver_g1")
      import C_INT,C_FLOAT,S_GAUGE_SIZE,S_CLV_SIZE
#else
    subroutine new_gpu_blk_solver_g1(fue,fuo,ieo) BIND(C,name="new_gpu_solver_g1")
      import C_INT,C_FLOAT,S_GAUGE_SIZE
#endif
      implicit none
      complex(C_FLOAT),  intent(in) :: fue(S_GAUGE_SIZE),fuo(S_GAUGE_SIZE)
      integer(C_INT),    intent(in) :: ieo
#ifdef _CLOVER
      complex(C_FLOAT),  intent(in) :: fclve(S_CLV_SIZE),fclvo(S_CLV_SIZE)
#endif
    end subroutine

    subroutine delete_gpu_blk_solver_g1() BIND(C,name="delete_gpu_solver_g1")
      implicit none
    end subroutine

    subroutine gpu_blk_solver_g1(tol,iter,kappa,ye,dye,ieo) BIND(C,name="bicgstab_g1")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

!=================================================================
! for GPU# 2
!=================================================================
#ifdef _CLOVER
    subroutine new_gpu_blk_solver_g2(fue,fuo,ieo,fclve,fclvo) BIND(C,name="new_gpu_solver_g2")
      import C_INT,C_FLOAT,S_GAUGE_SIZE,S_CLV_SIZE
#else
    subroutine new_gpu_blk_solver_g2(fue,fuo,ieo) BIND(C,name="new_gpu_solver_g2")
      import C_INT,C_FLOAT,S_GAUGE_SIZE
#endif
      implicit none
      complex(C_FLOAT),  intent(in) :: fue(S_GAUGE_SIZE),fuo(S_GAUGE_SIZE)
      integer(C_INT),    intent(in) :: ieo
#ifdef _CLOVER
      complex(C_FLOAT),  intent(in) :: fclve(S_CLV_SIZE),fclvo(S_CLV_SIZE)
#endif
    end subroutine

    subroutine delete_gpu_blk_solver_g2() BIND(C,name="delete_gpu_solver_g2")
      implicit none
    end subroutine

    subroutine gpu_blk_solver_g2(tol,iter,kappa,ye,dye,ieo) BIND(C,name="bicgstab_g2")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in) :: ieo
    end subroutine


!=================================================================
! for GPU# 3
!=================================================================
#ifdef _CLOVER
    subroutine new_gpu_blk_solver_g3(fue,fuo,ieo,fclve,fclvo) BIND(C,name="new_gpu_solver_g3")
      import C_INT,C_FLOAT,S_GAUGE_SIZE,S_CLV_SIZE
#else
    subroutine new_gpu_blk_solver_g3(fue,fuo,ieo) BIND(C,name="new_gpu_solver_g3")
      import C_INT,C_FLOAT,S_GAUGE_SIZE
#endif
      implicit none
      complex(C_FLOAT),  intent(in) :: fue(S_GAUGE_SIZE),fuo(S_GAUGE_SIZE)
      integer(C_INT),    intent(in) :: ieo
#ifdef _CLOVER
      complex(C_FLOAT),  intent(in) :: fclve(S_CLV_SIZE),fclvo(S_CLV_SIZE)
#endif
    end subroutine

    subroutine delete_gpu_blk_solver_g3() BIND(C,name="delete_gpu_solver_g3")
      implicit none
    end subroutine

    subroutine gpu_blk_solver_g3(tol,iter,kappa,ye,dye,ieo) BIND(C,name="bicgstab_g3")
      import C_INT,C_FLOAT,C_DOUBLE,S_FERMI_SIZE
      implicit none
      real(C_DOUBLE),    intent(in)    :: tol,kappa
      integer(C_INT),    intent(inout) :: iter
      complex(C_FLOAT),  intent(in)    ::  ye(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: dye(S_FERMI_SIZE)
      integer(C_INT),    intent(in) :: ieo
    end subroutine


!====================================================================
! precision and array conversion routines for GPU<=>CPU
!====================================================================
    subroutine conv_q_d2s(vec,fye,ieo,norm) BIND(C,name="conv_q_vec_d2s_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_FERMI_SIZE,S_FERMI_SIZE
      implicit none
      complex(C_DOUBLE), intent(in)    :: vec(S_FERMI_SIZE)
      complex(C_FLOAT),  intent(inout) :: fye(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(inout) :: norm
    end subroutine

    subroutine conv_q_s2d(fye,vec,ieo,norm) BIND(C,name="conv_q_vec_s2d_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_FERMI_SIZE,S_FERMI_SIZE
      implicit none
      complex(C_FLOAT),  intent(in)    :: fye(S_FERMI_SIZE)
      complex(C_DOUBLE), intent(inout) :: vec(S_FERMI_SIZE)
      integer(C_INT),    intent(in)    :: ieo
      real(C_DOUBLE),    intent(in)    :: norm
    end subroutine

    subroutine conv_u(due,fue,ieo) BIND(C,name="conv_u_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_GAUGE_SIZE,S_GAUGE_SIZE
      implicit none
      complex(C_DOUBLE), intent(in)    :: due(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
      complex(C_FLOAT),  intent(inout) :: fue(S_GAUGE_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine

#ifdef _CLOVER
    subroutine conv_clv(dclve,fclve,ieo) BIND(C,name="conv_clv_")
      use lattice_class
      import C_INT,C_FLOAT,C_DOUBLE,D_CLV_SIZE,S_CLV_SIZE
      implicit none
      complex(C_DOUBLE), intent(in)    :: dclve(CLSPH,2,NTH,NZ,NY,NX)
      complex(C_FLOAT),  intent(inout) :: fclve(S_CLV_SIZE)
      integer(C_INT),    intent(in)    :: ieo
    end subroutine
#endif

  end interface

contains

#ifdef _CLOVER
subroutine new_gpu_blk_slv_f(this,u,ieo,fclinv)
#else
subroutine new_gpu_blk_slv_f(this,u,ieo)
#endif
  use field_gauge_class
#ifdef _CLOVER
  use field_clover_class
#endif
  use gpu_solver_class, only : bind_cpu_to_gpu
  implicit none
  type(gpu_block_solver),  intent(inout) :: this
  type(vfield_gluon_wg),   intent(in)    :: u
  integer, intent(in)                    :: ieo
#ifdef _CLOVER
  type(field_clover_term), intent(in)    :: fclinv
  integer :: ieoxyz,ics
#endif
  complex(DP), allocatable :: uu(:,:,:,:,:,:,:)
  integer :: ic,jc,ith,iz,iy,ix,mu
  integer :: ioe,igpu

  this%iter = 0
  if (allocated(this%iter_all)) deallocate(this%iter_all)
  allocate(this%iter_all(this%num_pf))
  this%iter_all(:) = 0

  this%ieo = ieo
  ioe = 1-ieo

  if (allocated(this%su)) deallocate(this%su)
  allocate(this%su(S_GAUGE_SIZE,0:1))

  allocate(uu(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM))
  do mu=1,NDIM
  do ix=0,NX1
  do iy=0,NY1
  do iz=0,NZ1
  do ith=0,NTH
  do jc=1,COL
  do ic=1,COL
    uu(ic,jc,ith,iz,iy,ix,mu) = u%eo(ieo)%mu(mu)%s(ith,iz,iy,ix)%u(ic,jc)
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  call conv_u(uu,this%su(:,ieo),ieo)

  do mu=1,NDIM
  do ix=0,NX1
  do iy=0,NY1
  do iz=0,NZ1
  do ith=0,NTH
  do jc=1,COL
  do ic=1,COL
    uu(ic,jc,ith,iz,iy,ix,mu) = u%eo(ioe)%mu(mu)%s(ith,iz,iy,ix)%u(ic,jc)
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  enddo
  call conv_u(uu,this%su(:,ioe),ioe)
  deallocate(uu)

#ifdef _CLOVER
  if (allocated(this%sclv)) deallocate(this%sclv)
  allocate(this%sclv(S_CLV_SIZE,0:1))
  call conv_clv(fclinv%eo(ieo)%l,this%sclv(:,ieo),ieo)
  call conv_clv(fclinv%eo(ioe)%l,this%sclv(:,ioe),ioe)
#endif

  call bind_cpu_to_gpu()
  call set_gpu_task_table(this%num_pf,this%num_gpu)

!$OMP PARALLEL DO PRIVATE(igpu)
  do igpu=1,this%num_gpu
    select case(igpu)
    case(1)
      call new_gpu_blk_solver_g1(this%su(:,ieo),this%su(:,ioe),ieo,this%sclv(:,ieo),this%sclv(:,ioe))
    case(2)
      call new_gpu_blk_solver_g2(this%su(:,ieo),this%su(:,ioe),ieo,this%sclv(:,ieo),this%sclv(:,ioe))
    case(3)
      call new_gpu_blk_solver_g3(this%su(:,ieo),this%su(:,ioe),ieo,this%sclv(:,ieo),this%sclv(:,ioe))
    end select
  enddo

  return
end subroutine

subroutine delete_gpu_blk_slv_f(this)
  implicit none
  type(gpu_block_solver), intent(inout) :: this
  integer :: igpu
  if (allocated(this%iter_all)) deallocate(this%iter_all)
  if (allocated(this%su))  deallocate(this%su)
  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
#ifdef _CLOVER
  if (allocated(this%sclv)) deallocate(this%sclv)
#endif
!$OMP PARALLEL DO PRIVATE(igpu)
  do igpu=1,this%num_gpu
    select case(igpu)
    case(1)
      call delete_gpu_blk_solver_g1()
    case(2)
      call delete_gpu_blk_solver_g2()
    case(3)
      call delete_gpu_blk_solver_g3()
    end select
  enddo
  return
end subroutine

subroutine set_gpu_task_table(num_pf,ngpu)
!
! set GPU and PF-field assignment table
!
  implicit none
  integer, intent(in) :: num_pf
  integer, intent(in) :: ngpu
  integer :: ipf,igpu
  external :: get_num_gpu   ! function defined in GPUSolverLIB(CUDA function)

  if (num_pf <= 0) then
    call error_stop("Number of PF field should be positive.")
  endif

  if (allocated(GPU_TASK_TABLE)) deallocate(GPU_TASK_TABLE)
  allocate(GPU_TASK_TABLE(ngpu,3))

  GPU_TASK_TABLE(:,:)=0

  !=======================================
  ! count ps-fields for each GPU.
  !=======================================
  do ipf=1,num_pf
    igpu = mod(ipf-1,ngpu)+1
    GPU_TASK_TABLE(igpu,3) = GPU_TASK_TABLE(igpu,3) + 1
  enddo

  igpu=1
  if (GPU_TASK_TABLE(igpu,3) > 0) then
    GPU_TASK_TABLE(igpu,1) = 1
    GPU_TASK_TABLE(igpu,2) = GPU_TASK_TABLE(igpu,1)+GPU_TASK_TABLE(igpu,3)-1
  endif
  do igpu=2,ngpu
    if (GPU_TASK_TABLE(igpu,3) > 0) then
      GPU_TASK_TABLE(igpu,1) = GPU_TASK_TABLE(igpu-1,2)+1
      GPU_TASK_TABLE(igpu,2) = GPU_TASK_TABLE(igpu,1)+GPU_TASK_TABLE(igpu,3)-1
    endif
  enddo

  if (0==nodeid) then
    write(*,'("# GPU START_PF END_PF NUM_PF")')
    do igpu=1,ngpu
      write(*,'("# ",I3,2X,3I7)')igpu,GPU_TASK_TABLE(igpu,1),&
 &                                    GPU_TASK_TABLE(igpu,2),&
 &                                    GPU_TASK_TABLE(igpu,3)
    enddo
  endif

  return
end subroutine

subroutine assign_inv_mult_eoprec_wd_blk_gpu(this,kappa,x,y)
!
! x = D \ y
!
  use field_fermion_class
  implicit none
  type(gpu_block_solver), intent(inout) :: this
  real(DP),    intent(in)    :: kappa
  complex(DP), intent(inout) :: x(:,:),y(:,:)

  real(C_DOUBLE), allocatable :: norm(:)
  real(C_DOUBLE) :: tol
  complex(DP), allocatable :: vv(:)
  integer :: igpu,ipf,ieo,num_pf,j

  ieo = this%ieo
  num_pf = this%num_pf
  tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)
  write(*,'("% target_tol, residual, gpu_tol:",3E24.15)')this%tol,this%err_norm,tol

  allocate(norm(num_pf))
  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  allocate(this%swe(S_FERMI_SIZE,num_pf))
  allocate(this%sve(S_FERMI_SIZE,num_pf))
  allocate(vv(S_FERMI_SIZE))

  do ipf=1,num_pf
!$OMP PARALLEL WORKSHARE
    vv(:) = y(ipf,:)
!$OMP END PARALLEL WORKSHARE
    call conv_q_vec_d2s(vv,this%swe(:,ipf),ieo,norm(ipf))
  enddo

!$OMP PARALLEL DO PRIVATE(igpu,ipf)
  do igpu=1,this%num_gpu
    select case(igpu)
    case(1)
      do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
        call gpu_blk_solver_g1(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
      enddo
    case(2)
      do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
        call gpu_blk_solver_g2(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
      enddo
    case(3)
      do ipf=GPU_TASK_TABLE(igpu,1),GPU_TASK_TABLE(igpu,2)
        call gpu_blk_solver_g3(tol,this%iter_all(ipf),kappa,this%swe(:,ipf),this%sve(:,ipf),ieo)
      enddo
    end select
  enddo

  this%iter = SUM(this%iter_all(:))

  do ipf=1,num_pf
    call conv_q_vec_s2d(this%sve(:,ipf),vv,ieo,norm(ipf))
!$OMP PARALLEL WORKSHARE
    x(ipf,:) = vv(:)
!$OMP END PARALLEL WORKSHARE
  enddo

  deallocate(this%sve)
  deallocate(this%swe)
  deallocate(vv)
  deallocate(norm)

  return
end subroutine

end module
