#include "config.h"

module gpu_solver_class
  use iso_c_binding
  use comlib
  use lattice_class
  use counter_class
  use field_gauge_class
  use field_fermion_class
  use field_clover_class
  use single_field_class
  implicit none
  private

  public :: gpu_solver
  public :: new, delete
  public :: assign_inv_mult_eoprec_wd

  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP

  logical, save :: m_is_gpu_binded_to_cpu = .false.
  logical, save :: m_is_gpu_initialized   = .false.

  integer(C_INT), save, bind(C,name="cu_mult_count") :: gpu_mult_count = 0

  type gpu_solver
    private
    type(s_wqf_eo_obj),  allocatable :: swe
    type(s_wqf_eo_obj),  allocatable :: sve
    type(s_gvf_eo_obj),  allocatable :: sue
    type(s_gvf_eo_obj),  allocatable :: suo
    type(s_clvf_eo_obj), allocatable :: sfcle
    type(s_clvf_eo_obj), allocatable :: sfclo
    real(DP), public :: err_norm
    real(DP), public :: tol
    integer,  public :: iter
    integer  :: idummy
  end type

  interface new
    module procedure new_gpu_solver_f
  end interface

  interface delete
    module procedure delete_gpu_solver_f
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_gpu
  end interface

  interface 
   !
   ! Fortran interfaces for GPU code using F2003 C binding
   !

    subroutine bind_cpu_to_gpu() BIND(C,name="bind_cpu_to_gpu_")
    end subroutine

    subroutine unbind_cpu_to_gpu() BIND(C,name="unbind_cpu_to_gpu_")
    end subroutine

    subroutine new_gpu_solver_clover(ue, uo, fclinve, fclinvo) BIND(C,name="new_gpu_solver_clover_")
      import s_gvf_eo_obj, s_clvf_eo_obj
      implicit none
      type(s_gvf_eo_obj),  intent(in) :: ue,uo
      type(s_clvf_eo_obj), intent(in) :: fclinve,fclinvo
    end subroutine

    subroutine new_gpu_solver_wilson(ue, uo) BIND(C,name="new_gpu_solver_wilson_")
      import s_gvf_eo_obj;
      implicit none
      type(s_gvf_eo_obj),  intent(in) :: ue,uo
    end subroutine

    subroutine delete_gpu_solver() BIND(C,name="delete_gpu_solver_")
    end subroutine

    subroutine s_bicgstab_gpu(tol, iter, kappa, ye, dye) BIND(C,name="s_bicgstab_hmc_gpu_")
      import c_double, c_int, s_wqf_eo_obj
      implicit none
      real(c_double),     intent(in) :: tol, kappa
      integer(c_int),     intent(in) :: iter
      type(s_wqf_eo_obj), intent(inout) :: ye, dye
    end subroutine

  end interface
  
contains

subroutine new_gpu_solver_f(this,u,fclinv)
  implicit none
  type(gpu_solver),               intent(inout) :: this
  type(vfield_gluon_wg),             intent(in) :: u
  type(field_clover_term), optional, intent(in) :: fclinv
  logical :: is_clover

  if (present(fclinv)) then
    is_clover = .true.
  else
    is_clover = .false.
  endif

  if (.not.m_is_gpu_binded_to_cpu) then
    call bind_cpu_to_gpu()
    m_is_gpu_binded_to_cpu = .true.
  endif

  this%iter = 0
  gpu_mult_count = 0

  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  if (allocated(this%sue)) deallocate(this%sue)
  if (allocated(this%suo)) deallocate(this%suo)
  allocate(this%swe)
  allocate(this%sve)
  allocate(this%sue)
  allocate(this%suo)
  call new(this%sue,0)
  call new(this%suo,1)

  if (is_clover) then
    if (allocated(this%sfcle)) deallocate(this%sfcle)
    if (allocated(this%sfclo)) deallocate(this%sfclo)
    allocate(this%sfcle)
    allocate(this%sfclo)
    call new(this%sfcle,0)
    call new(this%sfclo,1)
  endif

  call conv_u(this%sue,u%eo(0))
  call conv_u(this%suo,u%eo(1))
  if (is_clover) then
    call conv_clv(this%sfcle,fclinv%eo(0))
    call conv_clv(this%sfclo,fclinv%eo(1))
  endif

  if (is_clover) then
    call new_gpu_solver_clover(this%sue,this%suo,this%sfcle,this%sfclo)
  else
    call new_gpu_solver_wilson(this%sue,this%suo)
  endif

  if (allocated(this%sue)) deallocate(this%sue)
  if (allocated(this%suo)) deallocate(this%suo)
  if (is_clover) then
    if (allocated(this%sfcle)) deallocate(this%sfcle)
    if (allocated(this%sfclo)) deallocate(this%sfclo)
  endif

  m_is_gpu_initialized = .true.
  return
end subroutine

subroutine delete_gpu_solver_f(this)
  implicit none
  type(gpu_solver),      intent(inout) :: this

  this%iter = 0
  gpu_mult_count = 0

  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  if (allocated(this%sue)) deallocate(this%sue)
  if (allocated(this%suo)) deallocate(this%suo)
  if (allocated(this%sfcle)) deallocate(this%sfcle)
  if (allocated(this%sfclo)) deallocate(this%sfclo)

  if (m_is_gpu_initialized) then
    call delete_gpu_solver
  endif

  m_is_gpu_initialized = .false.

  return
end subroutine

subroutine assign_inv_mult_eoprec_wd_gpu(this,kappa,dye,ye,ieo)
!
!       v
! dye = Dee \ ye
!
  implicit none
  type(gpu_solver), intent(inout) :: this
  real(DP),         intent(in)    :: kappa
  complex(DP),      intent(inout) ::  ye(:)  ! packed data
  complex(DP),      intent(inout) :: dye(:)  ! packed data
  integer,          intent(in)    :: ieo
  type(field_quark_eo_wg), allocatable :: we,ve
  real(DP) :: norm, tol

  gpu_mult_count = 0
  tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)

  allocate(we,ve)
  call new(we,ieo)
  call new(ve,ieo)
  call new(this%swe,ieo)
  !
  ! swe <= ye/|ye|, norm = |ye|
  !
  call unpack(ye,we)
  call conv_q_d2s(this%swe,we,norm)
!  call unpack_d2s(ye,this%swe,norm)

  !        v
  ! sve <= Dee \ swe
  !
  call s_bicgstab_gpu(tol, this%iter, kappa, this%swe, this%sve)
!  this%sve = this%swe

  !
  ! dye <= sve*norm
  !
  call conv_q_s2d(ve,this%sve,norm)
  call pack(ve,dye)
  deallocate(ve,we)
!  call pack_s2d(this%sve,dye,norm)    

  call inc(mult_iter,gpu_mult_count)
  gpu_mult_count = 0

  return
end subroutine

subroutine unpack_d2s(vec,sq,norm)
!
! norm = |vec|
! sq <= vec/norm
!
  implicit none
  complex(DP),        intent(inout) :: vec(:)
  type(s_wqf_eo_obj), intent(inout) :: sq
  real(DP),           intent(out)   :: norm
  type(su3ferm) :: yy
  real(DP) :: rtmp
  complex(DP) :: ztmp
  integer :: igx,ibx,iby,ibz,ithz,itht,itht0
  integer :: ix,iy,iz,ieoxyz,ith,is,ic,isite,ieo

  ieo = sq%ieo

  norm = 0.0_DP
!$OMP PARALLEL DO PRIVATE(isite) REDUCTION(+:norm)
  do isite=1,SIZE(vec)
    norm = norm + REAL(vec(isite))**2 + AIMAG(vec(isite))**2
  enddo
#ifndef _singlePU
  call comlib_sumcast(norm)
#endif
  norm = SQRT(norm)
  rtmp = 1.0_DP/norm 

!$OMP PARALLEL DO PRIVATE(igx,ibx,iby,ibz,ithz,itht,itht0,ix,iy,iz,ieoxyz,ic,is,isite,yy,ztmp)
  do igx=1,NGX
    do ibx=1,NBX
    ix = ibx + (igx-1)*NBX
    do iby=1,NBY
    do ibz=1,NBZ
      iy = iby
      do ithz=1,NTHZ
        iz = ithz + (ibz-1)*NTHZ
        ieoxyz=mod(ipeo+ieo+ix+iy+iz,2)
        do itht=1-ieoxyz,NTHT-ieoxyz
          itht0=itht+ieoxyz

          do is=1,SPIN
          do ic=1,COL
            isite = ic  &
 &            +    (is-1)*COL  &
 &            + (itht0-1)*CLSP  &
 &            +    (iz-1)*CLSP*NTH  &
 &            +    (iy-1)*CLSP*NTH*NZ  &
 &            +    (ix-1)*CLSP*NTH*NZ*NY

            ztmp = vec(isite)*rtmp
            yy%yr(ic,is) =  REAL(ztmp)
            yy%yi(ic,is) = AIMAG(ztmp)
          enddo
          enddo

          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%x = yy%yr(1,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%y = yy%yr(2,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%z = yy%yr(3,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%w = yy%yr(1,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%x = yy%yr(2,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%y = yy%yr(3,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%z = yy%yr(1,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%w = yy%yr(2,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%x = yy%yr(3,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%y = yy%yr(1,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%z = yy%yr(2,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%w = yy%yr(3,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%x = yy%yi(1,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%y = yy%yi(2,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%z = yy%yi(3,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%w = yy%yi(1,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%x = yy%yi(2,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%y = yy%yi(3,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%z = yy%yi(1,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%w = yy%yi(2,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%x = yy%yi(3,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%y = yy%yi(1,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%z = yy%yi(2,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%w = yy%yi(3,4)

        enddo
      enddo
    enddo
    enddo
    enddo
  enddo

  call copy_boundary(sq)

  return
end subroutine

subroutine pack_s2d(sq,vec,norm)
!
! vec <= sq * norm
!
  implicit none
  type(s_wqf_eo_obj), intent(inout) :: sq
  complex(DP),        intent(inout) :: vec(:)
  real(DP),           intent(in)    :: norm
  type(su3ferm) :: yy
  integer :: igx,ibx,iby,ibz,ithz,itht,itht0
  integer :: ix,iy,iz,ieoxyz,ith,is,ic,isite,ieo

  ieo = sq%ieo

!$OMP PARALLEL DO PRIVATE(igx,ibx,iby,ibz,ithz,itht,itht0,ix,iy,iz,ieoxyz,ic,is,isite,yy)
  do igx=1,NGX
    do ibx=1,NBX
    ix = ibx + (igx-1)*NBX
    do iby=1,NBY
    do ibz=1,NBZ
      iy = iby
      do ithz=1,NTHZ
        iz = ithz + (ibz-1)*NTHZ
        ieoxyz=mod(ipeo+ieo+ix+iy+iz,2)
        do itht=1-ieoxyz,NTHT-ieoxyz
          itht0=itht+ieoxyz

          yy%yr(1,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%x
          yy%yr(2,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%y
          yy%yr(3,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%z
          yy%yr(1,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%w
          yy%yr(2,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%x
          yy%yr(3,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%y
          yy%yr(1,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%z
          yy%yr(2,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%w
          yy%yr(3,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%x
          yy%yr(1,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%y
          yy%yr(2,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%z
          yy%yr(3,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%w
          yy%yi(1,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%x
          yy%yi(2,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%y
          yy%yi(3,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%z
          yy%yi(1,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%w
          yy%yi(2,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%x
          yy%yi(3,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%y
          yy%yi(1,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%z
          yy%yi(2,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%w
          yy%yi(3,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%x
          yy%yi(1,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%y
          yy%yi(2,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%z
          yy%yi(3,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%w

          do is=1,SPIN
          do ic=1,COL
            isite = ic  &
 &            +    (is-1)*COL  &
 &            + (itht0-1)*CLSP  &
 &            +    (iz-1)*CLSP*NTH  &
 &            +    (iy-1)*CLSP*NTH*NZ  &
 &            +    (ix-1)*CLSP*NTH*NZ*NY

            vec(isite) = CMPLX(yy%yr(ic,is),yy%yi(ic,is),kind=KIND(DP))*norm
          enddo
          enddo

        enddo
      enddo
    enddo
    enddo
    enddo
  enddo

  return
end subroutine

subroutine conv_q_d2s(sq,dq,norm)
!
! sq <= dq/|dq|
! norm = |dq|
!
  implicit none
  type(s_wqf_eo_obj),      intent(out) :: sq
  type(field_quark_eo_wg), intent(in)  :: dq
  real(DP),                intent(out) :: norm
  real(DP) :: rtmp
  type(su3ferm) :: yy
  complex(DP) :: ztmp
  integer :: itht,ithz,ibz,iby,ibx,igx,ic,is,itht0
  integer :: ieo,ix,iy,iz,ith,ieoxyz

  ieo = dq%ieo
  sq%ieo = ieo
  sq%ipeo = ipeo

  norm = 0.0_DP
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ic,is,ztmp) REDUCTION(+:norm)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ieo+ix+iy+iz,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      do is=1,SPIN
      do ic=1,COL
        ztmp = dq%s(ith,iz,iy,ix)%y(ic,is)
        norm = norm + REAL(ztmp)**2 + AIMAG(ztmp)**2
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo
!$OMP END PARALLEL DO
#ifndef _singlePU
  call comlib_sumcast(norm)
#endif
  norm = SQRT(norm)

  rtmp = 1.0_DP/norm
!$OMP PARALLEL DO PRIVATE(igx,ibx,iby,ibz,ithz,itht,itht0,ix,iy,iz,ieoxyz,ic,is,ztmp,yy)
  do igx=1,NGX
    sq%g(igx)%ipeo=ipeo
    sq%g(igx)%ieo=ieo
    sq%g(igx)%igx=igx
    do ibx=1,NBX
    ix = ibx + (igx-1)*NBX
    do iby=1,NBY
    do ibz=1,NBZ
      iy=iby
      do ithz=1,NTHZ
        iz=ithz + (ibz-1)*NTHZ
        ieoxyz=mod(ipeo+ieo+ix+iy+iz,2)
        do itht=1-ieoxyz,NTHT-ieoxyz
          itht0=itht+ieoxyz
          do is=1,SPIN
          do ic=1,COL
            ztmp = dq%s(itht,iz,iy,ix)%y(ic,is)*rtmp
            yy%yr(ic,is) =  REAL(ztmp)
            yy%yi(ic,is) = AIMAG(ztmp)
          enddo
          enddo
          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%x = yy%yr(1,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%y = yy%yr(2,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%z = yy%yr(3,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%w = yy%yr(1,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%x = yy%yr(2,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%y = yy%yr(3,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%z = yy%yr(1,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%w = yy%yr(2,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%x = yy%yr(3,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%y = yy%yr(1,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%z = yy%yr(2,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%w = yy%yr(3,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%x = yy%yi(1,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%y = yy%yi(2,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%z = yy%yi(3,1)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%w = yy%yi(1,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%x = yy%yi(2,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%y = yy%yi(3,2)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%z = yy%yi(1,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%w = yy%yi(2,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%x = yy%yi(3,3)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%y = yy%yi(1,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%z = yy%yi(2,4)
          sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%w = yy%yi(3,4)
        enddo
      enddo
    enddo
    enddo
    enddo
  enddo
!$OMP END PARALLEL DO

  call copy_boundary(sq)

  return
end subroutine

subroutine conv_q_s2d(dq,sq,norm)
!
! dq <= sq * norm
!
!
  implicit none
  type(field_quark_eo_wg), intent(out)  :: dq
  type(s_wqf_eo_obj),      intent(in)   :: sq
  real(DP),                intent(in)   :: norm

  type(su3ferm) :: yy
  complex(DP) :: ztmp
  integer :: itht,ithz,ibz,iby,ibx,igx,ic,is,itht0
  integer :: ix,iy,iz,ieoxyz,ieo

  ieo = sq%ieo
  dq%ieo = ieo

!$OMP PARALLEL DO PRIVATE(igx,ibx,iby,ibz,ithz,itht,itht0,ix,iy,iz,ieoxyz,ic,is,yy)
  do igx=1,NGX
    do ibx=1,NBX
    ix = ibx + (igx-1)*NBX
    do iby=1,NBY
    do ibz=1,NBZ
      iy = iby
      do ithz=1,NTHZ
        iz = ithz + (ibz-1)*NTHZ
        ieoxyz=mod(ipeo+ieo+ix+iy+iz,2)
        do itht=1-ieoxyz,NTHT-ieoxyz
          itht0=itht+ieoxyz

          yy%yr(1,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%x
          yy%yr(2,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%y
          yy%yr(3,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%z
          yy%yr(1,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(1)%f(itht0,ithz)%w
          yy%yr(2,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%x
          yy%yr(3,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%y
          yy%yr(1,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%z
          yy%yr(2,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(2)%f(itht0,ithz)%w
          yy%yr(3,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%x
          yy%yr(1,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%y
          yy%yr(2,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%z
          yy%yr(3,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(3)%f(itht0,ithz)%w
          yy%yi(1,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%x
          yy%yi(2,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%y
          yy%yi(3,1) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%z
          yy%yi(1,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(4)%f(itht0,ithz)%w
          yy%yi(2,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%x
          yy%yi(3,2) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%y
          yy%yi(1,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%z
          yy%yi(2,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(5)%f(itht0,ithz)%w
          yy%yi(3,3) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%x
          yy%yi(1,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%y
          yy%yi(2,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%z
          yy%yi(3,4) = sq%g(igx)%blk(ibz,iby,ibx)%thd(6)%f(itht0,ithz)%w
          do is=1,SPIN
          do ic=1,COL
            dq%s(itht,iz,iy,ix)%y(ic,is) = cmplx(yy%yr(ic,is),yy%yi(ic,is),kind=KIND(DP))*norm
          enddo
          enddo
        enddo
      enddo
    enddo
    enddo
    enddo
  enddo
!$OMP END PARALLEL DO
  
  return
end subroutine

subroutine conv_u(su,du)
  implicit none
  type(s_gvf_eo_obj),       intent(out) :: su
  type(vfield_gluon_eo_wg), intent(in)  :: du

  type(su3mat) :: uu
  complex(DP) :: ztmp
  integer :: itht,ithz,ibz,iby,ibx,igx,ic,jc,mu,itht0
  integer :: ix,iy,iz,ieoxyz,ieo
  integer :: itx,ity,ipu
  character(len=256) :: str(0:NPU-1)

  ieo = du%ieo
  su%ieo = ieo
  su%ipeo = ipeo

!$OMP PARALLEL DO PRIVATE(igx,ibx,iby,ibz,ithz,itht,itht0,ix,iy,iz,ieoxyz,ic,jc,mu,ztmp,uu)
  do igx=1,NGX
    su%g(igx)%ipeo=ipeo
    su%g(igx)%ieo=ieo
    su%g(igx)%igx=igx
    do ibx=1,NBX
    ix = ibx + (igx-1)*NBX
    do iby=1,NBY
    do ibz=1,NBZ
      iy = iby
      do mu=1,NDIM
      do ithz=1,NTHZ
        iz = ithz + (ibz-1)*NTHZ
        ieoxyz=mod(ipeo+ieo+ix+iy+iz,2)
        do itht=1-ieoxyz,NTHT-ieoxyz
          itht0=itht+ieoxyz
          do jc=1,COL-1
          do ic=1,COL
            ztmp = du%mu(mu)%s(itht,iz,iy,ix)%u(ic,jc)
            uu%ur(ic,jc) =  REAL(ztmp)
            uu%ui(ic,jc) = AIMAG(ztmp)
          enddo
          enddo
          su%g(igx)%blk(ibz,iby,ibx)%thd(1,mu)%f(itht0,ithz)%x = uu%ur(1,1)
          su%g(igx)%blk(ibz,iby,ibx)%thd(1,mu)%f(itht0,ithz)%y = uu%ur(2,1)
          su%g(igx)%blk(ibz,iby,ibx)%thd(1,mu)%f(itht0,ithz)%z = uu%ur(3,1)
          su%g(igx)%blk(ibz,iby,ibx)%thd(1,mu)%f(itht0,ithz)%w = uu%ur(1,2)
          su%g(igx)%blk(ibz,iby,ibx)%thd(2,mu)%f(itht0,ithz)%x = uu%ur(2,2)
          su%g(igx)%blk(ibz,iby,ibx)%thd(2,mu)%f(itht0,ithz)%y = uu%ur(3,2)
          su%g(igx)%blk(ibz,iby,ibx)%thd(2,mu)%f(itht0,ithz)%z = uu%ui(1,1)
          su%g(igx)%blk(ibz,iby,ibx)%thd(2,mu)%f(itht0,ithz)%w = uu%ui(2,1)
          su%g(igx)%blk(ibz,iby,ibx)%thd(3,mu)%f(itht0,ithz)%x = uu%ui(3,1)
          su%g(igx)%blk(ibz,iby,ibx)%thd(3,mu)%f(itht0,ithz)%y = uu%ui(1,2)
          su%g(igx)%blk(ibz,iby,ibx)%thd(3,mu)%f(itht0,ithz)%z = uu%ui(2,2)
          su%g(igx)%blk(ibz,iby,ibx)%thd(3,mu)%f(itht0,ithz)%w = uu%ui(3,2)
        enddo
      enddo
      enddo
    enddo
    enddo
    enddo
  enddo
!$OMP END PARALLEL DO

  call copy_boundary(su)

  return
end subroutine

#ifdef _CLOVER
subroutine conv_clv(sclv,dclv)
  implicit none
  type(s_clvf_eo_obj),        intent(out) :: sclv
  type(field_clover_term_eo), intent(in)  :: dclv

  complex(DP) :: ztmp
  type(clvmat) :: cc
  integer :: itht,ithz,ibz,iby,ibx,igx,ics,itht0
  integer :: ix,iy,iz,ieoxyz,ieo

  ieo = dclv%ieo
  sclv%ieo = ieo
  sclv%ipeo = ipeo

!$OMP PARALLEL DO PRIVATE(igx,ibx,iby,ibz,ithz,itht0,ix,iy,iz,ieoxyz,ics,cc)
  do igx=1,NGX
    sclv%g(igx)%ipeo=ipeo
    sclv%g(igx)%ieo=ieo
    sclv%g(igx)%igx=igx
    do ibx=1,NBX
    ix = ibx + (igx-1)*NBX
    do iby=1,NBY
    do ibz=1,NBZ
      iy = iby
      do ithz=1,NTHZ
        iz = ithz + (ibz-1)*NTHZ
        ieoxyz=mod(ipeo+ieo+ix+iy+iz,2)
        do itht0=1,NTHT
          do ics=1,CLSPH
            cc%c1r(ics) =  REAL(dclv%l(ics,1,itht0,iz,iy,ix))
            cc%c1i(ics) = AIMAG(dclv%l(ics,1,itht0,iz,iy,ix))
            cc%c2r(ics) =  REAL(dclv%l(ics,2,itht0,iz,iy,ix))
            cc%c2i(ics) = AIMAG(dclv%l(ics,2,itht0,iz,iy,ix))
          enddo
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 1)%f(itht0,ithz)%x = cc%c1r( 1)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 1)%f(itht0,ithz)%y = cc%c1r( 2)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 1)%f(itht0,ithz)%z = cc%c1r( 3)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 1)%f(itht0,ithz)%w = cc%c1r( 4)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 2)%f(itht0,ithz)%x = cc%c1r( 5)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 2)%f(itht0,ithz)%y = cc%c1r( 6)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 2)%f(itht0,ithz)%z = cc%c1r( 7)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 2)%f(itht0,ithz)%w = cc%c1r( 8)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 3)%f(itht0,ithz)%x = cc%c1r( 9)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 3)%f(itht0,ithz)%y = cc%c1r(10)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 3)%f(itht0,ithz)%z = cc%c1r(11)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 3)%f(itht0,ithz)%w = cc%c1r(12)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 4)%f(itht0,ithz)%x = cc%c1r(13)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 4)%f(itht0,ithz)%y = cc%c1r(14)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 4)%f(itht0,ithz)%z = cc%c1r(15)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 4)%f(itht0,ithz)%w = cc%c1r(16)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 5)%f(itht0,ithz)%x = cc%c1r(17)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 5)%f(itht0,ithz)%y = cc%c1r(18)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 5)%f(itht0,ithz)%z = cc%c1r(19)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 5)%f(itht0,ithz)%w = cc%c1r(20)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 6)%f(itht0,ithz)%x = cc%c1r(21)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 6)%f(itht0,ithz)%y = cc%c1i( 1)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 6)%f(itht0,ithz)%z = cc%c1i( 2)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 6)%f(itht0,ithz)%w = cc%c1i( 3)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 7)%f(itht0,ithz)%x = cc%c1i( 4)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 7)%f(itht0,ithz)%y = cc%c1i( 5)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 7)%f(itht0,ithz)%z = cc%c1i( 6)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 7)%f(itht0,ithz)%w = cc%c1i( 7)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 8)%f(itht0,ithz)%x = cc%c1i( 8)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 8)%f(itht0,ithz)%y = cc%c1i( 9)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 8)%f(itht0,ithz)%z = cc%c1i(10)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 8)%f(itht0,ithz)%w = cc%c1i(11)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 9)%f(itht0,ithz)%x = cc%c1i(12)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 9)%f(itht0,ithz)%y = cc%c1i(13)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 9)%f(itht0,ithz)%z = cc%c1i(14)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd( 9)%f(itht0,ithz)%w = cc%c1i(15)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(10)%f(itht0,ithz)%x = cc%c1i(16)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(10)%f(itht0,ithz)%y = cc%c1i(17)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(10)%f(itht0,ithz)%z = cc%c1i(18)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(10)%f(itht0,ithz)%w = cc%c1i(19)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(11)%f(itht0,ithz)%x = cc%c1i(20)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(11)%f(itht0,ithz)%y = cc%c1i(21)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(11)%f(itht0,ithz)%z = cc%c2r( 1)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(11)%f(itht0,ithz)%w = cc%c2r( 2)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(12)%f(itht0,ithz)%x = cc%c2r( 3)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(12)%f(itht0,ithz)%y = cc%c2r( 4)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(12)%f(itht0,ithz)%z = cc%c2r( 5)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(12)%f(itht0,ithz)%w = cc%c2r( 6)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(13)%f(itht0,ithz)%x = cc%c2r( 7)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(13)%f(itht0,ithz)%y = cc%c2r( 8)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(13)%f(itht0,ithz)%z = cc%c2r( 9)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(13)%f(itht0,ithz)%w = cc%c2r(10)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(14)%f(itht0,ithz)%x = cc%c2r(11)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(14)%f(itht0,ithz)%y = cc%c2r(12)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(14)%f(itht0,ithz)%z = cc%c2r(13)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(14)%f(itht0,ithz)%w = cc%c2r(14)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(15)%f(itht0,ithz)%x = cc%c2r(15)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(15)%f(itht0,ithz)%y = cc%c2r(16)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(15)%f(itht0,ithz)%z = cc%c2r(17)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(15)%f(itht0,ithz)%w = cc%c2r(18)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(16)%f(itht0,ithz)%x = cc%c2r(19)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(16)%f(itht0,ithz)%y = cc%c2r(20)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(16)%f(itht0,ithz)%z = cc%c2r(21)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(16)%f(itht0,ithz)%w = cc%c2i( 1)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(17)%f(itht0,ithz)%x = cc%c2i( 2)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(17)%f(itht0,ithz)%y = cc%c2i( 3)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(17)%f(itht0,ithz)%z = cc%c2i( 4)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(17)%f(itht0,ithz)%w = cc%c2i( 5)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(18)%f(itht0,ithz)%x = cc%c2i( 6)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(18)%f(itht0,ithz)%y = cc%c2i( 7)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(18)%f(itht0,ithz)%z = cc%c2i( 8)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(18)%f(itht0,ithz)%w = cc%c2i( 9)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(19)%f(itht0,ithz)%x = cc%c2i(10)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(19)%f(itht0,ithz)%y = cc%c2i(11)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(19)%f(itht0,ithz)%z = cc%c2i(12)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(19)%f(itht0,ithz)%w = cc%c2i(13)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(20)%f(itht0,ithz)%x = cc%c2i(14)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(20)%f(itht0,ithz)%y = cc%c2i(15)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(20)%f(itht0,ithz)%z = cc%c2i(16)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(20)%f(itht0,ithz)%w = cc%c2i(17)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(21)%f(itht0,ithz)%x = cc%c2i(18)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(21)%f(itht0,ithz)%y = cc%c2i(19)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(21)%f(itht0,ithz)%z = cc%c2i(20)
          sclv%g(igx)%blk(ibz,iby,ibx)%thd(21)%f(itht0,ithz)%w = cc%c2i(21)
        enddo
      enddo
    enddo
    enddo
    enddo
  enddo
!$OMP END PARALLEL DO

  call copy_boundary(sclv)
  
  return
end subroutine
#endif

end module
