#include "config.h"
#include "gpu_config.h"

#if _NRAS > 0      
#define _PREC_
#else
#undef _PREC_
#endif

module single_solver_class
  use timer_class
  use comlib
  use lattice_class
  use error_class
  use field_fermion_class
  use field_gauge_class
  use field_clover_class
  use single_field_class
  implicit none
  private

  public :: single_solver
  public :: new
  public :: delete
  public :: assign_inv_mult_eoprec_wd

  integer,  parameter :: MAX_ITER = 5000
  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP
  logical, save :: m_is_single_solver_initialized = .false.

  type single_solver
    private
    type(s_gvf_obj),  allocatable :: su
#ifdef _CLOVER
    type(s_clvf_obj), allocatable :: sfcl
#endif
    real(DP), public :: err_norm = 0.0_SP
    real(DP), public :: tol  = 1.0e-4_SP
    integer,  public :: iter = 0
    integer,  public :: iout = 6
  end type

  interface new
    module procedure new_single_solver
  end interface

  interface delete
    module procedure delete_single_solver
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_sgl
  end interface

contains

subroutine new_single_solver(this,u,fclinv)
  implicit none
  type(single_solver),               intent(inout) :: this
  type(vfield_gluon_wg),             intent(in) :: u
  type(field_clover_term), optional, intent(in) :: fclinv
  logical :: is_clover

  if (present(fclinv)) then
    is_clover = .true.
  else
    is_clover = .false.
  endif

  this%iter = 0

  if (allocated(this%su)) deallocate(this%su)
  allocate(this%su)
  call new(this%su)

#ifdef _CLOVER
  if (is_clover) then
    if (allocated(this%sfcl)) deallocate(this%sfcl)
    allocate(this%sfcl)
    call new(this%sfcl)
  endif
#endif
  call conv_u(this%su,u)
#ifdef _CLOVER
  if (is_clover) then
    call conv_clv(this%sfcl,fclinv)
  endif
#endif

  m_is_single_solver_initialized = .true.

  return
end subroutine

subroutine delete_single_solver(this)
  implicit none
  type(single_solver), intent(inout) :: this

  this%iter = 0

  if (allocated(this%su))   deallocate(this%su)
#ifdef _CLOVER
  if (allocated(this%sfcl)) deallocate(this%sfcl)
#endif

  m_is_single_solver_initialized = .false.

  return
end subroutine

subroutine update_xr_blk(alp,xe,re,pe,qe,rnorm2)
!
!     x = x + alpha p 
!     r = r - alpha q
!     rnorm2 = |r|^2
!
  implicit none
  complex(SP), intent(in) :: alp
  type(s_wqf_eo_blk_obj), intent(inout) :: xe,re
  type(s_wqf_eo_blk_obj), intent(in)    :: pe,qe
  real(SP), intent(out) :: rnorm2
  real(SP) :: ar,ai,rtmp0
  integer :: ix,iy,iz,ieoxyz,ics,jcs
  integer :: ibx,iby,ibz,itht,ithz
  character(len=CHARLEN) :: str
  if (xe%ieo /= re%ieo .or. xe%igx /= re%igx .or.   &
 &    xe%ieo /= qe%ieo .or. xe%igx /= qe%igx .or.   &
 &    pe%ieo /= qe%ieo .or. pe%igx /= qe%igx .or.   &
 &    pe%ieo /= re%ieo .or. pe%igx /= re%igx ) then
    call error_message("Index error in update_xr_blk xe/=re/=pe/=qe.")
    write(str,'("xe%ieo =",I2)')xe%ieo; call error_message(TRIM(str))
    write(str,'("re%ieo =",I2)')re%ieo; call error_message(TRIM(str))
    write(str,'("pe%ieo =",I2)')pe%ieo; call error_message(TRIM(str))
    write(str,'("qe%ieo =",I2)')qe%ieo; call error_message(TRIM(str))
    write(str,'("xe%igx =",I2)')xe%igx; call error_message(TRIM(str))
    write(str,'("re%igx =",I2)')re%igx; call error_message(TRIM(str))
    write(str,'("pe%igx =",I2)')pe%igx; call error_message(TRIM(str))
    write(str,'("qe%igx =",I2)')qe%igx; call error_stop(TRIM(str))
  endif
  ar =  real(alp)
  ai = aimag(alp)
  rtmp0 = (0.0_SP,0.0_SP)
  do ibx=1,NBX
  do iby=1,NBY
  do ibz=1,NBZ
    do ithz=1,NTHZ
      ix=ibx+(xe%igx-1)*NBX
      iy=iby
      iz=ithz+(ibz-1)*NTHZ
      ieoxyz=mod(ipeo+ix+iy+iz+xe%ieo,2)
      do ics=1,COL
      jcs=ics+COL
      do itht=1,NTHT
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  -ai*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  -ai*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  -ai*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  -ai*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w

          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  +ai*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x
          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  +ai*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y
          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  +ai*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z
          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  +ar*pe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  +ai*pe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w

          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  +ai*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x
          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  +ai*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y
          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  +ai*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z
          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  +ai*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w

          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  -ai*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x
          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  -ai*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y
          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  -ai*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z
          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  -ar*qe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  -ai*qe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w

          rtmp0=rtmp0 + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w**2
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo
  rnorm2 = rtmp0
  return
end subroutine

subroutine update_xr_eo(alp,xe,re,pe,qe,rnorm2)
!
!     x = x + alpha p 
!     r = r - alpha q
!     rnorm2 = |r|^2
!
  implicit none
  complex(SP), intent(in) :: alp
  type(s_wqf_eo_obj), intent(inout) :: xe,re
  type(s_wqf_eo_obj), intent(in)    :: pe,qe
  real(SP), intent(out) :: rnorm2
  real(SP) :: rnorm2_g
  integer :: igx
  rnorm2 = 0.0_SP
!$OMP PARALLEL DO PRIVATE(rnorm2_g) REDUCTION(+:rnorm2)
  do igx=1,NGX
    call update_xr_blk(alp,xe%g(igx),re%g(igx),pe%g(igx),qe%g(igx),rnorm2_g)
    rnorm2 = rnorm2 + rnorm2_g
  enddo
  return
end subroutine

subroutine update_xr2_blk(alp,xe,re,ve,rnorm2)
!
!     x = x + alpha r 
!     r = r - alpha v
!     rnorm2 = |r|^2
!
  implicit none
  complex(SP), intent(in) :: alp
  type(s_wqf_eo_blk_obj), intent(inout) :: xe,re
  type(s_wqf_eo_blk_obj), intent(in)    :: ve
  real(SP), intent(out) :: rnorm2
  real(SP) :: ar,ai,rtmp0
  integer :: ix,iy,iz,ieoxyz,ics,jcs
  integer :: ibx,iby,ibz,itht,ithz
  if (xe%ieo /= re%ieo .or. xe%igx /= re%igx .or.   &
 &    xe%ieo /= ve%ieo .or. xe%igx /= ve%igx .or.   &
 &    re%ieo /= ve%ieo .or. re%igx /= ve%igx ) then
    call error_stop("index (igx,ieo) error in update_xr2_blk xe/=re/=ve.")
  endif
  ar =  real(alp)
  ai = aimag(alp)
  rtmp0 = (0.0_SP,0.0_SP)
  do ibx=1,NBX
  do iby=1,NBY
  do ibz=1,NBZ
    do ithz=1,NTHZ
      ix=ibx+(xe%igx-1)*NBX
      iy=iby
      iz=ithz+(ibz-1)*NTHZ
      ieoxyz=mod(ipeo+ix+iy+iz+xe%ieo,2)
      do ics=1,COL
      jcs=ics+COL
      do itht=1,NTHT
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  -ai*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  -ai*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  -ai*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z
          xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                     =xe%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  -ai*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w

          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  +ai*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x
          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  +ai*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y
          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  +ai*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z
          xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                     =xe%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  +ar*re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  +ai*re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w

          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x   &
 &                  +ai*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x
          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y   &
 &                  +ai*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y
          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z   &
 &                  +ai*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z
          re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                     =re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w   &
 &                  +ai*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w

          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x   &
 &                  -ai*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x
          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y   &
 &                  -ai*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y
          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z   &
 &                  -ai*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z
          re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                     =re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  -ar*ve%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w   &
 &                  -ai*ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w

          rtmp0=rtmp0 + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%x**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%y**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%z**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w**2   &
 &                    + re%blk(ibz,iby,ibx)%thd(jcs)%f(itht,ithz)%w**2
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo
  rnorm2 = rtmp0
  return
end subroutine

subroutine update_xr2_eo(alp,xe,re,ve,rnorm2)
!
!     x = x + alpha r 
!     r = r - alpha v
!     rnorm2 = |r|^2
!
  implicit none
  complex(SP), intent(in) :: alp
  type(s_wqf_eo_obj), intent(inout) :: xe,re
  type(s_wqf_eo_obj), intent(in)    :: ve
  real(SP), intent(out) :: rnorm2
  real(SP) :: rnorm2_g
  integer :: igx
  rnorm2 = 0.0_SP
!$OMP PARALLEL DO PRIVATE(rnorm2_g) REDUCTION(+:rnorm2)
  do igx=1,NGX
    call update_xr2_blk(alp,xe%g(igx),re%g(igx),ve%g(igx),rnorm2_g)
    rnorm2 = rnorm2 + rnorm2_g
  enddo
  return
end subroutine

subroutine s_pack_u(itht,ithz,mu,u,uu)
    implicit none
    integer, intent(in) :: itht,ithz,mu
    type(s_gvf_eo_thd_obj), intent(in) :: u
    type(su3mat), intent(out) :: uu
    uu%ur(1,1) = u%thd(1,mu)%f(itht,ithz)%x
    uu%ur(2,1) = u%thd(1,mu)%f(itht,ithz)%y
    uu%ur(3,1) = u%thd(1,mu)%f(itht,ithz)%z
    uu%ur(1,2) = u%thd(1,mu)%f(itht,ithz)%w
    uu%ur(2,2) = u%thd(2,mu)%f(itht,ithz)%x
    uu%ur(3,2) = u%thd(2,mu)%f(itht,ithz)%y
    uu%ui(1,1) = u%thd(2,mu)%f(itht,ithz)%z
    uu%ui(2,1) = u%thd(2,mu)%f(itht,ithz)%w
    uu%ui(3,1) = u%thd(3,mu)%f(itht,ithz)%x
    uu%ui(1,2) = u%thd(3,mu)%f(itht,ithz)%y
    uu%ui(2,2) = u%thd(3,mu)%f(itht,ithz)%z
    uu%ui(3,2) = u%thd(3,mu)%f(itht,ithz)%w
    return
end subroutine

subroutine s_pack_y(itht,ithz,y,yy)
    implicit none
    integer, intent(in) :: itht,ithz
    type(s_wqf_eo_thd_obj), intent(in) :: y
    type(su3ferm), intent(out) :: yy
    yy%yr(1,1) = y%thd(1)%f(itht,ithz)%x
    yy%yr(2,1) = y%thd(1)%f(itht,ithz)%y
    yy%yr(3,1) = y%thd(1)%f(itht,ithz)%z
    yy%yr(1,2) = y%thd(1)%f(itht,ithz)%w
    yy%yr(2,2) = y%thd(2)%f(itht,ithz)%x
    yy%yr(3,2) = y%thd(2)%f(itht,ithz)%y
    yy%yr(1,3) = y%thd(2)%f(itht,ithz)%z
    yy%yr(2,3) = y%thd(2)%f(itht,ithz)%w
    yy%yr(3,3) = y%thd(3)%f(itht,ithz)%x
    yy%yr(1,4) = y%thd(3)%f(itht,ithz)%y
    yy%yr(2,4) = y%thd(3)%f(itht,ithz)%z
    yy%yr(3,4) = y%thd(3)%f(itht,ithz)%w
    yy%yi(1,1) = y%thd(4)%f(itht,ithz)%x
    yy%yi(2,1) = y%thd(4)%f(itht,ithz)%y
    yy%yi(3,1) = y%thd(4)%f(itht,ithz)%z
    yy%yi(1,2) = y%thd(4)%f(itht,ithz)%w
    yy%yi(2,2) = y%thd(5)%f(itht,ithz)%x
    yy%yi(3,2) = y%thd(5)%f(itht,ithz)%y
    yy%yi(1,3) = y%thd(5)%f(itht,ithz)%z
    yy%yi(2,3) = y%thd(5)%f(itht,ithz)%w
    yy%yi(3,3) = y%thd(6)%f(itht,ithz)%x
    yy%yi(1,4) = y%thd(6)%f(itht,ithz)%y
    yy%yi(2,4) = y%thd(6)%f(itht,ithz)%z
    yy%yi(3,4) = y%thd(6)%f(itht,ithz)%w
    return
end subroutine

subroutine s_unpack_y(yy,itht,ithz,y)
    implicit none
    integer, intent(in) :: itht,ithz
    type(su3ferm), intent(in) :: yy
    type(s_wqf_eo_thd_obj), intent(inout) :: y
    y%thd(1)%f(itht,ithz)%x = yy%yr(1,1)
    y%thd(1)%f(itht,ithz)%y = yy%yr(2,1)
    y%thd(1)%f(itht,ithz)%z = yy%yr(3,1)
    y%thd(1)%f(itht,ithz)%w = yy%yr(1,2)
    y%thd(2)%f(itht,ithz)%x = yy%yr(2,2)
    y%thd(2)%f(itht,ithz)%y = yy%yr(3,2)
    y%thd(2)%f(itht,ithz)%z = yy%yr(1,3)
    y%thd(2)%f(itht,ithz)%w = yy%yr(2,3)
    y%thd(3)%f(itht,ithz)%x = yy%yr(3,3)
    y%thd(3)%f(itht,ithz)%y = yy%yr(1,4)
    y%thd(3)%f(itht,ithz)%z = yy%yr(2,4)
    y%thd(3)%f(itht,ithz)%w = yy%yr(3,4)
    y%thd(4)%f(itht,ithz)%x = yy%yi(1,1)
    y%thd(4)%f(itht,ithz)%y = yy%yi(2,1)
    y%thd(4)%f(itht,ithz)%z = yy%yi(3,1)
    y%thd(4)%f(itht,ithz)%w = yy%yi(1,2)
    y%thd(5)%f(itht,ithz)%x = yy%yi(2,2)
    y%thd(5)%f(itht,ithz)%y = yy%yi(3,2)
    y%thd(5)%f(itht,ithz)%z = yy%yi(1,3)
    y%thd(5)%f(itht,ithz)%w = yy%yi(2,3)
    y%thd(6)%f(itht,ithz)%x = yy%yi(3,3)
    y%thd(6)%f(itht,ithz)%y = yy%yi(1,4)
    y%thd(6)%f(itht,ithz)%z = yy%yi(2,4)
    y%thd(6)%f(itht,ithz)%w = yy%yi(3,4)
    return
end subroutine

#ifdef _CLOVER
subroutine s_pack_clv(itht,ithz,c,cc)
    implicit none
    integer, intent(in) :: itht,ithz
    type(s_clvf_eo_thd_obj), intent(in) :: c
    type(clvmat), intent(out) :: cc
    cc%c1r( 1) = c%thd( 1)%f(itht,ithz)%x
    cc%c1r( 2) = c%thd( 1)%f(itht,ithz)%y
    cc%c1r( 3) = c%thd( 1)%f(itht,ithz)%z
    cc%c1r( 4) = c%thd( 1)%f(itht,ithz)%w
    cc%c1r( 5) = c%thd( 2)%f(itht,ithz)%x
    cc%c1r( 6) = c%thd( 2)%f(itht,ithz)%y
    cc%c1r( 7) = c%thd( 2)%f(itht,ithz)%z
    cc%c1r( 8) = c%thd( 2)%f(itht,ithz)%w
    cc%c1r( 9) = c%thd( 3)%f(itht,ithz)%x
    cc%c1r(10) = c%thd( 3)%f(itht,ithz)%y
    cc%c1r(11) = c%thd( 3)%f(itht,ithz)%z
    cc%c1r(12) = c%thd( 3)%f(itht,ithz)%w
    cc%c1r(13) = c%thd( 4)%f(itht,ithz)%x
    cc%c1r(14) = c%thd( 4)%f(itht,ithz)%y
    cc%c1r(15) = c%thd( 4)%f(itht,ithz)%z
    cc%c1r(16) = c%thd( 4)%f(itht,ithz)%w
    cc%c1r(17) = c%thd( 5)%f(itht,ithz)%x
    cc%c1r(18) = c%thd( 5)%f(itht,ithz)%y
    cc%c1r(19) = c%thd( 5)%f(itht,ithz)%z
    cc%c1r(20) = c%thd( 5)%f(itht,ithz)%w
    cc%c1r(21) = c%thd( 6)%f(itht,ithz)%x

    cc%c1i( 1) = c%thd( 6)%f(itht,ithz)%y
    cc%c1i( 2) = c%thd( 6)%f(itht,ithz)%z
    cc%c1i( 3) = c%thd( 6)%f(itht,ithz)%w
    cc%c1i( 4) = c%thd( 7)%f(itht,ithz)%x
    cc%c1i( 5) = c%thd( 7)%f(itht,ithz)%y
    cc%c1i( 6) = c%thd( 7)%f(itht,ithz)%z
    cc%c1i( 7) = c%thd( 7)%f(itht,ithz)%w
    cc%c1i( 8) = c%thd( 8)%f(itht,ithz)%x
    cc%c1i( 9) = c%thd( 8)%f(itht,ithz)%y
    cc%c1i(10) = c%thd( 8)%f(itht,ithz)%z
    cc%c1i(11) = c%thd( 8)%f(itht,ithz)%w
    cc%c1i(12) = c%thd( 9)%f(itht,ithz)%x
    cc%c1i(13) = c%thd( 9)%f(itht,ithz)%y
    cc%c1i(14) = c%thd( 9)%f(itht,ithz)%z
    cc%c1i(15) = c%thd( 9)%f(itht,ithz)%w
    cc%c1i(16) = c%thd(10)%f(itht,ithz)%x
    cc%c1i(17) = c%thd(10)%f(itht,ithz)%y
    cc%c1i(18) = c%thd(10)%f(itht,ithz)%z
    cc%c1i(19) = c%thd(10)%f(itht,ithz)%w
    cc%c1i(20) = c%thd(11)%f(itht,ithz)%x
    cc%c1i(21) = c%thd(11)%f(itht,ithz)%y

    cc%c2r( 1) = c%thd(11)%f(itht,ithz)%z
    cc%c2r( 2) = c%thd(11)%f(itht,ithz)%w
    cc%c2r( 3) = c%thd(12)%f(itht,ithz)%x
    cc%c2r( 4) = c%thd(12)%f(itht,ithz)%y
    cc%c2r( 5) = c%thd(12)%f(itht,ithz)%z
    cc%c2r( 6) = c%thd(12)%f(itht,ithz)%w
    cc%c2r( 7) = c%thd(13)%f(itht,ithz)%x
    cc%c2r( 8) = c%thd(13)%f(itht,ithz)%y
    cc%c2r( 9) = c%thd(13)%f(itht,ithz)%z
    cc%c2r(10) = c%thd(13)%f(itht,ithz)%w
    cc%c2r(11) = c%thd(14)%f(itht,ithz)%x
    cc%c2r(12) = c%thd(14)%f(itht,ithz)%y
    cc%c2r(13) = c%thd(14)%f(itht,ithz)%z
    cc%c2r(14) = c%thd(14)%f(itht,ithz)%w
    cc%c2r(15) = c%thd(15)%f(itht,ithz)%x
    cc%c2r(16) = c%thd(15)%f(itht,ithz)%y
    cc%c2r(17) = c%thd(15)%f(itht,ithz)%z
    cc%c2r(18) = c%thd(15)%f(itht,ithz)%w
    cc%c2r(19) = c%thd(16)%f(itht,ithz)%x
    cc%c2r(20) = c%thd(16)%f(itht,ithz)%y
    cc%c2r(21) = c%thd(16)%f(itht,ithz)%z

    cc%c2i( 1) = c%thd(16)%f(itht,ithz)%w
    cc%c2i( 2) = c%thd(17)%f(itht,ithz)%x
    cc%c2i( 3) = c%thd(17)%f(itht,ithz)%y
    cc%c2i( 4) = c%thd(17)%f(itht,ithz)%z
    cc%c2i( 5) = c%thd(17)%f(itht,ithz)%w
    cc%c2i( 6) = c%thd(18)%f(itht,ithz)%x
    cc%c2i( 7) = c%thd(18)%f(itht,ithz)%y
    cc%c2i( 8) = c%thd(18)%f(itht,ithz)%z
    cc%c2i( 9) = c%thd(18)%f(itht,ithz)%w
    cc%c2i(10) = c%thd(19)%f(itht,ithz)%x
    cc%c2i(11) = c%thd(19)%f(itht,ithz)%y
    cc%c2i(12) = c%thd(19)%f(itht,ithz)%z
    cc%c2i(13) = c%thd(19)%f(itht,ithz)%w
    cc%c2i(14) = c%thd(20)%f(itht,ithz)%x
    cc%c2i(15) = c%thd(20)%f(itht,ithz)%y
    cc%c2i(16) = c%thd(20)%f(itht,ithz)%z
    cc%c2i(17) = c%thd(20)%f(itht,ithz)%w
    cc%c2i(18) = c%thd(21)%f(itht,ithz)%x
    cc%c2i(19) = c%thd(21)%f(itht,ithz)%y
    cc%c2i(20) = c%thd(21)%f(itht,ithz)%z
    cc%c2i(21) = c%thd(21)%f(itht,ithz)%w
    return
end subroutine
#endif

subroutine s_load_u(uu,u)
  implicit none
  type(su3mat),    intent(in) :: uu
  type(su3matful), intent(out):: u
  u%ur(1:COL,1:COL-1) = uu%ur(1:COL,1:COL-1)
  u%ui(1:COL,1:COL-1) = uu%ui(1:COL,1:COL-1)
  u%ur(1,COL) = uu%ur(2,1)*uu%ur(3,2)-uu%ui(2,1)*uu%ui(3,2)    &
 &            -(uu%ur(3,1)*uu%ur(2,2)-uu%ui(3,1)*uu%ui(2,2))
  u%ur(2,COL) = uu%ur(3,1)*uu%ur(1,2)-uu%ui(3,1)*uu%ui(1,2)    &
 &            -(uu%ur(1,1)*uu%ur(3,2)-uu%ui(1,1)*uu%ui(3,2))
  u%ur(3,COL) = uu%ur(1,1)*uu%ur(2,2)-uu%ui(1,1)*uu%ui(2,2)    &
 &            -(uu%ur(2,1)*uu%ur(1,2)-uu%ui(2,1)*uu%ui(1,2))
  u%ui(1,COL) = uu%ur(3,1)*uu%ui(2,2)+uu%ui(3,1)*uu%ur(2,2)    &
 &            -(uu%ur(2,1)*uu%ui(3,2)+uu%ui(2,1)*uu%ur(3,2))
  u%ui(2,COL) = uu%ur(1,1)*uu%ui(3,2)+uu%ui(1,1)*uu%ur(3,2)    &
 &            -(uu%ur(3,1)*uu%ui(1,2)+uu%ui(3,1)*uu%ur(1,2))
  u%ui(3,COL) = uu%ur(2,1)*uu%ui(1,2)+uu%ui(2,1)*uu%ur(1,2)    &
 &            -(uu%ur(1,1)*uu%ui(2,2)+uu%ui(1,1)*uu%ur(2,2))
  return
end subroutine

subroutine s_mult_u(uu,y)
  implicit none
  type(su3mat),  intent(in)    :: uu
  type(su3ferm), intent(inout) :: y ! spin=(1,2):input, spin=(3,4):output
  integer :: i
  type(su3matful) :: u
  call s_load_u(uu,u)
  do i=1,COL
  y%yr(i,3) = u%ur(i,1)*y%yr(1,1)    &
 &           +u%ur(i,2)*y%yr(2,1)    &
 &           +u%ur(i,3)*y%yr(3,1)    &
 &           -u%ui(i,1)*y%yi(1,1)    &
 &           -u%ui(i,2)*y%yi(2,1)    &
 &           -u%ui(i,3)*y%yi(3,1)
  y%yi(i,3) = u%ur(i,1)*y%yi(1,1)    &
 &           +u%ur(i,2)*y%yi(2,1)    &
 &           +u%ur(i,3)*y%yi(3,1)    &
 &           +u%ui(i,1)*y%yr(1,1)    &
 &           +u%ui(i,2)*y%yr(2,1)    &
 &           +u%ui(i,3)*y%yr(3,1)
  y%yr(i,4) = u%ur(i,1)*y%yr(1,2)    &
 &           +u%ur(i,2)*y%yr(2,2)    &
 &           +u%ur(i,3)*y%yr(3,2)    &
 &           -u%ui(i,1)*y%yi(1,2)    &
 &           -u%ui(i,2)*y%yi(2,2)    &
 &           -u%ui(i,3)*y%yi(3,2)
  y%yi(i,4) = u%ur(i,1)*y%yi(1,2)    &
 &           +u%ur(i,2)*y%yi(2,2)    &
 &           +u%ur(i,3)*y%yi(3,2)    &
 &           +u%ui(i,1)*y%yr(1,2)    &
 &           +u%ui(i,2)*y%yr(2,2)    &
 &           +u%ui(i,3)*y%yr(3,2)
  enddo
  return
end subroutine

subroutine s_mult_udag(uu,y)
  implicit none
  type(su3mat),  intent(in)   :: uu
  type(su3ferm), intent(inout):: y ! spin=(1,2):input, spin=(3,4):output
  integer :: i
  type(su3matful) :: u
  call s_load_u(uu,u)
  do i=1,COL
  y%yr(i,3) = u%ur(1,i)*y%yr(1,1)    &
 &           +u%ur(2,i)*y%yr(2,1)    &
 &           +u%ur(3,i)*y%yr(3,1)    &
 &           +u%ui(1,i)*y%yi(1,1)    &
 &           +u%ui(2,i)*y%yi(2,1)    &
 &           +u%ui(3,i)*y%yi(3,1)
  y%yi(i,3) = u%ur(1,i)*y%yi(1,1)    &
 &           +u%ur(2,i)*y%yi(2,1)    &
 &           +u%ur(3,i)*y%yi(3,1)    &
 &           -u%ui(1,i)*y%yr(1,1)    &
 &           -u%ui(2,i)*y%yr(2,1)    &
 &           -u%ui(3,i)*y%yr(3,1)
  y%yr(i,4) = u%ur(1,i)*y%yr(1,2)    &
 &           +u%ur(2,i)*y%yr(2,2)    &
 &           +u%ur(3,i)*y%yr(3,2)    &
 &           +u%ui(1,i)*y%yi(1,2)    &
 &           +u%ui(2,i)*y%yi(2,2)    &
 &           +u%ui(3,i)*y%yi(3,2)
  y%yi(i,4) = u%ur(1,i)*y%yi(1,2)    &
 &           +u%ur(2,i)*y%yi(2,2)    &
 &           +u%ur(3,i)*y%yi(3,2)    &
 &           -u%ui(1,i)*y%yr(1,2)    &
 &           -u%ui(2,i)*y%yr(2,2)    &
 &           -u%ui(3,i)*y%yr(3,2)
  enddo
  return
end subroutine

subroutine s_mult_forw_t(u,y,dy)
  implicit none
  type(su3mat),  intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,3)
  wy%yi(1:COL,1) = y%yi(1:COL,3)
  wy%yr(1:COL,2) = y%yr(1:COL,4)
  wy%yi(1:COL,2) = y%yi(1:COL,4)
  call s_mult_u(u,wy)
  dy%yr(1:COL,3) = 2*wy%yr(1:COL,3)
  dy%yi(1:COL,3) = 2*wy%yi(1:COL,3)
  dy%yr(1:COL,4) = 2*wy%yr(1:COL,4)
  dy%yi(1:COL,4) = 2*wy%yi(1:COL,4)
  return
end subroutine

subroutine s_mult_back_t(u,y,dy)
  implicit none
  type(su3mat),   intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,1)
  wy%yi(1:COL,1) = y%yi(1:COL,1)
  wy%yr(1:COL,2) = y%yr(1:COL,2)
  wy%yi(1:COL,2) = y%yi(1:COL,2)
  call s_mult_udag(u,wy)
  dy%yr(1:COL,1) = 2*wy%yr(1:COL,3)
  dy%yi(1:COL,1) = 2*wy%yi(1:COL,3)
  dy%yr(1:COL,2) = 2*wy%yr(1:COL,4)
  dy%yi(1:COL,2) = 2*wy%yi(1:COL,4)
  return
end subroutine

subroutine s_mult_forw_z(u,y,dy)
  implicit none
  type(su3mat),   intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,1) - y%yi(1:COL,3)
  wy%yi(1:COL,1) = y%yi(1:COL,1) + y%yr(1:COL,3)
  wy%yr(1:COL,2) = y%yr(1:COL,2) + y%yi(1:COL,4)
  wy%yi(1:COL,2) = y%yi(1:COL,2) - y%yr(1:COL,4)
  call s_mult_u(u,wy)
  dy%yr(1:COL,1) = dy%yr(1:COL,1) + wy%yr(1:COL,3)
  dy%yi(1:COL,1) = dy%yi(1:COL,1) + wy%yi(1:COL,3)
  dy%yr(1:COL,2) = dy%yr(1:COL,2) + wy%yr(1:COL,4)
  dy%yi(1:COL,2) = dy%yi(1:COL,2) + wy%yi(1:COL,4)
  dy%yr(1:COL,3) = dy%yr(1:COL,3) + wy%yi(1:COL,3)
  dy%yi(1:COL,3) = dy%yi(1:COL,3) - wy%yr(1:COL,3)
  dy%yr(1:COL,4) = dy%yr(1:COL,4) - wy%yi(1:COL,4)
  dy%yi(1:COL,4) = dy%yi(1:COL,4) + wy%yr(1:COL,4)
  return
end subroutine

subroutine s_mult_back_z(u,y,dy)
  implicit none
  type(su3mat),   intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,1) + y%yi(1:COL,3)
  wy%yi(1:COL,1) = y%yi(1:COL,1) - y%yr(1:COL,3)
  wy%yr(1:COL,2) = y%yr(1:COL,2) - y%yi(1:COL,4)
  wy%yi(1:COL,2) = y%yi(1:COL,2) + y%yr(1:COL,4)
  call s_mult_udag(u,wy)
  dy%yr(1:COL,1) = dy%yr(1:COL,1) + wy%yr(1:COL,3)
  dy%yi(1:COL,1) = dy%yi(1:COL,1) + wy%yi(1:COL,3)
  dy%yr(1:COL,2) = dy%yr(1:COL,2) + wy%yr(1:COL,4)
  dy%yi(1:COL,2) = dy%yi(1:COL,2) + wy%yi(1:COL,4)
  dy%yr(1:COL,3) = dy%yr(1:COL,3) - wy%yi(1:COL,3)
  dy%yi(1:COL,3) = dy%yi(1:COL,3) + wy%yr(1:COL,3)
  dy%yr(1:COL,4) = dy%yr(1:COL,4) + wy%yi(1:COL,4)
  dy%yi(1:COL,4) = dy%yi(1:COL,4) - wy%yr(1:COL,4)
  return
end subroutine

subroutine s_mult_forw_y(u,y,dy)
  implicit none
  type(su3mat),   intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,1) + y%yr(1:COL,4)
  wy%yi(1:COL,1) = y%yi(1:COL,1) + y%yi(1:COL,4)
  wy%yr(1:COL,2) = y%yr(1:COL,2) - y%yr(1:COL,3)
  wy%yi(1:COL,2) = y%yi(1:COL,2) - y%yi(1:COL,3)
  call s_mult_u(u,wy)
  dy%yr(1:COL,1) = dy%yr(1:COL,1) + wy%yr(1:COL,3)
  dy%yi(1:COL,1) = dy%yi(1:COL,1) + wy%yi(1:COL,3)
  dy%yr(1:COL,2) = dy%yr(1:COL,2) + wy%yr(1:COL,4)
  dy%yi(1:COL,2) = dy%yi(1:COL,2) + wy%yi(1:COL,4)
  dy%yr(1:COL,3) = dy%yr(1:COL,3) - wy%yr(1:COL,4)
  dy%yi(1:COL,3) = dy%yi(1:COL,3) - wy%yi(1:COL,4)
  dy%yr(1:COL,4) = dy%yr(1:COL,4) + wy%yr(1:COL,3)
  dy%yi(1:COL,4) = dy%yi(1:COL,4) + wy%yi(1:COL,3)
  return
end subroutine

subroutine s_mult_back_y(u,y,dy)
  implicit none
  type(su3mat),   intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,1) - y%yr(1:COL,4)
  wy%yi(1:COL,1) = y%yi(1:COL,1) - y%yi(1:COL,4)
  wy%yr(1:COL,2) = y%yr(1:COL,2) + y%yr(1:COL,3)
  wy%yi(1:COL,2) = y%yi(1:COL,2) + y%yi(1:COL,3)
  call s_mult_udag(u,wy)
  dy%yr(1:COL,1) = dy%yr(1:COL,1) + wy%yr(1:COL,3)
  dy%yi(1:COL,1) = dy%yi(1:COL,1) + wy%yi(1:COL,3)
  dy%yr(1:COL,2) = dy%yr(1:COL,2) + wy%yr(1:COL,4)
  dy%yi(1:COL,2) = dy%yi(1:COL,2) + wy%yi(1:COL,4)
  dy%yr(1:COL,3) = dy%yr(1:COL,3) + wy%yr(1:COL,4)
  dy%yi(1:COL,3) = dy%yi(1:COL,3) + wy%yi(1:COL,4)
  dy%yr(1:COL,4) = dy%yr(1:COL,4) - wy%yr(1:COL,3)
  dy%yi(1:COL,4) = dy%yi(1:COL,4) - wy%yi(1:COL,3)
  return
end subroutine

subroutine s_mult_forw_x(u,y,dy)
  implicit none
  type(su3mat),   intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,1) - y%yi(1:COL,4)
  wy%yi(1:COL,1) = y%yi(1:COL,1) + y%yr(1:COL,4)
  wy%yr(1:COL,2) = y%yr(1:COL,2) - y%yi(1:COL,3)
  wy%yi(1:COL,2) = y%yi(1:COL,2) + y%yr(1:COL,3)
  call s_mult_u(u,wy)
  dy%yr(1:COL,1) = dy%yr(1:COL,1) + wy%yr(1:COL,3)
  dy%yi(1:COL,1) = dy%yi(1:COL,1) + wy%yi(1:COL,3)
  dy%yr(1:COL,2) = dy%yr(1:COL,2) + wy%yr(1:COL,4)
  dy%yi(1:COL,2) = dy%yi(1:COL,2) + wy%yi(1:COL,4)
  dy%yr(1:COL,3) = dy%yr(1:COL,3) + wy%yi(1:COL,4)
  dy%yi(1:COL,3) = dy%yi(1:COL,3) - wy%yr(1:COL,4)
  dy%yr(1:COL,4) = dy%yr(1:COL,4) + wy%yi(1:COL,3)
  dy%yi(1:COL,4) = dy%yi(1:COL,4) - wy%yr(1:COL,3)
  return
end subroutine

subroutine s_mult_back_x(u,y,dy)
  implicit none
  type(su3mat),  intent(in) :: u
  type(su3ferm), intent(in) :: y
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: wy
  wy%yr(1:COL,1) = y%yr(1:COL,1) + y%yi(1:COL,4)
  wy%yi(1:COL,1) = y%yi(1:COL,1) - y%yr(1:COL,4)
  wy%yr(1:COL,2) = y%yr(1:COL,2) + y%yi(1:COL,3)
  wy%yi(1:COL,2) = y%yi(1:COL,2) - y%yr(1:COL,3)
  call s_mult_udag(u,wy)
  dy%yr(1:COL,1) = dy%yr(1:COL,1) + wy%yr(1:COL,3)
  dy%yi(1:COL,1) = dy%yi(1:COL,1) + wy%yi(1:COL,3)
  dy%yr(1:COL,2) = dy%yr(1:COL,2) + wy%yr(1:COL,4)
  dy%yi(1:COL,2) = dy%yi(1:COL,2) + wy%yi(1:COL,4)
  dy%yr(1:COL,3) = dy%yr(1:COL,3) - wy%yi(1:COL,4)
  dy%yi(1:COL,3) = dy%yi(1:COL,3) + wy%yr(1:COL,4)
  dy%yr(1:COL,4) = dy%yr(1:COL,4) - wy%yi(1:COL,3)
  dy%yi(1:COL,4) = dy%yi(1:COL,4) + wy%yr(1:COL,3)
  return
end subroutine

#ifdef _CLOVER
subroutine s_mult_clover(fcl,dy)
  implicit none
  type(clvmat),  intent(in) :: fcl
  type(su3ferm), intent(inout):: dy
  type(su3ferm) :: yy,fy
  real(SP) :: ar(COL,SPIN),ai(COL,SPIN)
  real(SP) :: br(COL,SPIN),bi(COL,SPIN)
  integer :: ics,ic,is

  br(1:COL,1:2) = dy%yr(1:COL,1:2) - dy%yr(1:COL,3:4)
  bi(1:COL,1:2) = dy%yi(1:COL,1:2) - dy%yi(1:COL,3:4)
  ar(1:COL,1:2) = dy%yr(1:COL,1:2) + dy%yr(1:COL,3:4)
  ai(1:COL,1:2) = dy%yi(1:COL,1:2) + dy%yi(1:COL,3:4)
  call s_mult_6x6(fcl%c1r,fcl%c1i,br,bi)
  call s_mult_6x6(fcl%c2r,fcl%c2i,ar,ai)
  dy%yr(1:COL,1:2) = ar(1:COL,3:4) +  br(1:COL,3:4)
  dy%yi(1:COL,1:2) = ai(1:COL,3:4) +  bi(1:COL,3:4)
  dy%yr(1:COL,3:4) = ar(1:COL,3:4) -  br(1:COL,3:4)
  dy%yi(1:COL,3:4) = ai(1:COL,3:4) -  bi(1:COL,3:4)

  return
end subroutine

subroutine s_mult_6x6(cr,ci,ar,ai)
  implicit none
  real(SP), intent(in)      :: cr(CLSPH),ci(CLSPH)
  real(SP), intent(inout)   :: ar(COL,SPIN),ai(COL,SPIN)

  ar(1,3)=cr( 1)*ar(1,1)    &
 &       +cr( 2)*ar(2,1)    &
 &       +cr( 3)*ar(3,1)    &
 &       +cr( 4)*ar(1,2)    &
 &       +cr( 5)*ar(2,2)    &
 &       +cr( 6)*ar(3,2)    &
!     &       -ci( 1)*ai(1,1)
 &       -ci( 2)*ai(2,1)    &
 &       -ci( 3)*ai(3,1)    &
 &       -ci( 4)*ai(1,2)    &
 &       -ci( 5)*ai(2,2)    &
 &       -ci( 6)*ai(3,2)
  ai(1,3)=cr( 1)*ai(1,1)    &
 &       +cr( 2)*ai(2,1)    &
 &       +cr( 3)*ai(3,1)    &
 &       +cr( 4)*ai(1,2)    &
 &       +cr( 5)*ai(2,2)    &
 &       +cr( 6)*ai(3,2)    &
!     &       +ci( 1)*ar(1,1)
 &       +ci( 2)*ar(2,1)    &
 &       +ci( 3)*ar(3,1)    &
 &       +ci( 4)*ar(1,2)    &
 &       +ci( 5)*ar(2,2)    &
 &       +ci( 6)*ar(3,2)

  ar(2,3)=cr( 2)*ar(1,1)    &
 &       +cr( 7)*ar(2,1)    &
 &       +cr( 8)*ar(3,1)    &
 &       +cr( 9)*ar(1,2)    &
 &       +cr(10)*ar(2,2)    &
 &       +cr(11)*ar(3,2)    &
 &       +ci( 2)*ai(1,1)    &
!     &       -ci( 7)*ai(2,1)
 &       -ci( 8)*ai(3,1)    &
 &       -ci( 9)*ai(1,2)    &
 &       -ci(10)*ai(2,2)    &
 &       -ci(11)*ai(3,2)
  ai(2,3)=cr( 2)*ai(1,1)    &
 &       +cr( 7)*ai(2,1)    &
 &       +cr( 8)*ai(3,1)    &
 &       +cr( 9)*ai(1,2)    &
 &       +cr(10)*ai(2,2)    &
 &       +cr(11)*ai(3,2)    &
 &       -ci( 2)*ar(1,1)    &
!     &       +ci( 7)*ar(2,1)
 &       +ci( 8)*ar(3,1)    &
 &       +ci( 9)*ar(1,2)    &
 &       +ci(10)*ar(2,2)    &
 &       +ci(11)*ar(3,2)

  ar(3,3)=cr( 3)*ar(1,1)    &
 &       +cr( 8)*ar(2,1)    &
 &       +cr(12)*ar(3,1)    &
 &       +cr(13)*ar(1,2)    &
 &       +cr(14)*ar(2,2)    &
 &       +cr(15)*ar(3,2)    &
 &       +ci( 3)*ai(1,1)    &
 &       +ci( 8)*ai(2,1)    &
!     &       -ci(12)*ai(3,1)
 &       -ci(13)*ai(1,2)    &
 &       -ci(14)*ai(2,2)    &
 &       -ci(15)*ai(3,2)
  ai(3,3)=cr( 3)*ai(1,1)    &
 &       +cr( 8)*ai(2,1)    &
 &       +cr(12)*ai(3,1)    &
 &       +cr(13)*ai(1,2)    &
 &       +cr(14)*ai(2,2)    &
 &       +cr(15)*ai(3,2)    &
 &       -ci( 3)*ar(1,1)    &
 &       -ci( 8)*ar(2,1)    &
!     &       +ci(12)*ar(3,1)
 &       +ci(13)*ar(1,2)    &
 &       +ci(14)*ar(2,2)    &
 &       +ci(15)*ar(3,2)

  ar(1,4)=cr( 4)*ar(1,1)    &
 &       +cr( 9)*ar(2,1)    &
 &       +cr(13)*ar(3,1)    &
 &       +cr(16)*ar(1,2)    &
 &       +cr(17)*ar(2,2)    &
 &       +cr(18)*ar(3,2)    &
 &       +ci( 4)*ai(1,1)    &
 &       +ci( 9)*ai(2,1)    &
 &       +ci(13)*ai(3,1)    &
!     &       -ci(16)*ai(1,2)
 &       -ci(17)*ai(2,2)    &
 &       -ci(18)*ai(3,2)
  ai(1,4)=cr( 4)*ai(1,1)    &
 &       +cr( 9)*ai(2,1)    &
 &       +cr(13)*ai(3,1)    &
 &       +cr(16)*ai(1,2)    &
 &       +cr(17)*ai(2,2)    &
 &       +cr(18)*ai(3,2)    &
 &       -ci( 4)*ar(1,1)    &
 &       -ci( 9)*ar(2,1)    &
 &       -ci(13)*ar(3,1)    &
!     &       +ci(16)*ar(1,2)
 &       +ci(17)*ar(2,2)    &
 &       +ci(18)*ar(3,2)

  ar(2,4)=cr( 5)*ar(1,1)    &
 &       +cr(10)*ar(2,1)    &
 &       +cr(14)*ar(3,1)    &
 &       +cr(17)*ar(1,2)    &
 &       +cr(19)*ar(2,2)    &
 &       +cr(20)*ar(3,2)    &
 &       +ci( 5)*ai(1,1)    &
 &       +ci(10)*ai(2,1)    &
 &       +ci(14)*ai(3,1)    &
 &       +ci(17)*ai(1,2)    &
!     &       -ci(19)*ai(2,2)
 &       -ci(20)*ai(3,2)
  ai(2,4)=cr( 5)*ai(1,1)    &
 &       +cr(10)*ai(2,1)    &
 &       +cr(14)*ai(3,1)    &
 &       +cr(17)*ai(1,2)    &
 &       +cr(19)*ai(2,2)    &
 &       +cr(20)*ai(3,2)    &
 &       -ci( 5)*ar(1,1)    &
 &       -ci(10)*ar(2,1)    &
 &       -ci(14)*ar(3,1)    &
 &       -ci(17)*ar(1,2)    &
!     &       +ci(19)*ar(2,2)
 &       +ci(20)*ar(3,2)

  ar(3,4)=cr( 6)*ar(1,1)    &
 &       +cr(11)*ar(2,1)    &
 &       +cr(15)*ar(3,1)    &
 &       +cr(18)*ar(1,2)    &
 &       +cr(20)*ar(2,2)    &
 &       +cr(21)*ar(3,2)    &
 &       +ci( 6)*ai(1,1)    &
 &       +ci(11)*ai(2,1)    &
 &       +ci(15)*ai(3,1)    &
 &       +ci(18)*ai(1,2)    &
 &       +ci(20)*ai(2,2)
!     &       -ci(21)*ai(3,2)
  ai(3,4)=cr( 6)*ai(1,1)    &
 &       +cr(11)*ai(2,1)    &
 &       +cr(15)*ai(3,1)    &
 &       +cr(18)*ai(1,2)    &
 &       +cr(20)*ai(2,2)    &
 &       +cr(21)*ai(3,2)    &
 &       -ci( 6)*ar(1,1)    &
 &       -ci(11)*ar(2,1)    &
 &       -ci(15)*ar(3,1)    &
 &       -ci(18)*ar(1,2)    &
 &       -ci(20)*ar(2,2)
!     &       +ci(21)*ar(3,2)
  return
end subroutine
#endif

#ifdef _CLOVER
subroutine assign_mult_hop_eo_blk_all(Mye,yo,ue,uo,fclinve)
#else
subroutine assign_mult_hop_eo_blk_all(Mye,yo,ue,uo)
#endif
!
! multiply Hopping operator (clover) : FMy = F^-1 M y
!
!        FMye = F^-1ee Meo yo (ieo=0)
!
! acts on all sites in a block (including extended sites).
!
!    FMye : an even/odd site fermion field
!      yo : an odd/even site fermion field
!      ue : even/odd site gauge link
!      uo : odd/even site gauge link
!
  use counter_class
  use field_fermion_class, only : mult_iter
  implicit none
  type(s_wqf_eo_blk_obj),  intent(inout) :: Mye
  type(s_wqf_eo_blk_obj),  intent(in)    :: yo
  type(s_gvf_eo_blk_obj),  intent(in)    :: ue
  type(s_gvf_eo_blk_obj),  intent(in)    :: uo
#ifdef _CLOVER
  type(s_clvf_eo_blk_obj), intent(in)    :: fclinve
#endif
  type(su3ferm) :: yt
  type(su3mat)  :: uu
  type(su3ferm) :: yy
#ifdef _CLOVER
  type(clvmat)  :: cc
#endif
  integer :: iflag
  integer :: ix,iy,iz,ic,is,ieo,igx,ics
  integer :: ibx,iby,ibz,ithz,itht,itht0,ieoxyz
  integer :: izp,izd,it
  integer :: ithtp,ithtd
  integer :: ithzp,ithzd
  integer :: ibxp,ibxd
  integer :: ibyp,ibyd
  integer :: ibzp,ibzd

  ieo=ue%ieo
  igx=ue%igx
  Mye%ieo=ieo
  Mye%igx=igx

  do ibx=1-NDEPTH,NBX+NDEPTH
    ix=ibx+(igx-1)*NBX

  do iby=1-NDEPTH,NBY+NDEPTH
    iy=iby

  do ibz=1,NBZ
    do ithz=1,NTHZ
      iz=ithz+(ibz-1)*NTHZ
      ieoxyz=abs(mod(ipeo+ix+iy+iz+ieo,2))

      do itht=1,NTHT

        yt%yr(:,:)=0.0_SP
        yt%yi(:,:)=0.0_SP

        it = (itht-ieoxyz)*2+ieoxyz
        ithtp=(mod(it-1+1+NT,NT)+1)/2+1-ieoxyz
        ithtd=(mod(it-1-1+NT,NT)+1)/2+1-ieoxyz
        call s_pack_u(itht ,ithz,4,ue%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtp,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_forw_t(uu,yy,yt)
        call s_pack_u(ithtd,ithz,4,uo%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtd,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_back_t(uu,yy,yt)

        izp=mod(iz-1+1+NZ,NZ)+1
        izd=mod(iz-1-1+NZ,NZ)+1
        ithzp=mod(izp-1,NTHZ)+1
        ithzd=mod(izd-1,NTHZ)+1
        ibzp=(izp-1)/NTHZ+1
        ibzd=(izd-1)/NTHZ+1
        call s_pack_u(itht,ithz ,3,ue%blk(ibz ,iby,ibx),uu)
        call s_pack_y(itht,ithzp,  yo%blk(ibzp,iby,ibx),yy)
        call s_mult_forw_z(uu,yy,yt)
        call s_pack_u(itht,ithzd,3,uo%blk(ibzd,iby,ibx),uu)
        call s_pack_y(itht,ithzd,  yo%blk(ibzd,iby,ibx),yy)
        call s_mult_back_z(uu,yy,yt)

        iflag=1-(iby/(NBY+NDEPTH))
        ibyp=iby+iflag
        call s_pack_u(itht,ithz,2,ue%blk(ibz,iby ,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyp,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_y(uu,yy,yt)

        iflag=1-((NBY+1-iby)/(NBY+NDEPTH))
        ibyd=iby-iflag
        call s_pack_u(itht,ithz,2,uo%blk(ibz,ibyd,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyd,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_y(uu,yy,yt)

        iflag=1-(ibx/(NBX+NDEPTH))
        ibxp=ibx+iflag
        call s_pack_u(itht,ithz,1,ue%blk(ibz,iby,ibx ),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxp),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_x(uu,yy,yt)

        iflag=(1-((NBX+1-ibx)/(NBX+NDEPTH)))
        ibxd=ibx-iflag
        call s_pack_u(itht,ithz,1,uo%blk(ibz,iby,ibxd),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxd),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_x(uu,yy,yt)

#ifdef _CLOVER
        call s_pack_clv(itht,ithz,fclinve%blk(ibz,iby,ibx),cc)
        call s_mult_clover(cc,yt)
#endif

        call s_unpack_y(yt,itht,ithz,Mye%blk(ibz,iby,ibx))

      enddo  ! end of do itht
    enddo  ! end of do ithz
  enddo
  enddo
  enddo  ! end of do ibz,iby,ibx

  call inc(mult_iter,1)

  return
end subroutine

#ifdef _CLOVER
subroutine assign_mult_hop_eo_all(Mye,yo,ue,uo,fclinve)
#else
subroutine assign_mult_hop_eo_all(Mye,yo,ue,uo)
#endif
!
! multiply Hopping operator (clover) : FMy = F^-1 M y
!        FMye = F^-1ee Meo yo (ieo=0)
!
! acts on all sites in a block (including extended sites)
!
!    FMye : an even/odd site fermion field
!      yo : an odd/even site fermion field
!      ue : even/odd site gauge link
!      uo : odd/even site gauge link
!
  implicit none
  type(s_wqf_eo_obj),  intent(inout) :: Mye
  type(s_wqf_eo_obj),  intent(in)    :: yo
  type(s_gvf_eo_obj),  intent(in)    :: ue
  type(s_gvf_eo_obj),  intent(in)    :: uo
#ifdef _CLOVER
  type(s_clvf_eo_obj), intent(in)    :: fclinve
#endif
  integer :: igx
!$OMP PARALLEL DO
  do igx=1,NGX
#ifdef _CLOVER
    call assign_mult_hop_eo_blk_all(Mye%g(igx),yo%g(igx),ue%g(igx),uo%g(igx),fclinve%g(igx))
#else
    call assign_mult_hop_eo_blk_all(Mye%g(igx),yo%g(igx),ue%g(igx),uo%g(igx))
#endif
  enddo
  return
end subroutine

#ifdef _CLOVER
subroutine assign_mult_hop_eo_blk_in(Mye,yo,ue,uo,fclinve)
#else
subroutine assign_mult_hop_eo_blk_in(Mye,yo,ue,uo)
#endif
!
! multiply Hopping operator (clover) : FMy = F^-1 M y
!                        FMye = F^-1ee Meo yo (ieo=0)
!
! acts on the interior sites of a physical block.
!
!      yo : an odd/even site fermion feild
!    FMye : an even/odd site fermion feild
!      ue : even/odd site gauge link
!      uo : odd/even site gauge link
!
  use counter_class
  use field_fermion_class, only : mult_iter
  implicit none
  type(s_wqf_eo_blk_obj),  intent(inout) :: Mye
  type(s_wqf_eo_blk_obj),  intent(in)    :: yo
  type(s_gvf_eo_blk_obj),  intent(in)    :: ue
  type(s_gvf_eo_blk_obj),  intent(in)    :: uo
#ifdef _CLOVER
  type(s_clvf_eo_blk_obj), intent(in)    :: fclinve
#endif
  type(su3ferm) :: yt
  type(su3mat)  :: uu
  type(su3ferm) :: yy
#ifdef _CLOVER
  type(clvmat)  :: cc
#endif
  integer :: iflag
  integer :: ix,iy,iz,ic,is,ieo,igx,ics
  integer :: ibx,iby,ibz,ithz,itht,itht0,ieoxyz
  integer :: izp,izd,it
  integer :: ithtp,ithtd
  integer :: ithzp,ithzd
  integer :: ibxp,ibxd
  integer :: ibyp,ibyd
  integer :: ibzp,ibzd

  ieo=ue%ieo
  igx=ue%igx
  Mye%ieo=ieo
  Mye%igx=igx

  do ibx=2,NBX-1
    ix=ibx+(igx-1)*NBX
  do iby=2,NBY-1
  do ibz=1,NBZ
    do ithz=1,NTHZ
      iy=iby
      iz=ithz+(ibz-1)*NTHZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
#ifdef _SF
      do itht=1+ieoxyz,NTHT
#else
      do itht=1,NTHT
#endif

        yt%yr(:,:)=0.0_SP
        yt%yi(:,:)=0.0_SP

        it = (itht-ieoxyz)*2+ieoxyz
        ithtp=(mod(it-1+1+NT,NT)+1)/2+1-ieoxyz
        ithtd=(mod(it-1-1+NT,NT)+1)/2+1-ieoxyz
        call s_pack_u(itht ,ithz,4,ue%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtp,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_forw_t(uu,yy,yt)
        call s_pack_u(ithtd,ithz,4,uo%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtd,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_back_t(uu,yy,yt)

        izp=mod(iz-1+1+NZ,NZ)+1
        izd=mod(iz-1-1+NZ,NZ)+1
        ithzp=mod(izp-1,NTHZ)+1
        ithzd=mod(izd-1,NTHZ)+1
        ibzp=(izp-1)/NTHZ+1
        ibzd=(izd-1)/NTHZ+1
        call s_pack_u(itht,ithz ,3,ue%blk(ibz ,iby,ibx),uu)
        call s_pack_y(itht,ithzp,  yo%blk(ibzp,iby,ibx),yy)
        call s_mult_forw_z(uu,yy,yt)
        call s_pack_u(itht,ithzd,3,uo%blk(ibzd,iby,ibx),uu)
        call s_pack_y(itht,ithzd,  yo%blk(ibzd,iby,ibx),yy)
        call s_mult_back_z(uu,yy,yt)

        iflag=1-(iby/(NBY+1))
        ibyp=iby+iflag
        call s_pack_u(itht,ithz,2,ue%blk(ibz,iby ,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyp,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_y(uu,yy,yt)

        iflag=(1-((NBY+1-iby)/(NBY+1)))
        ibyd=iby-iflag
        call s_pack_u(itht,ithz,2,uo%blk(ibz,ibyd,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyd,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_y(uu,yy,yt)

        iflag=1-(ibx/(NBX+NDEPTH))
        ibxp=ibx+iflag
        call s_pack_u(itht,ithz,1,ue%blk(ibz,iby,ibx ),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxp),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_x(uu,yy,yt)

        iflag=(1-((NBX+1-ibx)/(NBX+NDEPTH)))
        ibxd=ibx-iflag
        call s_pack_u(itht,ithz,1,uo%blk(ibz,iby,ibxd),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxd),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_x(uu,yy,yt)

#ifdef _CLOVER
        call s_pack_clv(itht,ithz,fclinve%blk(ibz,iby,ibx),cc)
        call s_mult_clover(cc,yt)
#endif

        call s_unpack_y(yt,itht,ithz,Mye%blk(ibz,iby,ibx))

      enddo  ! end of do itht
#ifdef _SF
      if (ieoxyz==1) then
        do ics=1,COL*2
          Mye%blk(ibz,iby,ibx)%thd(ics)%f(ieoxyz,ithz)%x = 0.0_SP
          Mye%blk(ibz,iby,ibx)%thd(ics)%f(ieoxyz,ithz)%y = 0.0_SP
          Mye%blk(ibz,iby,ibx)%thd(ics)%f(ieoxyz,ithz)%z = 0.0_SP
          Mye%blk(ibz,iby,ibx)%thd(ics)%f(ieoxyz,ithz)%w = 0.0_SP
        enddo
      endif
#endif
    enddo  ! end of do ithz
  enddo
  enddo
  enddo  ! end of do ibz,iby,ibx

  call inc(mult_iter,1)

  return
end subroutine

#ifdef _CLOVER
subroutine assign_mult_hop_eo_in(Mye,yo,ue,uo,fclinve)
#else
subroutine assign_mult_hop_eo_in(Mye,yo,ue,uo)
#endif
!
! multiply Hopping operator (clover) : FMy = F^-1 M y
!                        FMye = F^-1ee Meo yo (ieo=0)
!
! acts on the interior sites of a physical block.
!
!    FMye : an even/odd site fermion feild
!      yo : an odd/even site fermion feild
!      ue : even/odd site gauge link
!      uo : odd/even site gauge link
!
  implicit none
  type(s_wqf_eo_obj),  intent(inout) :: Mye
  type(s_wqf_eo_obj),  intent(in)    :: yo
  type(s_gvf_eo_obj),  intent(in)    :: ue
  type(s_gvf_eo_obj),  intent(in)    :: uo
#ifdef _CLOVER
  type(s_clvf_eo_obj), intent(in)    :: fclinve
#endif
  integer :: igx
!$OMP PARALLEL DO
  do igx=1,NGX
#ifdef _CLOVER
    call assign_mult_hop_eo_blk_in(Mye%g(igx),yo%g(igx),ue%g(igx),uo%g(igx),fclinve%g(igx))
#else
    call assign_mult_hop_eo_blk_in(Mye%g(igx),yo%g(igx),ue%g(igx),uo%g(igx))
#endif
  enddo
  return
end subroutine

#ifdef _CLOVER
subroutine assign_mult_hop_eo_blk_out(Mye,yo,ue,uo,fclinve)
#else
subroutine assign_mult_hop_eo_blk_out(Mye,yo,ue,uo)
#endif
! 
! multiply Hopping operator (clover) : FMy = F^-1 M y
!                            FMye = F^-1ee Meo yo (ieo=0)
!
! acts on the edge sites of a physical block.
!
!    FMye : an even/odd site fermion field
!      yo : an odd/even site fermion field
!      ue : even/odd site gauge link
!      uo : odd/even site gauge link
!
  implicit none
  type(s_wqf_eo_blk_obj),  intent(inout) :: Mye
  type(s_wqf_eo_blk_obj),  intent(in)    :: yo
  type(s_gvf_eo_blk_obj),  intent(in)    :: ue
  type(s_gvf_eo_blk_obj),  intent(in)    :: uo
#ifdef _CLOVER
  type(s_clvf_eo_blk_obj), intent(in)    :: fclinve
#endif
  type(su3ferm) :: yt
  type(su3mat)  :: uu
  type(su3ferm) :: yy
#ifdef _CLOVER
  type(clvmat)  :: cc
#endif
  integer :: iflag
  integer :: ix,iy,iz,ic,is,ieo,igx,ics
  integer :: ibx,iby,ibz,ithz,itht,itht0,ieoxyz
  integer :: izp,izd,it
  integer :: ithtp,ithtd
  integer :: ithzp,ithzd
  integer :: ibxp,ibxd
  integer :: ibyp,ibyd
  integer :: ibzp,ibzd

  ieo=ue%ieo
  igx=ue%igx
  Mye%ieo=ieo
  Mye%igx=igx

  do ibx=1,NBX
    ix=ibx+(igx-1)*NBX
  do iby=1,NBY,NBY-1
  do ibz=1,NBZ
    do ithz=1,NTHZ
      iy=iby
      iz=ithz+(ibz-1)*NTHZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itht=1,NTHT

        yt%yr(:,:)=0.0_SP
        yt%yi(:,:)=0.0_SP

        it = (itht-ieoxyz)*2+ieoxyz
        ithtp=(mod(it-1+1+NT,NT)+1)/2+1-ieoxyz
        ithtd=(mod(it-1-1+NT,NT)+1)/2+1-ieoxyz
        call s_pack_u(itht ,ithz,4,ue%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtp,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_forw_t(uu,yy,yt)
        call s_pack_u(ithtd,ithz,4,uo%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtd,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_back_t(uu,yy,yt)

        izp=mod(iz-1+1+NZ,NZ)+1
        izd=mod(iz-1-1+NZ,NZ)+1
        ithzp=mod(izp-1,NTHZ)+1
        ithzd=mod(izd-1,NTHZ)+1
        ibzp=(izp-1)/NTHZ+1
        ibzd=(izd-1)/NTHZ+1
        call s_pack_u(itht,ithz ,3,ue%blk(ibz ,iby,ibx),uu)
        call s_pack_y(itht,ithzp,  yo%blk(ibzp,iby,ibx),yy)
        call s_mult_forw_z(uu,yy,yt)
        call s_pack_u(itht,ithzd,3,uo%blk(ibzd,iby,ibx),uu)
        call s_pack_y(itht,ithzd,  yo%blk(ibzd,iby,ibx),yy)
        call s_mult_back_z(uu,yy,yt)

        iflag=1-(iby/(NBY+1))
        ibyp=iby+iflag
        call s_pack_u(itht,ithz,2,ue%blk(ibz,iby ,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyp,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_y(uu,yy,yt)

        iflag=(1-((NBY+1-iby)/(NBY+1)))
        ibyd=iby-iflag
        call s_pack_u(itht,ithz,2,uo%blk(ibz,ibyd,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyd,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_y(uu,yy,yt)

        iflag=1-(ibx/(NBX+NDEPTH))
        ibxp=ibx+iflag
        call s_pack_u(itht,ithz,1,ue%blk(ibz,iby,ibx ),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxp),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_x(uu,yy,yt)

        iflag=(1-((NBX+1-ibx)/(NBX+NDEPTH)))
        ibxd=ibx-iflag
        call s_pack_u(itht,ithz,1,uo%blk(ibz,iby,ibxd),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxd),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_x(uu,yy,yt)

#ifdef _CLOVER
        call s_pack_clv(itht,ithz,fclinve%blk(ibz,iby,ibx),cc)
        call s_mult_clover(cc,yt)
#endif

        call s_unpack_y(yt,itht,ithz,Mye%blk(ibz,iby,ibx))

      enddo  ! end of do itht
    enddo  ! end of do ithz
  enddo
  enddo
  enddo  ! end of do ibz,iby,ibx

  do ibx=1,NBX,NBX-1
    ix=ibx+(igx-1)*NBX
  do iby=2,NBY-1
  do ibz=1,NBZ
    do ithz=1,NTHZ
      iy=iby
      iz=ithz+(ibz-1)*NTHZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do itht=1,NTHT

        yt%yr(:,:)=0.0_SP
        yt%yi(:,:)=0.0_SP

        it = (itht-ieoxyz)*2+ieoxyz
        ithtp=(mod(it-1+1+NT,NT)+1)/2+1-ieoxyz
        ithtd=(mod(it-1-1+NT,NT)+1)/2+1-ieoxyz
        call s_pack_u(itht ,ithz,4,ue%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtp,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_forw_t(uu,yy,yt)
        call s_pack_u(ithtd,ithz,4,uo%blk(ibz,iby,ibx),uu)
        call s_pack_y(ithtd,ithz,  yo%blk(ibz,iby,ibx),yy)
        call s_mult_back_t(uu,yy,yt)

        izp=mod(iz-1+1+NZ,NZ)+1
        izd=mod(iz-1-1+NZ,NZ)+1
        ithzp=mod(izp-1,NTHZ)+1
        ithzd=mod(izd-1,NTHZ)+1
        ibzp=(izp-1)/NTHZ+1
        ibzd=(izd-1)/NTHZ+1
        call s_pack_u(itht,ithz ,3,ue%blk(ibz ,iby,ibx),uu)
        call s_pack_y(itht,ithzp,  yo%blk(ibzp,iby,ibx),yy)
        call s_mult_forw_z(uu,yy,yt)
        call s_pack_u(itht,ithzd,3,uo%blk(ibzd,iby,ibx),uu)
        call s_pack_y(itht,ithzd,  yo%blk(ibzd,iby,ibx),yy)
        call s_mult_back_z(uu,yy,yt)

        iflag=1-(iby/(NBY+1))
        ibyp=iby+iflag
        call s_pack_u(itht,ithz,2,ue%blk(ibz,iby ,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyp,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_y(uu,yy,yt)

        iflag=(1-((NBY+1-iby)/(NBY+1)))
        ibyd=iby-iflag
        call s_pack_u(itht,ithz,2,uo%blk(ibz,ibyd,ibx),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,ibyd,ibx),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_y(uu,yy,yt)

        iflag=1-(ibx/(NBX+NDEPTH))
        ibxp=ibx+iflag
        call s_pack_u(itht,ithz,1,ue%blk(ibz,iby,ibx ),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxp),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_forw_x(uu,yy,yt)

        iflag=(1-((NBX+1-ibx)/(NBX+NDEPTH)))
        ibxd=ibx-iflag
        call s_pack_u(itht,ithz,1,uo%blk(ibz,iby,ibxd),uu)
        call s_pack_y(itht,ithz,  yo%blk(ibz,iby,ibxd),yy)
        yy%yr(:,:) = yy%yr(:,:)*real(iflag)
        yy%yi(:,:) = yy%yi(:,:)*real(iflag)
        call s_mult_back_x(uu,yy,yt)

#ifdef _CLOVER
        call s_pack_clv(itht,ithz,fclinve%blk(ibz,iby,ibx),cc)
        call s_mult_clover(cc,yt)
#endif

        call s_unpack_y(yt,itht,ithz,Mye%blk(ibz,iby,ibx))

      enddo  ! end of do itht
    enddo  ! end of do ithz
  enddo
  enddo
  enddo  ! end of do ibz,iby,ibx

  return
end subroutine

#ifdef _CLOVER
subroutine assign_mult_hop_eo_out(Mye,yo,ue,uo,fclinve)
#else
subroutine assign_mult_hop_eo_out(Mye,yo,ue,uo)
#endif
! 
! multiply Hopping operator (clover) : FMy = F^-1 M y
!                            FMye = F^-1ee Meo yo (ieo=0)
!
! acts on the edge sites of a physical block
!
!    FMye : an even/odd site fermion field
!      yo : an odd/even site fermion field
!      ue : even/odd site gauge link
!      uo : odd/even site gauge link
!
  implicit none
  type(s_wqf_eo_obj),  intent(inout) :: Mye
  type(s_wqf_eo_obj),  intent(in)    :: yo
  type(s_gvf_eo_obj),  intent(in)    :: ue
  type(s_gvf_eo_obj),  intent(in)    :: uo
#ifdef _CLOVER
  type(s_clvf_eo_obj), intent(in)    :: fclinve
#endif
  integer :: igx
!$OMP PARALLEL DO
  do igx=1,NGX
#ifdef _CLOVER
    call assign_mult_hop_eo_blk_out(Mye%g(igx),yo%g(igx),ue%g(igx),uo%g(igx),fclinve%g(igx))
#else
    call assign_mult_hop_eo_blk_out(Mye%g(igx),yo%g(igx),ue%g(igx),uo%g(igx))
#endif
  enddo
  return
end subroutine

#ifdef _CLOVER
subroutine assign_mult_eoprec_wd_eo(kappa,dye,ye,u,fclinv)
#else
subroutine assign_mult_eoprec_wd_eo(kappa,dye,ye,u)
#endif
!
! Single precision
!                                                v
! Multiply preconditioned Dirac operator : dye = Dee ye,
! where  v
!        Dee = 1ee - kappa^2 Moe Moe
!
! acts on physical sites.
!
!   kappa : hopping parameter
!     dye : even/odd site fermion field 
!      ye : even/odd site fermion field 
!       u : gauge link
!  fclinv : inverse clover term
!
  implicit none
  real(SP),            intent(in)    :: kappa
  type(s_wqf_eo_obj),  intent(inout) :: dye
  type(s_wqf_eo_obj),  intent(inout) ::  ye
  type(s_gvf_obj),     intent(in)    :: u
#ifdef _CLOVER
  type(s_clvf_obj),    intent(in)    :: fclinv
#endif
!************
  type(s_wqf_eo_obj), allocatable :: Myo
  real(SP) :: mkappa2
  integer :: ibx,iby,ibz,itht,ithz,ieoxyz,idx
  integer :: ix,iy,iz,ic,is,igx,ics
  integer :: ieo,ioe

  ieo = ye%ieo
  ioe = 1-ieo

!        v
!  dye = D ye ( 1 - kappa**2 Meo Moe ) ye

  allocate(Myo)
  !
  !  Myo = F^-1oo Moe ye
  !
  call copy_boundary(ye)
#ifdef _CLOVER
  call assign_mult_hop_eo_in( Myo,ye,u%eo(ioe),u%eo(ieo),fclinv%eo(ioe))
  call assign_mult_hop_eo_out(Myo,ye,u%eo(ioe),u%eo(ieo),fclinv%eo(ioe))
#else
  call assign_mult_hop_eo_in( Myo,ye,u%eo(ioe),u%eo(ieo))
  call assign_mult_hop_eo_out(Myo,ye,u%eo(ioe),u%eo(ieo))
#endif

  !
  !  dye = F^-1ee Meo Myo = F^-1ee Meo F^-1oo Moe ye
  !
  call copy_boundary(Myo)
#ifdef _CLOVER
  call assign_mult_hop_eo_in( dye,Myo,u%eo(ieo),u%eo(ioe),fclinv%eo(ieo))
  call assign_mult_hop_eo_out(dye,Myo,u%eo(ieo),u%eo(ioe),fclinv%eo(ieo))
#else
  call assign_mult_hop_eo_in( dye,Myo,u%eo(ieo),u%eo(ioe))
  call assign_mult_hop_eo_out(dye,Myo,u%eo(ieo),u%eo(ioe))
#endif
  deallocate(Myo)

  !
  !  dye = dye * (-kappa^2) + ye
  !      = ye - kappa^2 Meo Moe ye
  !
  mkappa2= -kappa**2
  call accum_mult_add(dye,mkappa2,ye)

  return
end subroutine

#ifdef _CLOVER
subroutine assign_inv_mult_eoprec_wd_eo_blk_all(kappa,dye,ye,ue,uo,fclinve,fclinvo)
#else
subroutine assign_inv_mult_eoprec_wd_eo_blk_all(kappa,dye,ye,ue,uo)
#endif
!
! Multiply approximated inverse of e/e preconditioned wilson dirac operator
! in a block (including extended sites).
!
!            v
!  dye_bkl = Dee_blk \ ye_blk 
!
!
  implicit none
  real(SP),                intent(in)    :: kappa
  type(s_wqf_eo_blk_obj),  intent(inout) :: dye
  type(s_wqf_eo_blk_obj),  intent(in)    ::  ye
  type(s_gvf_eo_blk_obj),  intent(in)    :: ue,uo
#ifdef _CLOVER
  type(s_clvf_eo_blk_obj), intent(in)    :: fclinve,fclinvo
#endif
!************
  type(s_wqf_eo_blk_obj), allocatable :: Myo,ve
  real(SP) :: kappa2,omega
  integer :: ibx,iby,ibz,itht,ithz,ieoxyz
  integer :: ix,iy,iz,ic,is,igx,ioe,ieo,ics
  integer :: iiter

  kappa2=kappa**2
  igx=ye%igx
  ieo=ue%ieo
  ioe=1-ieo

  dye%ieo=ieo
  dye%igx=igx

  allocate(Myo,ve)

  Myo%ieo=ioe
  Myo%igx=igx
   ve%ieo=ieo
   ve%igx=igx

  !======================
  ! dye = ye
  !======================
  call assign(dye,ye)

  do iiter=1,NMR
    !===============================
    ! Myo = Moe yde
    !  ve = Meo Myo = Meo Moe yde
    !===============================
#ifdef _CLOVER
    call assign_mult_hop_eo_blk_all(Myo,dye,uo,ue,fclinvo)
    call assign_mult_hop_eo_blk_all( ve,Myo,ue,uo,fclinve)
#else
    call assign_mult_hop_eo_blk_all(Myo,dye,uo,ue)
    call assign_mult_hop_eo_blk_all( ve,Myo,ue,uo)
#endif

    !===============================
    ! yde = ye + kappa2 * ve
    !===============================
    do ibx=1-NDEPTH,NBX+NDEPTH
    do iby=1-NDEPTH,NBY+NDEPTH
    do ibz=1,NBZ
      do ics=1,COL*2
        do ithz=1,NTHZ
        do itht=1,NTHT
            dye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x     &
 &         = ye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x     &
 &         + ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%x * kappa2
            dye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y     &
 &         = ye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y     &
 &         + ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%y * kappa2
            dye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z     &
 &         = ye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z     &
 &         + ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%z * kappa2
            dye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w     &
 &         = ye%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w     &
 &         + ve%blk(ibz,iby,ibx)%thd(ics)%f(itht,ithz)%w * kappa2
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo

  enddo
  deallocate(Myo,ve)
  return
end subroutine

#ifdef _CLOVER
subroutine assign_inv_mult_eoprec_wd_eo_all(kappa,dye,ye,ue,uo,fclinve,fclinvo)
#else
subroutine assign_inv_mult_eoprec_wd_eo_all(kappa,dye,ye,ue,uo)
#endif
!
! Multiply approximated inverse of e/e preconditioned wilson dirac operator
! in a block (including extended sites).
!
!            v
!  dye_bkl = Dee_blk \ ye_blk 
!
!
  implicit none
  real(SP),            intent(in)    :: kappa
  type(s_wqf_eo_obj),  intent(inout) :: dye
  type(s_wqf_eo_obj),  intent(in)    ::  ye
  type(s_gvf_eo_obj),  intent(in)    :: ue,uo
#ifdef _CLOVER
  type(s_clvf_eo_obj), intent(in)    :: fclinve,fclinvo
#endif
  integer :: igx
!$OMP PARALLEL DO
  do igx=1,NGX
#ifdef _CLOVER
    call assign_inv_mult_eoprec_wd_eo_blk_all(kappa,dye%g(igx),ye%g(igx),ue%g(igx),uo%g(igx), &
 &                                                        fclinve%g(igx),fclinvo%g(igx))
#else
    call assign_inv_mult_eoprec_wd_eo_blk_all(kappa,dye%g(igx),ye%g(igx),ue%g(igx),uo%g(igx))
#endif
  enddo
  return
end subroutine

#ifdef _CLOVER
subroutine assign_inv_mult_eoprec_wd_ras(kappa,dye,ye,u,fclinv)
#else
subroutine assign_inv_mult_eoprec_wd_ras(kappa,dye,ye,u)
#endif
!
! RAS preconditioner 
!         v
! dye sim Dee ye
!
  implicit none
  real(SP), intent(in) :: kappa
  type(s_wqf_eo_obj), intent(inout) :: dye
  type(s_wqf_eo_obj), intent(inout) ::  ye
  type(s_gvf_obj),    intent(in)    ::   u
#ifdef _CLOVER
  type(s_clvf_obj),   intent(in)    :: fclinv
#endif
!************
  type(s_wqf_eo_obj), allocatable :: qe,re,ve
  real(SP) :: kappa2,omega
  integer :: ibx,iby,ibz,itht,ithz,ieoxyz
  integer :: ix,iy,iz,ic,is,igx,ioe,ieo,ics
  integer :: iiter

  ieo = ye%ieo
  ioe = 1-ieo

  call new(dye,ieo)

  allocate(qe,re,ve)

  call new(qe,ieo)
  call new(re,ieo)
  call new(ve,ieo)

  call copy_boundary(ye)

  call assign(dye,ye)  ! dye = ye
  call assign( ve,ye)  !  ve = ye
  call assign( re,ye)  !  re = ye

  do iiter=1,NRAS

    !------------------------
    ! qe = Dee ve
    !------------------------
#ifdef _CLOVER
    call assign_mult_eoprec_wd_eo(kappa,qe,ve,u,fclinv)
#else
    call assign_mult_eoprec_wd_eo(kappa,qe,ve,u)
#endif

    call accum_sub(re,qe)    ! r = r - q

    !-------------------------------------
    ! ve = Dee' \ re (block inverse)
    !-------------------------------------
    call copy_boundary(re)
#ifdef _CLOVER
    call assign_inv_mult_eoprec_wd_eo_all(kappa,ve,re,u%eo(ieo),u%eo(ioe), &
 &                                               fclinv%eo(ieo),fclinv%eo(ioe))
#else
    call assign_inv_mult_eoprec_wd_eo_all(kappa,ve,re,u%eo(ieo),u%eo(ioe))
#endif

    !
    ! yd = yd + v
    !
    call accum_add(dye,ve)

  enddo

  deallocate(qe,re,ve)

  return
end subroutine

subroutine assign_inv_mult_eoprec_wd_sgl(this,kappa,dye,ye,ieo)
!
! $Id: single_solver_class.F90,v 1.3 2011/08/05 09:10:30 ishikawa Exp $
! BiCGStab
!
  implicit none
  type(single_solver), intent(inout) :: this
  real(DP),            intent(in)    :: kappa
  complex(DP),         intent(inout) ::  ye(:)
  complex(DP),         intent(inout) :: dye(:)
  integer,             intent(in)    :: ieo
!**********
  real(DP) :: norm,tol
  type(s_wqf_eo_obj), allocatable :: rte,pe,te,qe,re,ve,xe,be
  integer :: iter,CHKFLG
  complex(SP) :: rho1,rho0,bet,omg,alp
  complex(SP) :: ctmp0
  real(SP) :: rnorm,rtmp0,rsource,precnorm,skappa,stol,rtmp1(NGX)
  character(LEN=23) :: FMT='("@",I5," ERR:",ES15.7)'
  real(SP), parameter :: TOL_BK = TINY(1.0_SP)*100
  type(timer) :: etime
  real(DP) :: time_solver
  integer :: igx,iout,ioe


  call new(etime)
  call tic(etime)

  ioe = 1-ieo

  if (ABS(this%err_norm) > EPSILON(1.0_DP)) then
    tol  = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)
  else
    tol  = MAX_TOL
  endif
  iout = this%iout

  skappa = kappa
  stol   = tol

#ifdef _DEBUG
  if (0 == nodeid) then
    write(*,'("%ASSIGN_INV_MULT_EOPREC_WD_SGL")')
    write(*,'("% TOL=",E24.15," ERR_NORM=",E24.15)')this%tol,this%err_norm
    write(*,'("%STOL=",E24.15)')stol
  endif
  CHKFLG=1
#else
  CHKFLG=0
#endif

  allocate(rte,re,pe,te,qe,ve,xe,be)
  call new(rte,ieo)
  call new( re,ieo)
  call new( pe,ieo)
  call new( te,ieo)
  call new( qe,ieo)
  call new( ve,ieo)
  call new( xe,ieo)
  call new( be,ieo)

  call unpack(ye,be,norm)   ! be <= ye/|ye|

!***********************************
! precondition (ipre=1)
!***********************************


!********
!  x = 0 
!  r = b 
!  rt = r 
!  p = r 
!  rho0 = <rt|r>
!********
  call clear(xe)
  call assign(re ,be)
  call assign(rte,re)
  call assign( pe,re)
  rho0 = prod(rte,re)

  iter = this%iter

  if (nodeid == 0 .AND. CHKFLG == 1) then
    write(*,'("# NRAS=",I3," NMR=",I3," NDEPTH=",I3," NGX=",I3)')NRAS,NMR,NDEPTH,NGX
    rnorm = sqrt(real(rho0))
    write(*,FMT)iter,rnorm
  endif

  do while (iter < MAX_ITER)

    !-------------
    !  v = M p
    !  q = A v
    !-------------
#ifdef _PREC_
#ifdef _CLOVER
    call assign_inv_mult_eoprec_wd_ras(skappa,ve,pe,this%su,this%sfcl)
    call      assign_mult_eoprec_wd_eo(skappa,qe,ve,this%su,this%sfcl)
#else
    call assign_inv_mult_eoprec_wd_ras(skappa,ve,pe,this%su)
    call      assign_mult_eoprec_wd_eo(skappa,qe,ve,this%su)
#endif
#else
#ifdef _CLOVER
    call assign_mult_eoprec_wd_eo(skappa,qe,pe,this%su,this%sfcl)
#else
    call assign_mult_eoprec_wd_eo(skappa,qe,pe,this%su)
#endif
#endif
    iter = iter + 1

    ctmp0 = prod(rte,qe)
    if (ABS(ctmp0) < TOL_BK) then
      exit
    endif
    alp = rho0/ctmp0

    !--------------------
    ! x = x + alp v 
    ! r = r - alp q
    ! rnorm = |r|
    !--------------------
#ifdef _PREC_
    call update_xr_eo(alp,xe,re,ve,qe,rtmp0)
#else
    call update_xr_eo(alp,xe,re,pe,qe,rtmp0)
#endif
    rnorm = sqrt(rtmp0)

    if (nodeid == 0 .AND. CHKFLG == 1) then
      write(*,FMT)iter,rnorm
    endif
    if ( rnorm <= stol ) then
      exit
    endif

    !----------------
    !  v = M r
    !  t = A v
    !----------------
#ifdef _PREC_
#ifdef _CLOVER
    call assign_inv_mult_eoprec_wd_ras(skappa,ve,re,this%su,this%sfcl)
    call      assign_mult_eoprec_wd_eo(skappa,te,ve,this%su,this%sfcl)
#else
    call assign_inv_mult_eoprec_wd_ras(skappa,ve,re,this%su)
    call      assign_mult_eoprec_wd_eo(skappa,te,ve,this%su)
#endif
#else
#ifdef _CLOVER
    call assign_mult_eoprec_wd_eo(skappa,te,re,this%su,this%sfcl)
#else
    call assign_mult_eoprec_wd_eo(skappa,te,re,this%su)
#endif
#endif
    iter = iter + 1

    rtmp0 = abs2(te)
    if (ABS(rtmp0) < TOL_BK) then
      exit
    endif
    ctmp0 = prod(te,re)
    omg = ctmp0/rtmp0

    !------------------------
    !  x = x + omg v
    !  r = r - omg t
    !  rnorm = |r|
    !------------------------
    rnorm = 0.0_SP
#ifdef _PREC_
    call update_xr_eo( omg,xe,re,ve,te,rtmp0)
#else
    call update_xr2_eo(omg,xe,re,te,rtmp0)
#endif
    rnorm = sqrt(rtmp0)

    if (nodeid == 0 .AND. CHKFLG == 1) then
      write(*,FMT)iter,rnorm
    endif
    if ( rnorm <= stol ) then
      exit
    endif

!***********************
! rho1 = <rt|r>
!***********************
    rho1 = prod(rte,re)
    if ((ABS(rho0) < TOL_BK) .or. (ABS(omg) < TOL_BK)) then
      exit
    endif
    bet = (rho1/rho0)*(alp/omg)
    rho0 = rho1

!*************************
! p = r + beta(p - omg q)
! p = p - omg*q
! p = p*bet + q
!*************************
    call accum_add_mult(pe,-omg,qe)
    call accum_mult_add(pe, bet,qe)
    
  enddo

#ifdef _DEBUG
#ifdef _CLOVER
  call assign_mult_eoprec_wd_eo(skappa,te,xe,this%su,this%sfcl)
#else
  call assign_mult_eoprec_wd_eo(skappa,te,xe,this%su)
#endif
  call assign(re,be)    ! r = b
  call accum_sub(re,te) ! r = r - t = b - A x
  rnorm = abs2(re)
  rnorm = SQRT(rnorm)
  if (nodeid == 0) then
    write(iout,FMT)iter,rnorm
  endif
  write(*,FMT)iter,rnorm
#endif

  call pack(xe,dye,norm)

  deallocate(rte,re,pe,te,qe,ve,xe,be)

  if ( rnorm <= stol ) then
    if (nodeid == 0) then
      write(iout,'("@BiCGStab_SP(NRAS=",I3," NMR=",I3," NDEPTH=",I3," NGX=",I3,") iteration converged.")') &
 &                  NRAS,NMR,NDEPTH,NGX
      write(iout,FMT)iter,rnorm
    endif
  endif
  if ( iter >= MAX_ITER ) then
    if (nodeid == 0) then
      write(iout,'("@BiCGStab_SP(NRAS=",I3," NMR=",I3," NDEPTH=",I3," NGX=",I3,") iteration does not converge.")') &
 &                  NRAS,NMR,NDEPTH,NGX
      write(iout,FMT)iter,rnorm
    endif
    call error_stop("@BiCGStab_SP does not converge. stop.")
  endif
  this%iter = iter

  call toc(etime)
  time_solver = get_elapse(etime)
  if (nodeid == 0) then
    write(iout,'("@Solve_SP ETIME:",F15.7)')time_solver
  endif
  if (nodeid == 0 .AND. CHKFLG == 1) then
    write(*,'("%Solve_SP ETIME:",F15.7)')time_solver
  endif

  return
end subroutine

!#include "s_copy_y_hf.F90"

end module
