#include "config.h"

module gpu_solver_class
  use iso_c_binding
  use comlib
  use lattice_class
  use counter_class
  use field_gauge_class
  use field_fermion_class
  use field_clover_class
  use single_field_class
  implicit none
  private

  public :: gpu_solver
  public :: new, delete
  public :: assign_inv_mult_eoprec_wd
  public :: bind_cpu_to_gpu
  public :: unbind_cpu_to_gpu

  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP

  logical, save :: m_is_gpu_binded_to_cpu = .false.
  logical, save :: m_is_gpu_initialized   = .false.

  integer(C_INT), save, bind(C,name="cu_mult_count") :: gpu_mult_count = 0

  type gpu_solver
    private
    type(s_wqf_eo_obj),  allocatable :: swe
    type(s_wqf_eo_obj),  allocatable :: sve
    type(s_gvf_obj),     allocatable :: su
#ifdef _CLOVER
    type(s_clvf_obj),    allocatable :: sfcl
#endif
    real(DP), public :: err_norm
    real(DP), public :: tol
    integer,  public :: iter
#ifdef _DEBUG
    integer,  public :: print_flag = 1
#else
    integer,  public :: print_flag = 0
#endif
  end type

  interface new
    module procedure new_gpu_solver_f
  end interface

  interface delete
    module procedure delete_gpu_solver_f
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_gpu
  end interface

  interface 
   !
   ! Fortran interfaces for GPU code using F2003 C binding
   !

    subroutine bind_cpu_to_gpu_loc() BIND(C,name="bind_cpu_to_gpu_")
    end subroutine

    subroutine unbind_cpu_to_gpu_loc() BIND(C,name="unbind_cpu_to_gpu_")
    end subroutine

#ifdef _CLOVER
    subroutine new_gpu_solver_clover(u, fclinv) BIND(C,name="new_gpu_solver_clover_")
      import s_gvf_obj
      import s_clvf_obj
      implicit none
      type(s_gvf_obj),  intent(in) :: u
      type(s_clvf_obj), intent(in) :: fclinv
    end subroutine
#endif

    subroutine new_gpu_solver_wilson(u) BIND(C,name="new_gpu_solver_wilson_")
      import s_gvf_obj
      implicit none
      type(s_gvf_obj),  intent(in) :: u
    end subroutine

    subroutine delete_gpu_solver(print_flag) BIND(C,name="delete_gpu_solver_")
      import C_INT
      implicit none
      integer(C_INT), intent(in) :: print_flag
    end subroutine

    subroutine s_bicgstab_gpu(tol, iter, kappa, ye, dye) BIND(C,name="s_bicgstab_hmc_gpu_")
      import c_double, c_int, s_wqf_eo_obj
      implicit none
      real(C_DOUBLE),     intent(in)    :: tol, kappa
      integer(C_INT),     intent(in)    :: iter
      type(s_wqf_eo_obj), intent(inout) :: ye, dye
    end subroutine

  end interface
  
contains

subroutine bind_cpu_to_gpu()
  implicit none
  if (.not.m_is_gpu_binded_to_cpu) then
    call bind_cpu_to_gpu_loc()
    m_is_gpu_binded_to_cpu = .true.
  endif
  return
end subroutine

subroutine unbind_cpu_to_gpu()
  implicit none
  if (m_is_gpu_binded_to_cpu) then
    call unbind_cpu_to_gpu_loc()
    m_is_gpu_binded_to_cpu = .false.
  endif
  return
end subroutine

#ifdef _CLOVER
subroutine new_gpu_solver_f(this,u,fclinv)
#else
subroutine new_gpu_solver_f(this,u)
#endif
  implicit none
  type(gpu_solver),               intent(inout) :: this
  type(vfield_gluon_wg),             intent(in) :: u
#ifdef _CLOVER
  type(field_clover_term), optional, intent(in) :: fclinv
#endif
  logical :: is_clover
  type(single_field_world) :: sp_lattice

#ifdef _CLOVER
  if (present(fclinv)) then
    is_clover = .true.
  else
    is_clover = .false.
  endif
#else
  is_clover = .false.
#endif

  call new(sp_lattice)
  call bind_cpu_to_gpu()

  this%iter = 0
  gpu_mult_count = 0

  if (allocated(this%swe)) deallocate(this%swe)
  if (allocated(this%sve)) deallocate(this%sve)
  if (allocated(this%su))  deallocate(this%su)
  allocate(this%swe)
  allocate(this%sve)
  allocate(this%su)
  call new(this%su)

#ifdef _CLOVER
  if (is_clover) then
    if (allocated(this%sfcl)) deallocate(this%sfcl)
    allocate(this%sfcl)
    call new(this%sfcl)
  endif
#endif

  call conv_u(this%su,u)
#ifdef _CLOVER
  if (is_clover) then
    call conv_clv(this%sfcl,fclinv)
  endif
#endif

  if (is_clover) then
#ifdef _CLOVER
    call new_gpu_solver_clover(this%su,this%sfcl)
#endif
  else
    call new_gpu_solver_wilson(this%su)
  endif

  if (allocated(this%su)) deallocate(this%su)
#ifdef _CLOVER
  if (is_clover) then
    if (allocated(this%sfcl)) deallocate(this%sfcl)
  endif
#endif
  m_is_gpu_initialized = .true.
  return
end subroutine

subroutine delete_gpu_solver_f(this)
  implicit none
  type(gpu_solver),      intent(inout) :: this

  this%iter = 0
  gpu_mult_count = 0

  if (allocated(this%swe))  deallocate(this%swe)
  if (allocated(this%sve))  deallocate(this%sve)
  if (allocated(this%su))   deallocate(this%su)
#ifdef _CLOVER
  if (allocated(this%sfcl)) deallocate(this%sfcl)
#endif

  if (m_is_gpu_initialized) then
    call delete_gpu_solver(this%print_flag)
  endif

  m_is_gpu_initialized = .false.

  return
end subroutine

subroutine assign_inv_mult_eoprec_wd_gpu(this,kappa,dye,ye,ieo)
!
!       v
! dye = Dee \ ye
!
  use error_class
  implicit none
  type(gpu_solver), intent(inout) :: this
  real(DP),         intent(in)    :: kappa
  complex(DP),      intent(inout) ::  ye(:)  ! packed data
  complex(DP),      intent(inout) :: dye(:)  ! packed data
  integer,          intent(in)    :: ieo
  character(len=CHARLEN) :: str
  real(DP) :: norm, tol

  gpu_mult_count = 0
  tol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)

  call new(this%swe,ieo)
  !
  ! swe <= ye/|ye|, norm = |ye|
  !
  call unpack(ye,this%swe,norm)
!write(*,*)abs2(this%swe)

  !        v
  ! sve <= Dee \ swe
  !
  call s_bicgstab_gpu(tol, this%iter, kappa, this%swe, this%sve)
!  this%sve = this%swe

  !
  ! dye <= sve*norm
  !
  call pack(this%sve,dye,norm)    

  call inc(mult_iter,gpu_mult_count)
  gpu_mult_count = 0

  !==========================
  ! check solver convergence
  !==========================
  if (this%iter < 0) then
    call delete(this)
    write(str,'("assign_inv_mult_eoprec_wd_gpu : ",A," at ",A)')__FILE__,__LINE__
    call error_message(TRIM(str))
    call error_stop("assign_inv_mult_eoprec_wd_gpu : Error, maximam itertion reached.")
  endif

  return
end subroutine

end module
