#include <omp.h>
#ifdef _QMX_
#include <qmx_lib.h>
#else
#include <mpi.h>
#endif

#include <cassert>
#include <mm_malloc.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <cutil.h>

#include "cconfig.h"
#include "cgpu_config.h"
#include "cuparam-size.h"
#include "cuMyComplex.h"
#include "flop_table.h"
#include "omp_thread_gpu_dev_table.h"

//
// define _MULTIGPU_ to use mutli gpus
//
#undef  _MULTIGPU_
#if _NGX > 1
  #define _MULTIGPU_
#else
  #undef  _MULTIGPU_
#endif

//
// define _PREC_ to use preconditioner
//
#undef _PREC_
#if _NRAS > 0
  #define _PREC_
#else
  #undef  _PREC_
#endif

//
// Fortran Module Variables
//
#include "cufortran_vars.h"

//
// Texture fetching for the link variables and
// the input fermion fields.
//
#ifdef _USE_TEXFETCH_
static texture<float4,1,cudaReadModeElementType> tex_ueri;
static texture<float4,1,cudaReadModeElementType> tex_uori;
static texture<float4,1,cudaReadModeElementType> tex_yori;
#endif
static texture<float2,1,cudaReadModeElementType> tex_Complex;
static texture<float ,1,cudaReadModeElementType> tex_Float;

//
// list vectors
//
// HOST side
static int4  phys_blk[_PHYS_BLOCK_];
static int4    in_blk[_IN_BLOCK_];
static int4   out_blk[_OUT_BLOCK_];
// GPU side
static int4_gpu phys_blk_d;
static int4_gpu   in_blk_d;
static int4_gpu  out_blk_d;

//
// MPI_Allreduce for reduction operations
//
#ifndef _singlePU
#ifdef _QMX_
static void sumcast(float     *f){
//  float ff = *f;
  qmx_sumcast_float(1,f);
}
static void sumcast(cuComplex *c){  
  float f[2];
  f[0] = (*c).x;
  f[1] = (*c).y;
  qmx_sumcast_float(2,f);
  (*c).x = f[0];
  (*c).y = f[1];
}
#else
static void sumcast(float     *f){
  float f2;
//  float ff = *f;
  MPI_Allreduce(f,&f2,1,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
  *f=f2;
//  printf("%%SUMCAST_FLOAT[%d]%14.6e =>%14.6e\n",_nodeid_,ff,*f);
}

static void sumcast(cuComplex *c){
  cuComplex c2;
//  cuComplex cc = *c;
  MPI_Allreduce(c,&c2,2,MPI_FLOAT,MPI_SUM,MPI_COMM_WORLD);
  *c=c2;
//  printf("%%SUMCAST_CMPLX[%d]%14.6e%14.6e =>%14.6e%14.6e\n",_nodeid_,cc.x,cc.y,(*c).x,(*c).y);
}
#endif
#endif

//
// Timer names
//
static unsigned int ycopy_timer;
static unsigned int gsend_timer;
static unsigned int ysend_timer;
static unsigned int yrecv_timer;
static unsigned int  solv_timer;
static unsigned int  mult_timer;
static unsigned int  prec_timer;
static unsigned int  binv_timer;

//
// flop/copy counters
//
static double ycopy_count;  // counts MPI_sendrecv calls
static double ycopy_dsize;  // counts MPI_sendrecv total data size
static double solv_flop_count;  // counts total flop in the solver
static double mult_flop_count;  // counts total flop for mult operation

//
// kernel switch
//
static bool is_clover = false;

#include "YMACRO.h"
#include "cuMatVec.h"
#include "cuWqfAlgebra.h"
#include "cu_mult_forw_t.h"
#include "cu_mult_forw_z.h"
#include "cu_mult_forw_y.h"
#include "cu_mult_forw_x.h"
#include "cu_mult_back_t.h"
#include "cu_mult_back_z.h"
#include "cu_mult_back_y.h"
#include "cu_mult_back_x.h"
#include "cu_mult_clover_term.h"
#include "copy_boundary.h"
#include "assign_mult_hop_kernel.h"
#include "assign_mult_hop_in_kernel.h"
#include "assign_mult_eoprec_wd_eo.h"
#include "assign_inv_mult_eoprec_wd_ras.h"

//
// global device memory for LINK/CLOVER fields.
//
static cu_gvf    *ud      = NULL;
static cu_clvf   *fclinvd = NULL;

extern "C" void 
bind_cpu_to_gpu_(){
//
// Bind GPU to OpenMP-threads
//

  int num_gpus;

  CUDA_SAFE_CALL(cudaGetDeviceCount(&num_gpus));
  if (num_gpus < 1) {
    if (_nodeid_==0) 
      printf("There is no CUDA Device. Program Stops. MPI_RANK %5d \n",_nodeid_);
    exit(-1);
  }
  if ( 0 == _nodeid_ ) {
    printf("%%==============================================================================\n");
    printf("%% GPUs are binded to CPU threads.\n");
#ifdef _OPENMP
    printf("%% Number of Host OpenMP Threads : %d\n",omp_get_max_threads());
#else
    printf("%%                 Single Thread : %d\n",1);
#endif
    printf("%% Number of CUDA Devices        : %d\n",num_gpus);
  }
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
#ifdef _OPENMP
    int omp_thread_num  = omp_get_thread_num();
#else
    int omp_thread_num  = 0;
#endif
//    int cuda_device_num = igx % num_gpus;
    int cuda_device_num = devidx[igx % num_gpus];
    if ( 0 == _nodeid_ )
      printf("%% OpenMP Thread = %d  Cuda Device = %d\n",omp_thread_num,cuda_device_num);
#ifdef _MULTIGPU_
    CUDA_SAFE_CALL(cudaSetDevice(cuda_device_num));
#endif
    
    CUDA_SAFE_CALL(cudaThreadSetCacheConfig(cudaFuncCachePreferL1));

  }  
  if ( 0 == _nodeid_ )
    printf("%%==============================================================================\n");

  //
  // create list vector for physical sites 
  //
  int jblk=0;
  for(int ibx=0;ibx<_NBX;ibx++) {
  for(int iby=0;iby<_NBY;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    phys_blk[jblk].x = ibx;
    phys_blk[jblk].y = iby;
    phys_blk[jblk].z = ibz;
    phys_blk[jblk].w = iblk;
    jblk++;
  }
  }
  }

  //
  // create list vector for physical bulk(internal) sites
  //
  jblk=0;
  for(int ibx=1;ibx<_NBX-1;ibx++) {
  for(int iby=1;iby<_NBY-1;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    in_blk[jblk].x = ibx;
    in_blk[jblk].y = iby;
    in_blk[jblk].z = ibz;
    in_blk[jblk].w = iblk;
    jblk++;
  }
  }
  }

  //
  // create list vector for physical surface(edge) sites
  //
  jblk=0;
  int ibx=0;
  for(int iby=0;iby<_NBY;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    out_blk[jblk].x = ibx;
    out_blk[jblk].y = iby;
    out_blk[jblk].z = ibz;
    out_blk[jblk].w = iblk;
    jblk++;
  }
  }
  for(    ibx=1;ibx<_NBX-1;ibx++) {
  for(int iby=0;iby<_NBY;iby=iby+_NBY-1) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    out_blk[jblk].x = ibx;
    out_blk[jblk].y = iby;
    out_blk[jblk].z = ibz;
    out_blk[jblk].w = iblk;
    jblk++;
  }
  }
  }
  ibx=_NBX-1;
  for(int iby=0;iby<_NBY;iby++) {
  for(int ibz=0;ibz<_NBZ;ibz++) {
    int iblk=ibz+(iby+_NDEPTH)*_NBZ+(ibx+_NDEPTH)*_NBZ*(_NBY+2*_NDEPTH);
    out_blk[jblk].x = ibx;
    out_blk[jblk].y = iby;
    out_blk[jblk].z = ibz;
    out_blk[jblk].w = iblk;
    jblk++;
  }
  }
#ifdef _CHECK_LIST_VECTOR_
  if ( 0 == _nodeid_ ) {
    for (int iblk=0;iblk<_PHYS_BLOCK_;iblk++) {
      printf("PHYS:%6d%6d%3d%3d%3d\n",iblk,phys_blk[iblk].w,phys_blk[iblk].z,phys_blk[iblk].y,phys_blk[iblk].x);
    }
    for (int iblk=0;iblk<_IN_BLOCK_;iblk++) {
      printf("IN:%6d%6d%3d%3d%3d\n",iblk,in_blk[iblk].w,in_blk[iblk].z,in_blk[iblk].y,in_blk[iblk].x);
    }
    for (int iblk=0;iblk<_OUT_BLOCK_;iblk++) {
      printf("OUT:%6d%6d%3d%3d%3d\n",iblk,out_blk[iblk].w,out_blk[iblk].z,out_blk[iblk].y,out_blk[iblk].x);
    }
  }
#endif
  fflush(stdout);
}

extern "C" void 
unbind_cpu_to_gpu_(){
//
// Unbind GPUs and OpenMP threads
//
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
#ifdef _MULTIGPU_
    CUDA_SAFE_CALL(cudaThreadExit());
#endif
  }
}

static void 
new_gpu_solver_cuda(f_gvf  *u,
                    f_clvf *fclinv = NULL) {

#ifdef _DEBUG_
  printf("%%      nodeid=%d\n",_nodeid_);
  printf("%%        ipeo=%d\n",_ipeo_);
  for (int mu=0;mu < NDIM-1; mu++) printf("%%nodeidup[%d]=%d\n",mu,_nodeidup_(mu));
  for (int mu=0;mu < NDIM-1; mu++) printf("%%nodeiddn[%d]=%d\n",mu,_nodeiddn_(mu));
  fflush(stdout);
#endif

#ifdef _DEBUG_
  if ( 0 == _nodeid_ ) {
    printf("%%SIZEOF(CHAR)    : %ld\n",sizeof(char));
    printf("%%SIZEOF(SHORT)   : %ld\n",sizeof(short));
    printf("%%SIZEOF(INT)     : %ld\n",sizeof(int));
    printf("%%SIZEOF(FLOAT)   : %ld\n",sizeof(float));
    printf("%%SIZEOF(LINKeo)  : %ld\n",sizeof(cu_gvf_eo_blk));
    printf("%%SIZEOF(CLOVEReo): %ld\n",sizeof(cu_clvf_eo_blk));
    fflush(stdout);
  }
#endif

  if ( 0 == _nodeid_ ) fflush(stdout);

//
// Initialize GPU solver and
//   send LINK/CLOVER data from HOST to GPU
//

  //
  // Initialize timers and counters.
  //
  cutCreateTimer(&ycopy_timer);
  cutCreateTimer(&gsend_timer);
  cutCreateTimer(&ysend_timer);
  cutCreateTimer(&yrecv_timer);
  cutCreateTimer( &solv_timer);
  cutCreateTimer( &mult_timer);
  cutCreateTimer( &prec_timer);
  cutCreateTimer( &binv_timer);
  cutResetTimer(ycopy_timer);
  cutResetTimer(gsend_timer);
  cutResetTimer(ysend_timer);
  cutResetTimer(yrecv_timer);
  cutResetTimer( solv_timer);
  cutResetTimer( mult_timer);
  cutResetTimer( prec_timer);
  cutResetTimer( binv_timer);

  ycopy_count     = 0.0;
  ycopy_dsize     = 0.0;
  solv_flop_count = 0.0;
  mult_flop_count = 0.0;

  //
  // create ghost site copy buffers
  //
  new_copy_boundary();

  cutStartTimer(gsend_timer);
  //
  // create device memory and send LINK/CLOVER data to GPU
  //
  ud = (cu_gvf *)_mm_malloc(sizeof(cu_gvf),16);
  (*ud).eo[0].ieo = 0;
  (*ud).eo[1].ieo = 1;
  if (is_clover) {
    fclinvd = (cu_clvf *)_mm_malloc(sizeof(cu_clvf),16);
    (*fclinvd).eo[0].ieo = 0;
    (*fclinvd).eo[1].ieo = 1;
  }

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    CUDA_SAFE_CALL(cudaMalloc((void**)&(ipeo_d.g[igx]),sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&( igx_d.g[igx]),sizeof(int)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&( EVN_D.g[igx]),sizeof(int)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&( ODD_D.g[igx]),sizeof(int)));

    CUDA_SAFE_CALL(cudaMemcpy(ipeo_d.g[igx],&_ipeo_,sizeof(int),cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy( igx_d.g[igx],   &igx,sizeof(int),cudaMemcpyHostToDevice));
    int EVN = 0;
    int ODD = 1;
    CUDA_SAFE_CALL(cudaMemcpy( EVN_D.g[igx],&EVN,sizeof(int),cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy( ODD_D.g[igx],&ODD,sizeof(int),cudaMemcpyHostToDevice));

    CUDA_SAFE_CALL(cudaMalloc((void**)&(phys_blk_d.g[igx]),sizeof(int4)*_PHYS_BLOCK_));
    CUDA_SAFE_CALL(cudaMalloc((void**)&(  in_blk_d.g[igx]),sizeof(int4)*_IN_BLOCK_));
    CUDA_SAFE_CALL(cudaMalloc((void**)&( out_blk_d.g[igx]),sizeof(int4)*_OUT_BLOCK_));

    CUDA_SAFE_CALL(cudaMemcpy(phys_blk_d.g[igx],phys_blk,sizeof(int4)*_PHYS_BLOCK_,cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy(  in_blk_d.g[igx],  in_blk,sizeof(int4)*_IN_BLOCK_  ,cudaMemcpyHostToDevice));
    CUDA_SAFE_CALL(cudaMemcpy( out_blk_d.g[igx], out_blk,sizeof(int4)*_OUT_BLOCK_ ,cudaMemcpyHostToDevice));

    CUDA_SAFE_CALL(cudaMalloc((void**)&((*ud).eo[0].g[igx]),sizeof(cu_gvf_eo_blk)));
    CUT_CHECK_ERROR("cudaMalloc ud.eo[0] faild");
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*ud).eo[1].g[igx]),sizeof(cu_gvf_eo_blk)));
    CUT_CHECK_ERROR("cudaMalloc ud.eo[1] faild");

    CUDA_SAFE_CALL(cudaMemcpy((*ud).eo[0].g[igx],&((*u).eo[0].g[igx]),sizeof(cu_gvf_eo_blk),cudaMemcpyHostToDevice));
    CUT_CHECK_ERROR("cudaMemcpy ud.eo[0] faild");
    CUDA_SAFE_CALL(cudaMemcpy((*ud).eo[1].g[igx],&((*u).eo[1].g[igx]),sizeof(cu_gvf_eo_blk),cudaMemcpyHostToDevice));
    CUT_CHECK_ERROR("cudaMemcpy ud.eo[1] faild");

    new_u((*ud).eo[0].g[igx],ipeo_d.g[igx],EVN_D.g[igx],igx_d.g[igx]);
    new_u((*ud).eo[1].g[igx],ipeo_d.g[igx],ODD_D.g[igx],igx_d.g[igx]);


    if (is_clover) {
      CUDA_SAFE_CALL(cudaMalloc((void**)&((*fclinvd).eo[0].g[igx]),sizeof(cu_clvf_eo_blk)));
      CUT_CHECK_ERROR("cudaMalloc fclinvd.eo[0] faild");

      CUDA_SAFE_CALL(cudaMalloc((void**)&((*fclinvd).eo[1].g[igx]),sizeof(cu_clvf_eo_blk)));
      CUT_CHECK_ERROR("cudaMalloc fclinvd.eo[1] faild");

      CUDA_SAFE_CALL(cudaMemcpy((*fclinvd).eo[0].g[igx],&((*fclinv).eo[0].g[igx]),sizeof(cu_clvf_eo_blk),cudaMemcpyHostToDevice));
      CUT_CHECK_ERROR("cudaMemcpy fclinvd.eo[0] faild");

      CUDA_SAFE_CALL(cudaMemcpy((*fclinvd).eo[1].g[igx],&((*fclinv).eo[1].g[igx]),sizeof(cu_clvf_eo_blk),cudaMemcpyHostToDevice));
      CUT_CHECK_ERROR("cudaMemcpy fclinvd.eo[1] faild");

      new_clv((*fclinvd).eo[0].g[igx],ipeo_d.g[igx],EVN_D.g[igx],igx_d.g[igx]);
      new_clv((*fclinvd).eo[1].g[igx],ipeo_d.g[igx],ODD_D.g[igx],igx_d.g[igx]);
    }

    CUDA_SAFE_CALL(cudaThreadSynchronize());
  }

  cutStopTimer(gsend_timer);
}

extern "C" void
new_gpu_solver_wilson_(f_gvf *u){
  is_clover = false;
  new_gpu_solver_cuda(u);
#ifdef _DEBUG_
  printf("%%NEW_GPU_SOLVER_WILSON\n");
  fflush(stdout);
#endif
}

extern "C" void
new_gpu_solver_clover_(f_gvf *u, f_clvf *fclinv){
  is_clover = true;
  new_gpu_solver_cuda(u,fclinv);
#ifdef _DEBUG_
  printf("%%NEW_GPU_SOLVER_CLOVER\n");
  fflush(stdout);
#endif
}

extern "C" void
delete_gpu_solver_(const int *print_flag) {
//
// Delete GPU solver and summarize GPU statistics
//
// *pring_flag == 1 : print runtime statistics
// *pring_flag != 1 : no output
//

  //
  // delete ghost site copy buffers
  //
  delete_copy_boundary();

  //
  // delete device memorys
  //
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {

    CUDA_SAFE_CALL(cudaFree(ipeo_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree( igx_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree( EVN_D.g[igx]));
    CUDA_SAFE_CALL(cudaFree( ODD_D.g[igx]));

    CUDA_SAFE_CALL(cudaFree(phys_blk_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(  in_blk_d.g[igx]));
    CUDA_SAFE_CALL(cudaFree( out_blk_d.g[igx]));

    CUDA_SAFE_CALL(cudaFree((*ud).eo[0].g[igx]));        CUT_CHECK_ERROR("cudaFree ud.eo[0] faild");
    CUDA_SAFE_CALL(cudaFree((*ud).eo[1].g[igx]));        CUT_CHECK_ERROR("cudaFree ud.eo[1] faild");

    if (is_clover) {
      CUDA_SAFE_CALL(cudaFree((*fclinvd).eo[0].g[igx])); CUT_CHECK_ERROR("cudaFree fclinvd.eo[0] faild");
      CUDA_SAFE_CALL(cudaFree((*fclinvd).eo[1].g[igx])); CUT_CHECK_ERROR("cudaFree fclinvd.eo[1] faild");
    }

    CUDA_SAFE_CALL(cudaThreadSynchronize());
  }

  _mm_free(ud);
  if (is_clover) {
    _mm_free(fclinvd);
  }

  //
  // summarize GPU statistics
  //
  double etime_gsend = ((double)cutGetTimerValue(gsend_timer))*1.0e-3;
  double etime_ysend = ((double)cutGetTimerValue(ysend_timer))*1.0e-3;
  double etime_yrecv = ((double)cutGetTimerValue(yrecv_timer))*1.0e-3;
  double etime_ycopy = ((double)cutGetTimerValue(ycopy_timer))*1.0e-3;
  double etime_solv  = ((double)cutGetTimerValue( solv_timer))*1.0e-3;
  double etime_mult  = ((double)cutGetTimerValue( mult_timer))*1.0e-3;
  double etime_prec  = ((double)cutGetTimerValue( prec_timer))*1.0e-3;
  double etime_binv  = ((double)cutGetTimerValue( binv_timer))*1.0e-3;

  cutDeleteTimer(gsend_timer);
  cutDeleteTimer(ysend_timer);
  cutDeleteTimer(yrecv_timer);
  cutDeleteTimer(ycopy_timer);
  cutDeleteTimer( solv_timer);
  cutDeleteTimer( mult_timer);
  cutDeleteTimer( prec_timer);
  cutDeleteTimer( binv_timer);


  double HNSITE = (double)(NT*NZ*NY*NX/2);
  double solv_tflop_count = HNSITE*solv_flop_count;
  double mult_tflop_count = HNSITE*mult_flop_count;
  double solv_gflops = 0.0;
  double mult_gflops = 0.0;
  if (etime_solv > 0.0) {
    solv_gflops = ((HNSITE*solv_flop_count)/etime_solv)*1.0e-9;
  }
  if (etime_mult > 0.0) {
    mult_gflops = ((HNSITE*mult_flop_count)/etime_mult)*1.0e-9;
  }

  if ( 0 == _nodeid_ && 1 == *print_flag ) {
    printf("%%==============================================================================\n");
    printf("%% Solve_GFLOPS_SP     : %15.7F\n",solv_gflops);
    printf("%% Solve_FLOP_SP       : %15lld\n",(long long int)solv_tflop_count);
    printf("%% Solve_SP  ETIME(sec): %15.7F\n",etime_solv);

    printf("%% Mult_GFLOPS_SP      : %15.7F\n",mult_gflops);
    printf("%% Mult_FLOP_SP        : %15lld\n",(long long int)mult_tflop_count);
    printf("%% Mult_SP   ETIME(sec): %15.7F\n",etime_mult);

    printf("%% Prec_SP   ETIME(sec): %15.7F\n",etime_prec);
    printf("%% Binv_SP   ETIME(sec): %15.7F\n",etime_binv);
  }

  double gsend_dsize;
  if (is_clover) {
    gsend_dsize =  (((double)sizeof(cu_gvf_eo_blk)*2
                   +(double)sizeof(cu_clvf_eo_blk)*2)*(_NGX))/1024.0/1024.0;
  } else {
    gsend_dsize =  (((double)sizeof(cu_gvf_eo_blk)*2)*(_NGX))/1024.0/1024.0;
  }
  double ysend_dsize = ((double)sizeof(cu_wqf_eo_blk)*(_NGX))/1024.0/1024.0;

  ycopy_dsize = ycopy_dsize/1024.0/1024.0;

  if ( 0 == _nodeid_  && 1 == *print_flag ) {
    if (etime_gsend > 0.0) {
      printf("%% GPU_GSend ETIME(sec): %15.7F\n",etime_gsend);
      printf("%% GPU_GSend  SIZE(MiB): %15.7F\n",gsend_dsize);
      printf("%% GPU_GSend  (MiB/sec): %15.7F\n",gsend_dsize/etime_gsend);
    }
    if (etime_ysend > 0.0) {
      printf("%% GPU_YSend ETIME(sec): %15.7F\n",etime_ysend);
      printf("%% GPU_YSend  SIZE(MiB): %15.7F\n",ysend_dsize);
      printf("%% GPU_YSend  (MiB/sec): %15.7F\n",ysend_dsize/etime_ysend);
    }
    if (etime_yrecv > 0.0) {
      printf("%% GPU_YRecv ETIME(sec): %15.7F\n",etime_yrecv);
      printf("%% GPU_YRecv  SIZE(MiB): %15.7F\n",ysend_dsize);
      printf("%% GPU_YRecv  (MiB/sec): %15.7F\n",ysend_dsize/etime_yrecv);
    }
    if (etime_ycopy > 0.0) {
      printf("%% GPU_YCopy ETIME(sec): %15.7F\n",etime_ycopy);
      printf("%% GPU_YCopy    Called : %15.7F\n",ycopy_count);
      printf("%% GPU_YCopy  SIZE(MiB): %15.7F\n",ycopy_dsize);
      printf("%% GPU_YCopy  (MiB/sec): %15.7F\n",ycopy_dsize/etime_ycopy);
    }
    printf("%%==============================================================================\n");
    fflush(stdout);
  }

  is_clover = false;
}

extern "C" void
s_bicgstab_hmc_gpu_(double *tol,int *iter, double *kappa, f_wqf_eo *ye, f_wqf_eo *yde) {
//
// MAIN GPU Solver
//

//===========================================================
// host variables
//===========================================================
  const int niter = 5000;
  int ieo;
  float mkappa2 = -(float)((*kappa)*(*kappa));
  float    tol2 =  (float)((*tol)*(*tol));
  float err;
  float rtmp0;
  cuComplex  alp, bet,  omg, rho0, rho1, ctmp0, ctmp1;
  cuComplex nalp,      nomg;

//===========================================================
// device variables
//===========================================================
  int_gpu        ieo_d;
  float_gpu      mkappa2d;
  cuComplex_gpu  alpd, betd, omgd;
  cuComplex_gpu nalpd,      nomgd;

  cu_wqf_eo *yed,*yded,*red,*rted,*ped,*qed,*ted,*ved;

  CUDA_SAFE_CALL(cudaThreadSynchronize());
  cutStartTimer(solv_timer);

#ifdef _DEBUG_
  fflush(stdout);
  if (is_clover) {
    printf("%%S_BICGSTAB_CLOVER\n");
  } else {
    printf("%%S_BICGSTAB_WILSON\n");
  }
  fflush(stdout);
#endif

   ieo = (*ye).ieo;
  (*yde).ieo  = ieo;
  (*yde).ipeo = _ipeo_;

   yed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  yded = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   red = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
  rted = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ped = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   qed = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ted = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);
   ved = (cu_wqf_eo *)_mm_malloc(sizeof(cu_wqf_eo),16);

   (*yed).ieo = ieo;
  (*yded).ieo = ieo;
   (*red).ieo = ieo;
  (*rted).ieo = ieo;
   (*ped).ieo = ieo;
   (*qed).ieo = ieo;
   (*ted).ieo = ieo;
   (*ved).ieo = ieo;

#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
//======================================
// allocate device memory
//======================================

    CUDA_SAFE_CALL(cudaMalloc((void**)   &(ieo_d.g[igx]),sizeof(int)));
    CUDA_SAFE_CALL(cudaMemcpy(ieo_d.g[igx],&ieo,sizeof(int),cudaMemcpyHostToDevice));

    CUDA_SAFE_CALL(cudaMalloc((void**)    &(alpd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)   &(nalpd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)    &(betd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)    &(omgd.g[igx]),sizeof(cuComplex)));
    CUDA_SAFE_CALL(cudaMalloc((void**)   &(nomgd.g[igx]),sizeof(cuComplex)));

    CUDA_SAFE_CALL(cudaMalloc((void**)&(mkappa2d.g[igx]),sizeof(float)));

    CUDA_SAFE_CALL(cudaMalloc((void**) &((*yed).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*yded).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*red).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**)&((*rted).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*ped).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*qed).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*ted).g[igx]) ,sizeof(cu_wqf_eo_blk)));
    CUDA_SAFE_CALL(cudaMalloc((void**) &((*ved).g[igx]) ,sizeof(cu_wqf_eo_blk)));

    new_q((*yded).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
    new_q( (*yed).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
    new_q( (*red).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
    new_q((*rted).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
    new_q( (*ped).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
    new_q( (*qed).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
    new_q( (*ted).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
    new_q( (*ved).g[igx],ipeo_d.g[igx],ieo_d.g[igx],igx_d.g[igx]);
  }

#ifdef _DEBUG_
  printf("%% cudaMalloc OK.\n");
#endif

//======================================
// Send data from host to device
//======================================
  cutStartTimer(ysend_timer);
  assign(mkappa2d,mkappa2);
  assign(yed,ye);
  cutStopTimer(ysend_timer);

///////////////////////////////////////////////////////////////////////////////
  // set x_0 = 0 (yde=0)
  clear(yded,ieo_d);

  // r  = b ( = y : rhs of equation )
  // rt = r
  // p  = r
  assign( red,yed);
  assign(rted,red);
  assign( ped,red);

  // rho0 = |re|^2
   rtmp0 = abs2(red);
  rho0.x = rtmp0;
  rho0.y = 0.0f;
  solv_flop_count += QABS2;

#ifdef _DEBUG_
  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%% NRAS=%3d NMR=%3d NDEPTH=%3d NGX=%3d\n",_NRAS,_NMR,_NDEPTH,_NGX);
  if ( 0 == _nodeid_ ) printf("%% S_TOL=%24.15e\n",sqrt(tol2));
  if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

  while (*iter < niter) {
    // v = M p 
    // q = A v 
    if (is_clover) {

#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,ped,ud,fclinvd);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ved,ud,fclinvd);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ped,ud,fclinvd);
#endif

    } else {

#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,ped,ud);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ved,ud);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,qed,ped,ud);
#endif

    }

    solv_flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    // alp = rho0/<rt|q>
    ctmp0 = prod(rted,qed);
      alp = assign_div(rho0,ctmp0);
    assign( alpd, alp);
     nalp = assign_neg(alp);
    assign(nalpd,nalp);

    solv_flop_count += QPROD;

//printf("  <rt|q> = %24.15e%24.15e\n",ctmp0.x,ctmp0.y);
//printf("    rho0 = %24.15e%24.15e\n",rho0.x,rho0.y);
//printf("     alp = %24.15e%24.15e\n",alp.x,alp.y);

    // x = x + alp v
    // r = r - alp q
#ifdef _PREC_
    accum_add_mult(yded, alpd,ved);
#else
    accum_add_mult(yded, alpd,ped);
#endif
    accum_add_mult( red,nalpd,qed);

    solv_flop_count += QMULT1C_ACCUM_ADD*2;

    rtmp0 = abs2(red);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    solv_flop_count += QABS2;
    if (rtmp0 < tol2) break;

    // v = M r
    // t = A v
    if (is_clover) {
#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,red,ud,fclinvd);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,ted,ved,ud,fclinvd);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,ted,red,ud,fclinvd);
#endif
    } else {
#ifdef _PREC_
      assign_inv_mult_eoprec_wd_ras(mkappa2d,ved,red,ud);
      solv_flop_count += MULT_PREC;
           assign_mult_eoprec_wd_eo(mkappa2d,ted,ved,ud);
#else
           assign_mult_eoprec_wd_eo(mkappa2d,ted,red,ud);
#endif
    }
    solv_flop_count += MULT_MB_PRE;
    *iter = *iter + 1;

    // omg = <t|r>/<t|t>
    rtmp0 = abs2(ted);
    ctmp0 = prod(ted,red);
      omg = assign_div(ctmp0,rtmp0);
    assign( omgd, omg);
     nomg = assign_neg(omg);
    assign(nomgd,nomg);
    solv_flop_count += QABS2 + QPROD;

//printf("   <t|t> = %24.15e\n",rtmp0);
//printf("   <t|r> = %24.15e%24.15e\n",ctmp0.x,ctmp0.y);
//printf("     omg = %24.15e%24.15e\n",omg.x,omg.y);

    // x = x + omg v
    // r = r - omg t
#ifdef _PREC_
    accum_add_mult(yded, omgd,ved);
#else
    accum_add_mult(yded, omgd,red);
#endif
    accum_add_mult( red,nomgd,ted);
    solv_flop_count += QMULT1C_ACCUM_ADD*2;

    rtmp0 = abs2(red);
#ifdef _DEBUG_
    err = sqrt(rtmp0);
    if ( 0 == _nodeid_ ) printf("@%5d ERR:%15.7E\n",*iter,err);
    fflush(stdout);
#endif
    solv_flop_count += QABS2;
    if (rtmp0 < tol2) break;

    // rho1 = <rt|r>
    rho1 = prod(rted,red);
    solv_flop_count += QPROD;

//printf("   rho1 = %24.15e%24.15e\n",rho1.x,rho1.y);

    // bet = (alp/omg)*(rho1/rho0)
    // rho0 = rho1;
    ctmp0 = assign_div(alp,omg);
    ctmp1 = assign_div(rho1,rho0);
      bet = assign_mul(ctmp0,ctmp1);
    assign(betd,bet);
     rho0 = rho1;

//printf("   alp/omg = %24.15e%24.15e\n",ctmp0.x,ctmp0.y);
//printf(" rho1/rho0 = %24.15e%24.15e\n",ctmp1.x,ctmp1.y);
//printf("       bet = %24.15e%24.15e\n",  bet.x,  bet.y);

    // p = r + bet ( p - omg q)
    accum_add_mult(ped,nomgd,qed);
    accum_mult_add(ped, betd,red);
    solv_flop_count += QMULT1C_ACCUM_ADD+QMULT2C_ACCUM_ADD;

  } // end while

//======================================
// check iteration counts (max iter)
// if iter exceeds max iteration
// flip the sign.
//======================================
  if (*iter >= niter) {
    *iter = -(*iter);
  }

//======================================
// receive result from device to host
//======================================
  cutStartTimer(yrecv_timer);
  assign(yde,yded);
  cutStopTimer(yrecv_timer);

#ifdef _DEBUG_
// Check residual
//
// ved = Dee yded
//
  if (is_clover) {
    assign_mult_eoprec_wd_eo(mkappa2d,ved,yded,ud,fclinvd);
  } else {
    assign_mult_eoprec_wd_eo(mkappa2d,ved,yded,ud);
  }
//
// red = yed - ved
//
  assign_sub(red,yed,ved);

  rtmp0 = abs2(red);

  err = sqrt(rtmp0);
  if ( 0 == _nodeid_ ) printf("%%%5d ERR:%15.7E\n",*iter,err);
  fflush(stdout);
#endif

//======================================
// deallocate device memory
//======================================
#pragma omp parallel for
  for (int igx=0;igx<_NGX;igx++) {
    CUDA_SAFE_CALL(cudaFree(   ieo_d.g[igx]));

    CUDA_SAFE_CALL(cudaFree(    alpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(   nalpd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(    betd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(    omgd.g[igx]));
    CUDA_SAFE_CALL(cudaFree(   nomgd.g[igx]));

    CUDA_SAFE_CALL(cudaFree(mkappa2d.g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*yed).g[igx]));
    CUDA_SAFE_CALL(cudaFree( (*yded).g[igx]));

    CUDA_SAFE_CALL(cudaFree(  (*red).g[igx]));
    CUDA_SAFE_CALL(cudaFree( (*rted).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*ped).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*qed).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*ted).g[igx]));
    CUDA_SAFE_CALL(cudaFree(  (*ved).g[igx]));
  }

  _mm_free( yed);
  _mm_free(yded);
  _mm_free( red);
  _mm_free(rted);
  _mm_free( ped);
  _mm_free( qed);
  _mm_free( ted);
  _mm_free( ved);

#ifdef _DEBUG_
  fflush(stdout);
#endif
  cutStopTimer(solv_timer);
}
