#include <config.h>

module quark_prop_class
  use comlib
  use lattice_class
  use timer_class
  use logfile_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use wavefunc_class
  implicit none
  private
  public :: quark_prop
  public :: qprop_params
  public :: new
  public :: delete
  public :: read
  public :: print
  public :: print_statistics
  public :: get_prop
  public :: clear_prop
  public :: assign
  public :: assign_mult
  public :: set_source
!  public :: conv_prop_to_y
!  public :: conv_y_to_prop

  integer, parameter :: NSPACE=NX*NY*NZ

  type qprop_params
  !
  ! quark propagator parameter for physics measurement
  !
#ifdef _CLOVER
    type(quark_clover)              :: mass
#else
    type(quark_wilson)              :: mass
#endif
    type(timer)                     :: etime
    type(logfile), pointer          :: solver_log => NULL()
    character(len=CHARLEN)          :: fname
    character(len=CHARLEN)          :: path
    integer                         :: smear_sous ! 0 for local, 1 for smeared
    integer                         :: smear_sink ! 0 for local, 1 for smeared
    real(DP)                        :: tol
  end type

  type quark_prop
  !
  ! quark propagator for physics measurement
  !
    complex(DP),        allocatable :: prop(:,:,:,:,:,:) ! prop(sink,sous,sink,sous,sink,sink)
    real(DP)                        :: psum(NTT)         ! check sum
    type(qprop_params),     pointer :: parameter     => NULL()
    type(src_wavefunc_obj), pointer :: wavefunc_sous => NULL()
    type(snk_wavefunc_obj), pointer :: wavefunc_sink => NULL()
  end type

  interface new
    module procedure new_qprop
    module procedure new_qprop_params
  end interface

  interface delete
    module procedure delete_qprop
    module procedure delete_qprop_params
  end interface

  interface read
    module procedure read_qprop
    module procedure read_qprop_params
  end interface

  interface print
    module procedure print_qprop
    module procedure print_qprop_params
  end interface

  interface assign
    module procedure assign_qprop
    module procedure assign_qprop_params
  end interface

  interface get_prop
    module procedure get_qprop
  end interface

  interface clear_prop
    module procedure clear_qprop
  end interface

  interface print_statistics
    module procedure print_stat_qprop
  end interface

  interface assign_mult
    module procedure assign_mult_u_q
    module procedure assign_mult_u_y
  end interface

contains

subroutine new_qprop(this)
!
! set default values for a quark propagator
!
  implicit none
  type(quark_prop), intent(inout) :: this
  if (allocated(this%prop)) deallocate(this%prop)
  this%parameter     => NULL()
  this%wavefunc_sous => NULL()
  this%wavefunc_sink => NULL()
  return
end subroutine

subroutine delete_qprop(this)
!
! set default values for a quark propagator
!
  implicit none
  type(quark_prop), intent(inout) :: this
  if (allocated(this%prop)) deallocate(this%prop)
  this%parameter     => NULL()
  this%wavefunc_sous => NULL()
  this%wavefunc_sink => NULL()
  return
end subroutine

subroutine new_qprop_params(this)
!
! set default values for a quark propagator parameter
!
  implicit none
  type(qprop_params), intent(inout) :: this

  this%solver_log  => NULL()
  this%tol = 10*EPSILON(1.0_DP)
  call new(this%etime)

  return
end subroutine

subroutine delete_qprop_params(this)
  implicit none
  type(qprop_params), intent(inout) :: this
  this%solver_log => NULL()
  call delete(this%mass)
  return
end subroutine

subroutine assign_qprop_params(param1,param2)
!
! param1 := param2
!
  type(qprop_params), intent(inout) :: param1
  type(qprop_params), intent(in)    :: param2
  call assign(param1%mass,param2%mass)
  param1%solver_log  => param2%solver_log
  param1%fname       =  param2%fname
  param1%path        =  param2%path
  param1%smear_sous  =  param2%smear_sous
  param1%smear_sink  =  param2%smear_sink
  return
end subroutine

subroutine read_qprop(this,iout)
!
! read parameters for a quark propagator
!
  use error_class
  implicit none
  type(quark_prop), intent(inout) :: this
  integer,     intent(in)    :: iout
  call read(this%parameter,iout)
  return
end subroutine

subroutine read_qprop_params(this,iout)
!
! read parameters for a quark propagator parameter
!
  implicit none
  type(qprop_params), intent(inout) :: this
  integer,            intent(in)    :: iout
  real(DP) :: kappa,csw
  integer, parameter :: Nfone=1

  !
  ! read wilson quark mass parameters
  !

  if (nodeid==0) then
    !
    ! read Solver parameters
    !
#ifdef _CLOVER
    read(iout,*)kappa,csw
#else
    read(iout,*)kappa
#endif
    read(iout,*)this%tol
  endif

  if (NPU > 1) then
    call comlib_bcast(kappa,0)
#ifdef _CLOVER
    call comlib_bcast(csw,0)
#endif
    call comlib_bcast(this%tol,0)
  endif

  call set_nflavor(this%mass,Nfone)
  call set_kappa(this%mass,kappa) 
#ifdef _CLOVER
  call   set_csw(this%mass,csw)
#endif

  return
end subroutine

subroutine print_qprop(this)
!
! print parameters for a quark propagator
!
  implicit none
  type(quark_prop), intent(inout) :: this

  call print(this%parameter)

  if (nodeid==0) then
    write(*,'("==== Source Wavefunction type ====")')
  endif
  call print(this%wavefunc_sous)

  if (0 < this%parameter%smear_sink) then
    if (nodeid==0) then
      write(*,'("==== Sink Wavefunction type ====")')
    endif
    call print(this%wavefunc_sink)
  endif

  if (nodeid==0) then
    write(*,'(80("="))')
  endif

  return
end subroutine

subroutine print_qprop_params(this)
!
! print parameters for a quark propagator
!
  implicit none
  type(qprop_params), intent(inout) :: this

  if (nodeid==0) then
    write(*,'("==== Mass Params ====")')
    call print(this%mass)
  endif

  if (nodeid==0) then
    write(*,'("==== Solver Params ====")')
    write(*,'(3X," Solver tol      :",E24.16)')this%tol


    write(*,'("==== Smearing type ====")')
    write(*,'(3X,"          Source :",I3)')this%smear_sous
    write(*,'(3X,"            Sink :",I3)')this%smear_sink
  endif

  return
end subroutine

subroutine psum(this)
  implicit none
  type(quark_prop), intent(inout) :: this
  real(DP) :: rtmp
  integer :: it,itt,ic,is,jc,js,isp

  this%psum(:) = 0.0_DP

  do it=1,NT
!    itt = it + ipsite(4)*NT
    itt = it
    rtmp= 0.0_DP
!$OMP PARALLEL DO PRIVATE(isp,ic,is,jc,js) REDUCTION(+:rtmp)
    do isp=1,NSPACE
      do js=1,SPIN
      do is=1,SPIN
      do jc=1,COL
      do ic=1,COL
        rtmp=rtmp+ real(this%prop(ic,jc,is,js,isp,it))**2 &
 &               +aimag(this%prop(ic,jc,is,js,isp,it))**2
      enddo
      enddo
      enddo
      enddo
    enddo
    this%psum(itt)=rtmp
  enddo

  if (NPU > 1) then
    call comlib_sumcast(this%psum)
  endif

  return
end subroutine

subroutine print_stat_qprop(this)
  implicit none
  type(quark_prop), intent(inout) :: this
  integer :: itt
  real(DP) :: etime

  call psum(this)

  etime = get_elapse(this%parameter%etime)
  if (nodeid==0) then
    write(*,'("==== Solver Statistics ====")')
    write(*,'(3X,"             Time :",F12.4)')etime
  endif

  if (nodeid==0) then
    write(*,'("==== Quark Prop |S(itt)|^2 ====")')
    do itt=1,NTT
      write(*,'(I3,E24.16)')itt,this%psum(itt)
    enddo
    write(*,'("")')
  endif

  return
end subroutine

subroutine clear_qprop(this)
  implicit none
  type(quark_prop), intent(inout) :: this
!$OMP PARALLEL WORKSHARE
  this%prop(:,:,:,:,:,:) = Z0
!$OMP END PARALLEL WORKSHARE
  return
end subroutine

subroutine assign_qprop(qp1,qp2)
!
! q1 := q2
!
  implicit none
  type(quark_prop), intent(inout) :: qp1
  type(quark_prop), intent(in)    :: qp2
  if (allocated(qp1%prop)) deallocate(qp1%prop)
  allocate(qp1%prop(COL,COL,SPIN,SPIN,NX*NY*NZ,NTT))
  qp1%prop = qp2%prop
  qp1%psum = qp2%psum
  qp1%parameter     => qp2%parameter
  qp1%wavefunc_sous => qp2%wavefunc_sous
  qp1%wavefunc_sink => qp2%wavefunc_sink
  return
end subroutine

subroutine set_source(this)
!
! set source via wavefunction
!
  use error_class
  use print_status_class
  implicit none
  type(quark_prop), intent(inout) :: this
  integer :: it,ic,is,isp,jc,js
  integer :: ix,iy,iz,itx,ity,itz,itt
  character(len=CHARLEN) :: str
  complex(DP) :: ctmp

  if (allocated(this%prop)) deallocate(this%prop)
  allocate(this%prop(COL,COL,SPIN,SPIN,NX*NY*NZ,NTT))

  if (.not.associated(this%wavefunc_sous)) then
    call error_stop("set_source: Error, assign wavefunc_sous for qprop.")
  endif

!$OMP PARALLEL DO PRIVATE(it,isp,ic,is,jc,js)
  do it=1,NT
  do isp=1,NSPACE
    do js=1,SPIN
    do is=1,SPIN
    do jc=1,COL
    do ic=1,COL
      this%prop(ic,jc,is,js,isp,it) = Z0
    enddo
    enddo
    enddo
    enddo
    do is=1,SPIN
    do ic=1,COL
      this%prop(ic,ic,is,is,isp,it) = this%wavefunc_sous%wavefunc(ic,is,isp,it)
    enddo
    enddo
  enddo
  enddo

#ifdef _DDD_
!==============
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do it=1,NT
    itx = ix + ipsite(1)*NX
    ity = iy + ipsite(2)*NY
    itz = iz + ipsite(3)*NZ
    itt = it
    isp = ispace(ix,iy,iz)
    do js=1,SPIN
    do jc=1,COL
    do is=1,SPIN
    do ic=1,COL
      ctmp = this%prop(ic,jc,is,js,isp,it)
      call comlib_sumcast(ctmp)
      if (ABS(ctmp) > 0.0_DP) then
        write(str,'(4I4,2I2,2I2,2E24.15," SETSRC")')& 
 &       itx,ity,itz,itt,ic,is,jc,js, &
 &       this%prop(ic,jc,is,js,isp,it)
         call print_status(str)
      endif
    enddo
    enddo
    enddo
    enddo
  enddo
  enddo
  enddo
  enddo
!==============
#endif

  return
end subroutine

subroutine conv_prop_to_y(this,y,ic0,is0)
  implicit none
  type(quark_prop),     intent(in)    :: this
  type(field_quark_wg), intent(inout) :: y
  integer, intent(in) :: ic0,is0
  integer :: ieoxyz,ith,ic,is,ieo,isp
  integer :: ix,iy,iz,it

  do ieo=0,1
    y%eo(ieo)%ieo = ieo
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,it,isp,is,ic)
    do ix=1,NX  ! site loop in 1 block.
    do iy=1,NY
    do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      it=ith*2+ieoxyz
      isp=ispace(ix,iy,iz)

      do is=1,SPIN
      do ic=1,COL
        y%eo(ieo)%s(ith,iz,iy,ix)%y(ic,is) = this%prop(ic,ic0,is,is0,isp,it)
      enddo
      enddo

    enddo
    enddo
    enddo
    enddo
  enddo

  call copy_boundary(y)

  return
end subroutine

subroutine conv_y_to_prop(y,this,ic0,is0)
  implicit none
  type(quark_prop),     intent(inout) :: this
  type(field_quark_wg), intent(in)    :: y
  integer, intent(in) :: ic0,is0
  integer :: ieoxyz,ith,ic,is,ieo,isp
  integer :: ix,iy,iz,it

  do ieo =0,1
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,it,isp,is,ic)
    do ix=1,NX  ! site loop in 1 block.
    do iy=1,NY
    do iz=1,NZ
    ieoxyz=mod(ipeo+iz+iy+ix+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      it=ith*2+ieoxyz
      isp=ispace(ix,iy,iz)

      do is=1,SPIN
      do ic=1,COL
        this%prop(ic,ic0,is,is0,isp,it)=y%eo(ieo)%s(ith,iz,iy,ix)%y(ic,is)
      enddo
      enddo

    enddo
    enddo
    enddo
    enddo
  enddo

  return
end subroutine

subroutine get_qprop(this,u)
  use error_class
  implicit none
  type(quark_prop),           intent(inout) :: this
  type(vfield_gluon_wg), intent(in)    :: u
  type(timer) :: solver_time
  type(field_quark_wg), allocatable :: y,Dy
  character(len=CHARLEN) :: str
  integer :: maxiter,mode,iter
  integer :: ic,is,iout
  real(DP) :: tol,etime,rtmp0,rtmp1
  real(DP) :: copy_y_time0,copy_y_time1,copy_y_etime
#ifdef _CLOVER
  type(tfield_gluon_wg), allocatable :: wl34,wl98
  real(DP) :: dummy
#endif

  call new(this%parameter%etime)
  call tic(this%parameter%etime)


#ifdef _CLOVER
  allocate(wl34,wl98)
  call make_two_links(u,wl34,wl98)
  call make_clover_leaf(u,wl34,wl98)
  deallocate(wl34,wl98)
  call make_inverse_clover_term(this%parameter%mass,0,dummy)
  call delete_clover_leaf()
#endif

  if (.not.allocated(this%prop)) then
    call error_stop("get_qprop: Error, source is not setted/allocated.")
  endif
!  if (allocated(this%prop)) deallocate(this%prop)
!  allocate(this%prop(COL,COL,SPIN,SPIN,NX*NY*NZ,NTT))
!  call clear_qprop(this)
!  call set_source(this)

  allocate(y,Dy)

  do is=1,SPIN
  do ic=1,COL

    call new(solver_time)
    call tic(solver_time)

    copy_y_time0 = get_elapse(copy_fq_time)

    write(str,'("# SOLVER FOR COL=",I2," SPIN=",I2)')ic,is
    call print(this%parameter%solver_log,TRIM(str))
    iout = get_file_unit(this%parameter%solver_log)

    call new(y)
    call new(Dy)

    !------------------------
    ! Dy = D \ y
    !------------------------
    tol    = this%parameter%tol
    call conv_prop_to_y(this, y,ic,is)
    rtmp0 = abs2(y)
    call assign_inv_mult_wd(this%parameter%mass,iout,tol,iter,Dy,y,u)
    rtmp1 = abs2(Dy)
    call conv_y_to_prop(Dy,this,ic,is)

    copy_y_time1 = get_elapse(copy_fq_time)
    copy_y_etime = copy_y_time1-copy_y_time0

    call toc(solver_time)
    etime = get_elapse(solver_time)
    if (nodeid==0) then
      write(*,'(" solver done ",2I3," ETime:",E24.16," COPY_Y_ETime:",E24.16," |y|^2:",E24.16," |Dy|^2:",E24.16)')&
&     ic,is,etime,copy_y_etime,rtmp0,rtmp1
    endif
  enddo
  enddo

  deallocate(y,Dy)
#ifdef _CLOVER
  call delete_inverse_clover_term(this%parameter%mass)
#endif

  call toc(this%parameter%etime)

  return
end subroutine

subroutine assign_inv_mult_wd(param,iout,tol,iiter,dy,y,u)
  use hmc_std_quark_wilson_class, only : assign_inv_mult_eoprec_wd
  implicit none
#ifdef _CLOVER
  type(quark_clover), intent(inout) :: param
#else
  type(quark_wilson), intent(inout) :: param
#endif
  integer,  intent(in)    :: iout
  real(DP), intent(in)    :: tol
  integer,  intent(inout) :: iiter
  type(field_quark_wg),  intent(in)    :: y
  type(field_quark_wg),  intent(inout) :: dy
  type(vfield_gluon_wg), intent(in)    :: u
  type(field_quark_wg),  allocatable :: w
#ifdef _DDD_
  type(field_quark_wg),  allocatable :: s
#endif
  real(DP) :: rtmp,rtmp2
  integer :: ieo,ioe
  ieo = 0
  ioe = 1-ieo


  allocate(w)
  call new(w)
#ifdef _CLOVER
  call assign_mult_clover_term(w%eo(ieo),param,y%eo(ieo))
  call assign_mult_clover_term(w%eo(ioe),param,y%eo(ioe))
#else
  call assign(w,y) ! w = y
#endif

#ifdef _DDD_
rtmp = abs2(w)
if(nodeid==0) write(*,'(" |F\y|^2:",E24.16)')rtmp
#endif

  !
  !  we = we + kappa Meo yo
  !
  call assign_mult_hop(param,dy%eo(ieo),w%eo(ioe),u)
  rtmp = get_kappa(param)
  call accum_add_mult(w%eo(ieo),dy%eo(ieo),rtmp)

#ifdef _DDD_
write(*,*)"kappa",rtmp
rtmp = abs2(w%eo(ieo))
if(nodeid==0) write(*,'(" |we|^2:",E24.16)')rtmp
#endif

  call assign_inv_mult_eoprec_wd(param,iout,tol,iiter,dy%eo(ieo),w%eo(ieo),u)

#ifdef _DDD_
rtmp = abs2(dy%eo(ieo))
if(nodeid==0) write(*,'(" |dye|^2:",E24.16)')rtmp
#endif

  !
  !  dyo = wo + kappa Moe dye
  !
  call assign(dy%eo(ioe),w%eo(ioe))
  call assign_mult_hop(param,w%eo(ioe),dy%eo(ieo),u)
  rtmp = get_kappa(param)
  call accum_add_mult(dy%eo(ioe),w%eo(ioe),rtmp)   ! dy = dy + w * rtmp


#ifdef _DDD_
rtmp = abs2(dy)
if(nodeid==0) write(*,'(" |dy|^2:",E24.16)')rtmp
!====================
! s = F\M dy
! w = dy
!====================
allocate(s)
call new(s)
call assign(w,dy)
call assign_mult_hop(param,s%eo(ieo),dy%eo(ioe),u)
call assign_mult_hop(param,s%eo(ioe),dy%eo(ieo),u)
rtmp = - get_kappa(param)
call accum_add_mult(w,s,rtmp)  ! w = w - s*kappa = F\D dy
!==============
! s = F\y - w
!==============
call assign_mult_clover_term(s%eo(ieo),param,y%eo(ieo))
call assign_mult_clover_term(s%eo(ioe),param,y%eo(ioe))
rtmp = abs2(s)
call accum_sub(s,w)
rtmp2 = abs2(s)
if(nodeid==0) write(*,'(" |F\(b-Dy)|/|F\b|:",E24.16)')sqrt(rtmp2/rtmp)
deallocate(s)
#endif

  deallocate(w)

!  call comlib_finalize()
!  stop
  return
end subroutine

subroutine assign_mult_u_q(uq,q,u,mu)
!  
! mu : direction
!  u : gauge link 
!  q : quark prop
! uq : quark prop * link
!  
!   uq(n) = u_mu(n) q(n+mu) 
!  
  implicit none
  integer, intent(in) :: mu
  type(vfield_gluon_wg), intent(in)    :: u
  type(quark_prop),      intent(in)    :: q
  type(quark_prop),      intent(inout) :: uq
  type(field_quark_wg), allocatable :: y,uy
  integer :: ic,is

  uq = q
  allocate(y,uy)
  call new(y)
  call new(uy)

  do is=1,SPIN
  do ic=1,COL
    call conv_prop_to_y(uq, y,ic,is)
    call assign_mult_u_y(uy,y,u,mu)
    call conv_y_to_prop(uy,uq,ic,is)
  enddo
  enddo

  deallocate(y,uy)

  return
end subroutine

subroutine assign_mult_u_y(uy,y,u,mu)
!  
!   uy(n) = u_mu(n) y(n+mu) 
!  
  implicit none
  integer, intent(in) :: mu
  type(vfield_gluon_wg), intent(in) :: u
  type(field_quark_wg),  intent(inout) :: y, uy

  call assign_mult_u_eo(uy%eo(0),y%eo(1),u%eo(0),mu)
  call assign_mult_u_eo(uy%eo(1),y%eo(0),u%eo(1),mu)

  return
end subroutine

subroutine assign_mult_u_eo(uye,yo,ue,mu)
!  
!   uy(n) = u_mu(n) y(n+mu) 
!  
  use error_class
  implicit none
  integer, intent(in) :: mu
  type(vfield_gluon_eo_wg), intent(in) :: ue
  type(field_quark_eo_wg),  intent(inout) :: yo, uye

  integer :: ix,iy,iz,ith,ieoxyz,ic,is
  integer :: ix2,iy2,iz2,ith2,ieo,ioe

  ioe = yo%ieo
  ieo = ue%ieo
  if (ieo == ioe) then
    call error_stop("assign_mult_u_eo yo%ieo = ue%ieo ieo error.")
  endif
  uye%ieo = ue%ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ix2,iy2,iz2,ieoxyz,ith,ith2,ic,is)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ

    ix2=ix+isx(mu)
    iy2=iy+isy(mu)
    iz2=iz+isz(mu)

    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ith2=ith+ieoxyz*ist(mu)
      do is=1,SPIN
      do ic=1,COL
        uye%s(ith,iz,iy,ix)%y(ic,is) = ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,1)  &
 &                                           *yo%s(ith2,iz2,iy2,ix2)%y(1,is)  &
 &                                   + ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,2)  &
 &                                           *yo%s(ith2,iz2,iy2,ix2)%y(2,is)  &
 &                                   + ue%mu(mu)%s(ith ,iz ,iy ,ix )%u(ic,3)  &
 &                                           *yo%s(ith2,iz2,iy2,ix2)%y(3,is)
      enddo
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

end module
