#include <config.h>

module wavefunc_param_class
  use comlib
  use lattice_class
  implicit none
  private
  public :: wavefunc_param_obj

  public :: MAXPOL_WAV
  public ::  LOCAL_WAV
  public ::    EXP_WAV
  public ::   WALL_WAV
  public ::   POLY_WAV
  public ::   USER_WAV
  public :: new,delete,read,print

  integer, parameter :: MAXPOL_WAV=50
  integer, parameter ::  LOCAL_WAV=1
  integer, parameter ::    EXP_WAV=2
  integer, parameter ::   WALL_WAV=3
  integer, parameter ::   POLY_WAV=4
  integer, parameter ::   USER_WAV=5

  integer, parameter :: NSPACE=NX*NY*NZ

  type wavefunc_param_obj
  !
  ! quark source function parameters
  !
    integer  :: type                ! 1 for local, 2 for exponential, 3 for wall, 4 for polynomial, 5 user supplied
    integer  :: Npolsf              ! number of polynomial coeffciient
    integer  :: idummy(2)
    real(DP) :: Asmear,Bsmear       ! exponetial parameter, psi(r)=A*exp(-r*B)
    real(DP) :: Psmear(MAXPOL_WAV)  ! polynomial coeffcient, psi(r)= 1 - sum_(i=1,Npol) P(i)*r^i
  end type

  interface new
    module procedure new_wavefunc_param
  end interface

  interface delete
    module procedure delete_wavefunc_param
  end interface

  interface read
    module procedure read_wavefunc_param
  end interface

  interface print
    module procedure print_wavefunc_param
  end interface

contains

subroutine new_wavefunc_param(this)
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this
  this%type      = LOCAL_WAV
  this%Npolsf    = 0
  this%Asmear    = 1.0_DP
  this%Bsmear    = 0.0_DP
  this%Psmear(:) = 0.0_DP
  return
end subroutine

subroutine delete_wavefunc_param(this)
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this
  call new(this)
  return
end subroutine

subroutine read_wavefunc_param(this,iout)
  use error_class
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this
  integer,                  intent(in)    :: iout
  character(len=CHARLEN) :: str
  integer :: ierr,i

  if (nodeid==0) then
    read(iout,*) this%type
  endif
  if (NPU>1) then
    call comlib_bcast(this%type,0)
  endif

  if ( this%type < 1 .and. this%type > 5) then
    call error_message(" Wave function type error.")
    call    error_stop(" type = 1:loc, 2:exp, 3:wall, 4:pol, 5:usr")
  endif
 
  ierr = 0
  if (nodeid==0) then
    select case(this%type)
    case(LOCAL_WAV)
      continue
    case(WALL_WAV)
      continue
    case(EXP_WAV)
      read(iout,*)this%Asmear,this%Bsmear 
    case(POLY_WAV)
      read(iout,*) this%Npolsf 
      if (this%Npolsf > MAXPOL_WAV) then
        ierr=1
        goto 100
      endif
      read(iout,*)(this%Psmear(i),i=1,this%Npolsf) ! read polynomial coeffcients
    end select
  endif

100 continue
  if (ierr == 1) then
    write(str,'(" Error Npolsf =",I3," > MAXPOL_WAV =",I3)') this%Npolsf,MAXPOL_WAV
    call error_stop(TRIM(str))
  endif

  if (NPU>1) then
    call comlib_bcast(this%Asmear,0)
    call comlib_bcast(this%Bsmear,0)
    call comlib_bcast(this%Npolsf,0)

    if (this%type == POLY_WAV) then
      do i=1,this%Npolsf
        call comlib_bcast(this%Psmear(i),0)
      enddo
    endif
  endif

  return
end subroutine

subroutine print_wavefunc_param(this)
  implicit none
  type(wavefunc_param_obj), intent(inout) :: this
  integer :: i

  if (nodeid==0) then
    write(*,'("==== Wave function Parameters ====")')

    select case(this%type)
    case (LOCAL_WAV)
    write(*,'(3X," Local Wave function")')
    case (EXP_WAV)
    write(*,'(3X," Exponential Wave function")')
    write(*,'(3X,"          Asmear :",E24.16)')this%Asmear
    write(*,'(3X,"          Bsmear :",E24.16)')this%Bsmear
    case (WALL_WAV)
    write(*,'(3X," Wall Wave function")')
    case (POLY_WAV)
    write(*,'(3X," Polynomial Wave function")')
    write(*,'(3X,"           Order :",I3)')this%Npolsf
    write(*,'(3X,"       Psmear(:) :",100E24.16)') (this%Psmear(i),i=1,this%Npolsf)
    end select
    write(*,'(80("="))')
  endif

  return
end subroutine

end module
