#include "../config.h"
#include "../solver_config.h"

module quark_prop_class
  use comlib
  use lattice_class
  use timer_class
  use solver_parameter_class
  use inverse_class
  use logging_class
  use gauge_field_dd_class
  use gauge_field_deo_class
  use gauge_field_eo_deo_class
  use wilson_quark_field_dd_class
  use wilson_quark_field_deo_class
  use wilson_quark_field_eo_deo_class
  use wilson_quark_dd_class
#ifdef _CLOVER
  use clover_quark_dd_class
  use clover_quark_deo_class
  use clover_quark_eo_deo_class
#endif
  use wilson_quark_parameter_class
  use wilson_quark_ddhmc_class
  use wavefunc_class
  use single_solver_class
  implicit none
  private
  public :: quark_prop_obj, quark_prop_parameter_obj
  public :: new,delete,read,print,print_statistics
  public :: get_prop,mult,clear_prop
  public :: ispace

  integer, parameter :: NSPACE=NX*NY*NZ

  type quark_prop_parameter_obj
  ! quark propagator parameter for physics measurement
    type(wilson_quark_parameter_obj)  :: quark_mass
    type(solver_parameter_obj)        :: solver_param
    type(ir_solver_parameter_obj)     :: ir_solver
    character(len=_CHARLEN_)          :: fname
    character(len=_CHARLEN_)          :: path
    type(wavefunc_param_obj)          :: wavefunc_sous
    type(wavefunc_param_obj)          :: wavefunc_sink
    type(timer_obj)                   :: etime
    type(logging_obj), pointer        :: solver_log => null()
    integer                           :: smear_sous ! 0 for local, 1 for smeared
    integer                           :: smear_sink ! 0 for local, 1 for smeared
    integer                           :: idummy1
    integer                           :: idummy2
  end type

  type quark_prop_obj
  ! quark propagator for physics measurement
    complex(8)                      :: q(COL,COL,SPIN,SPIN,NZ*NY*NX,NT) ! q(sink,sous,sink,sous,sink,sink)
    real(8)                         :: psum(NTT) ! check sum
    type(quark_prop_parameter_obj)  :: parameter
  end type

  interface new
    module procedure new_prop
    module procedure new_prop_parameter
  end interface

  interface delete
    module procedure delete_prop
    module procedure delete_prop_parameter
  end interface

  interface read
    module procedure read_prop
    module procedure read_prop_parameter
  end interface

!  interface save
!    module procedure save_qprop
!  end interface

  interface get_prop
    module procedure get_quark_prop
  end interface

  interface print
    module procedure print_prop_parameter
  end interface

  interface print_statistics
    module procedure print_stat_prop
  end interface

  interface mult
    module procedure mult_u_q
    module procedure mult_u_dd
  end interface

contains

function ispace(ix,iy,iz) 
  implicit none
  integer :: ix,iy,iz,ispace

  ispace= iz + (iy-1)*NZ + (ix-1)*NZ*NY

  return
end function

subroutine new_prop(this)
! set default values for a quark propagator
  implicit none
  type(quark_prop_obj), intent(inout) :: this

  call new(this%parameter)

  return
end subroutine

subroutine delete_prop(this)
! set default values for a quark propagator
  implicit none
  type(quark_prop_obj), intent(inout) :: this

  call delete(this%parameter)

  return
end subroutine

subroutine new_prop_parameter(this)
! set default values for a quark propagator parameter
  implicit none
  type(quark_prop_parameter_obj), intent(inout) :: this

  this%quark_mass%kappa     = 0.1d0
  this%quark_mass%csw       = 0.0d0

  this%solver_param%tol     = 1.d-15
  this%solver_param%maxiter = 1000

  this%ir_solver%omega    = 1.1d0
  this%ir_solver%nmr      = 2
  this%ir_solver%nsap     = 5
  this%ir_solver%nchro    = 0
  this%ir_solver%solver_log => this%solver_log

  call new(this%ir_solver%single_solver)

  call new(this%etime)
  call new(this%wavefunc_sous)
  call new(this%wavefunc_sink)

  return
end subroutine

subroutine delete_prop_parameter(this)
  implicit none
  type(quark_prop_parameter_obj), intent(inout) :: this

  nullify(this%ir_solver%solver_log)

  return
end subroutine

subroutine read_prop(iout,this)
! read parameters for a quark propagator
  implicit none
  integer, intent(in) :: iout
  type(quark_prop_obj), intent(inout) :: this

  call read(iout,this%parameter)

  return
end subroutine

subroutine read_prop_parameter(iout,this)
! read parameters for a quark propagator parameter
  implicit none
  integer, intent(in) :: iout
  type(quark_prop_parameter_obj), intent(inout) :: this

  if (nodeid==0) then
!******************************
!* read wilson quark mass parameters
!******************************
#ifdef _CLOVER
    read(iout,*)this%quark_mass%kappa,  &
 &              this%quark_mass%csw
#else
    read(iout,*)this%quark_mass%kappa
                this%quark_mass%csw=0.0d0
#endif

!******************************
!* output quark propagetor 
!* file parameters
!******************************
!    read(iout,'(A)')this%path   ! save path 
!    read(iout,'(A)')this%fname  ! file name

!******************************
!* read Solver parameters
!******************************
    read(iout,*)this%solver_param%tol,  &
 &              this%solver_param%maxiter

!******************************
!* read Solver parameters
!******************************
    read(iout,*)this%ir_solver%omega,  &
 &              this%ir_solver%nmr,    &
 &              this%ir_solver%nsap
  endif

  if (NPU > 1) then
    call comlib_bcast(this%quark_mass%kappa,0)
    call comlib_bcast(this%quark_mass%csw,0)

    call comlib_bcast(this%solver_param%tol,0)
    call comlib_bcast(this%solver_param%maxiter,0)

    call comlib_bcast(this%ir_solver%omega,0)
    call comlib_bcast(this%ir_solver%nmr,0)
    call comlib_bcast(this%ir_solver%nsap,0)
  endif

  return
end subroutine

subroutine print_prop_parameter(this)
! read parameters for a quark propagator parameter
  implicit none
  type(quark_prop_parameter_obj), intent(inout) :: this

  if (nodeid==0) then
    write(*,'("==== Mass Params ====")')
    write(*,'(3X,"           Kappa :",F24.16)')this%quark_mass%kappa
#ifdef _CLOVER
    write(*,'(3X,"             Csw :",F24.16)')this%quark_mass%csw
#endif

    write(*,'("==== Solver Params ====")')
    write(*,'(3X," Solver tol      :",E24.16)')this%solver_param%tol
    write(*,'(3X," Solver max iter :",I10)')   this%solver_param%maxiter

    write(*,'("==== Solver Params II ====")')
    write(*,'(" outer-BiCGStab + inner-BiCGStab/GCRODR + SAP(NSAP) + domain-SSOR(NMR)")')
    write(*,'(3X,"        OR-param :",E16.8)')this%ir_solver%omega
    write(*,'(3X,"             NMR :",I4)')this%ir_solver%nmr
    write(*,'(3X,"            NSAP :",I4)')this%ir_solver%nsap

    write(*,'("==== Smearing type ====")')
    write(*,'(3X,"          Source :",I3)')this%smear_sous
    write(*,'(3X,"            Sink :",I3)')this%smear_sink

  endif

  if (nodeid==0) then
    write(*,'("==== Source Wavefunction type ====")')
  endif

  call print(this%wavefunc_sous)

  if (nodeid==0) then
    write(*,'("==== Sink Wavefunction type ====")')
  endif

  call print(this%wavefunc_sink)

  if (nodeid==0) then
    write(*,'(80("="))')
  endif

  return
end subroutine

subroutine psum(this)
  implicit none
  type(quark_prop_obj), intent(inout) :: this
  real(8) :: rtmp
  integer :: it,itt,ic,is,jc,js,isp

  this%psum(:)=0.0d0

  do it=1,NT
    itt = it + ipsite(4)*NT
    rtmp=0.0d0
    do isp=1,NSPACE
      do js=1,SPIN
      do is=1,SPIN
      do ic=1,COL
      do jc=1,COL
        rtmp=rtmp+ real(this%q(ic,jc,is,js,isp,it))**2 &
 &               +aimag(this%q(ic,jc,is,js,isp,it))**2
      enddo
      enddo
      enddo
      enddo
    enddo
    this%psum(itt)=rtmp
  enddo

  if (NPU > 1) then
    do itt=1,NTT
      call comlib_sumcast(this%psum(itt))
    enddo
  endif

  return
end subroutine

subroutine print_stat_prop(this)
  implicit none
  type(quark_prop_obj), intent(inout) :: this
  integer :: itt

  call psum(this)

  call average(this%parameter%solver_param)
  call average(this%parameter%etime)

  if (nodeid==0) then
    write(*,'("==== Solver Statistics ====")')
    write(*,'(3X," Solver iteration :",F12.4)')this%parameter%solver_param%solver_counter%ave
    write(*,'(3X,"             Time :",F12.4)')this%parameter%etime%elapse
  endif

  if (nodeid==0) then
    write(*,'("==== Quark Prop |S(itt)|^2 ====")')
    do itt=1,NTT
      write(*,'(I3,E24.16)')itt,this%psum(itt)
    enddo
    write(*,'("")')
  endif

  return
end subroutine

subroutine clear_prop(this)
  implicit none
  type(quark_prop_obj), intent(inout) :: this
  complex(8), parameter :: z0=(0.0d0,0.0d0)

  this%q(:,:,:,:,:,:)=z0

  return
end subroutine

subroutine set_source(this)
  implicit none
  type(quark_prop_obj), intent(inout) :: this
  integer :: it,ic,is,isp
  complex(8), parameter :: z0=(0.0d0,0.0d0)

  this%q(:,:,:,:,:,:)=z0

  do it=1,NT
    do isp=1,NSPACE
      do is=1,SPIN
      do ic=1,COL
        this%q(ic,ic,is,is,isp,it)=this%parameter%wavefunc_sous%wavefunc(ic,is,isp,it)
      enddo
      enddo
    enddo
  enddo

  return
end subroutine

subroutine conv_prop_to_y(this,ic0,is0,y)
  implicit none
  type(quark_prop_obj), intent(in) :: this
  type(wqfield_dd_wg_obj), intent(inout) :: y
  integer, intent(in) :: ic0,is0
  integer :: id,ieoxyz,ibth,ibt,ibz,iby,ibx,ic,is,ieo,ideo,isp
  integer :: idx,idy,idz,idt,idd
  integer :: ix,iy,iz,it

  do ieo =0,1
  do ideo=0,1
  do id = 0,NDM-1
    idd=id*2+idmod(id,ideo) ! calc. block index in 1-pu
    idt=mod(idd,NDT)
    idz=mod(idd/NDT,NDZ)
    idy=mod(idd/NDT/NDZ,NDY)
    idx=mod(idd/NDT/NDZ/NDY,NDX)

    do ibx=1,NBX  ! site loop in 1 block.
    do iby=1,NBY
    do ibz=1,NBZ
    ieoxyz=mod(iddeo(id,ideo)+ibz+iby+ibx+ieo,2)
    do ibth=1-ieoxyz,NBTH-ieoxyz
      ibt=ibth*2+ieoxyz
      ix=ibx+idx*NBX ! calc. global site index
      iy=iby+idy*NBY
      iz=ibz+idz*NBZ
      it=ibt+idt*NBT

      isp=ispace(ix,iy,iz)

      do is=1,SPIN
      do ic=1,COL
        y%deo(ideo)%eo(ieo)%b(id)%s(ibth,ibz,iby,ibx)%y(ic,is)=this%q(ic,ic0,is,is0,isp,it)
      enddo
      enddo

    enddo
    enddo
    enddo
    enddo
  enddo
  enddo
  enddo

  call qboundary_copy(y)

  return
end subroutine

subroutine conv_y_to_prop(y,this,ic0,is0)
  implicit none
  type(quark_prop_obj), intent(inout)    :: this
  type(wqfield_dd_wg_obj), intent(inout) :: y
  integer, intent(in) :: ic0,is0
  integer :: id,ieoxyz,ibth,ibt,ibz,iby,ibx,ic,is,ieo,ideo,isp
  integer :: idx,idy,idz,idt,idd
  integer :: ix,iy,iz,it

  do ieo =0,1
  do ideo=0,1
  do id = 0,NDM-1
    idd=id*2+idmod(id,ideo) ! calc. block index in 1-pu
    idt=mod(idd,NDT)
    idz=mod(idd/NDT,NDZ)
    idy=mod(idd/NDT/NDZ,NDY)
    idx=mod(idd/NDT/NDZ/NDY,NDX)

    do ibx=1,NBX  ! site loop in 1 block.
    do iby=1,NBY
    do ibz=1,NBZ
    ieoxyz=mod(iddeo(id,ideo)+ibz+iby+ibx+ieo,2)
    do ibth=1-ieoxyz,NBTH-ieoxyz
      ibt=ibth*2+ieoxyz
      ix=ibx+idx*NBX ! calc. global site index
      iy=iby+idy*NBY
      iz=ibz+idz*NBZ
      it=ibt+idt*NBT

      isp=ispace(ix,iy,iz)

      do is=1,SPIN
      do ic=1,COL
        this%q(ic,ic0,is,is0,isp,it)=y%deo(ideo)%eo(ieo)%b(id)%s(ibth,ibz,iby,ibx)%y(ic,is)
      enddo
      enddo

    enddo
    enddo
    enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine get_quark_prop(this,u)
  implicit none
  type(quark_prop_obj), intent(inout) :: this
  type(gvfield_dd_wg_obj), intent(in) :: u
  type(timer_obj) :: timer
  type(wqfield_dd_wg_obj), pointer :: y,Dy
  integer :: maxiter,mode,iter
  integer :: ic,is
  real(8) :: tol
#ifdef _CLOVER
  real(8) :: dummy
  type(ctfield_dd_obj),   pointer :: ucl
  type(cchlfield_dd_obj), pointer :: fclinv
#endif

  call tic(this%parameter%etime)

#ifdef _CLOVER
!*******************************
!* calc clover term
!*******************************
  allocate(ucl,fclinv)
  call make_clover_term(ucl,u)
  call make_inverse_clover_term(this%parameter%quark_mass,fclinv,ucl,dummy,0)
  deallocate(ucl)
#endif

  call clear_prop(this)

  call set_source(this)

  allocate(y,Dy)

  call new(this%parameter%ir_solver%single_solver)

  do is=1,SPIN
  do ic=1,COL
    call new(timer)
    call tic(timer)
    call new(y)
    call new(Dy)
!*******************
!*  y = F^-1 src
!*******************
#ifdef _CLOVER
    call conv_prop_to_y(this,ic,is,Dy)
    call mult_clover_term(Dy,y,fclinv)
#else
    call conv_prop_to_y(this,ic,is,y)
#endif

!**********************
!* Dy = \tilde{D}^-1 y
!**********************
    tol    = this%parameter%solver_param%tol
    maxiter= this%parameter%solver_param%maxiter
    if ( tol <= 0.0d0) then
      mode = 1  ! stop by maxiter
    else
      mode = 0  ! stop by tol
    endif
    call set_inverse_parameters(tolerance=tol,max_iteration=maxiter,inverse_mode=mode)

#ifdef _CLOVER
    call inverse_mult_wilson_dirac(ir_solver=this%parameter%ir_solver,kappa=this%parameter%quark_mass%kappa,&
  &                                y=y,Dy=Dy,u=u,fclinv=fclinv)
#else
    call inverse_mult_wilson_dirac(ir_solver=this%parameter%ir_solver,kappa=this%parameter%quark_mass%kappa,&
  &                                y=y,Dy=Dy,u=u)
#endif

    call get_inverse_results(iteration=iter)
    call inc(this%parameter%solver_param,iter)

    call conv_y_to_prop(Dy,this,ic,is)
    call tac(timer)
    call average(timer)
    if (nodeid==0) write(*,'(" solver done ",2I3," ETime:",E24.16)')ic,is,timer%elapse

  enddo
  enddo

  call delete(this%parameter%ir_solver%single_solver)
  deallocate(y,Dy)

  call tac(this%parameter%etime)

  return
end subroutine

!#include "save_qprop.h90"
#include "mult_u.h90"

end module
