#include <config.h>

module main_singlet_class
  use config_profile_class
  use comlib
  use timer_class
  use random_class
  use print_status_class
  use lattice_class
  use logfile_class
  use field_gauge_class
  use field_fermion_class
#ifdef _CLOVER
  use quark_clover_class
#else
  use quark_wilson_class
#endif
  use quark_prop_class
  use wavefunc_class
  use hadron_class
  use main_hadron_class
  use gamma_matrix_class
  implicit none
  private

  public :: main_hadron_obj, main_singlet_obj
  public :: new,read,print,delete,save
  public :: new_meson_dloop
  public :: delete_meson_dloop
  public :: get_meson_dloop
  public :: save_meson_dloop
  public :: m_rand
  public :: shift

  type(rand_gfsr_obj), save :: m_rand
  logical, save :: m_is_initialized = .FALSE.
  integer, save :: m_js_out  =957
  integer, save :: m_seed_out=34

  type main_singlet_obj
    type(main_hadron_obj) :: hadrons
    character(len=CHARLEN) :: singlet_parameter_file
    character(len=CHARLEN) :: seed_fpath
    character(len=CHARLEN) :: seed_fname
    integer :: run_number=0            ! run number
    integer :: random_number_seed=1    ! stochastic trace estimator, noise seed
    integer :: nnoise=1                ! stochastic trace estimator, noise number
    integer :: idummy1
  end type

  interface new
    module procedure new_main_singlet
  end interface

  interface delete
    module procedure delete_main_singlet
  end interface

  interface read
    module procedure read_main_singlet
  end interface

  interface print
    module procedure print_main_singlet
  end interface

  interface save
    module procedure save_main_singlet
  end interface

contains

subroutine new_main_singlet(this)
  implicit none
  type(main_singlet_obj), intent(inout) :: this
  type(gamma_matrix_obj), save :: gamma_matrix

  call new(this%hadrons)
  this%singlet_parameter_file=""
  this%seed_fpath=""
  this%seed_fname="singlet_seed."

  call new(gamma_matrix)

  return
end subroutine

subroutine delete_main_singlet(this)
  implicit none
  type(main_singlet_obj), intent(inout) :: this
  call delete(m_rand)
  call delete(this%hadrons)
  return
end subroutine

subroutine read_main_singlet(iout,this)
  use file_tools_class
  implicit none
  integer,                intent(in)    :: iout
  type(main_singlet_obj), intent(inout) :: this
  integer :: istat

  call read(this%hadrons,iout)

  if (nodeid == 0) then
    read(iout,*)
    read(iout,'(A)')this%singlet_parameter_file
  endif
  call comlib_bcast(this%singlet_parameter_file,0)

  istat=0
  if (nodeid == 0) then
    m_js_out = search_free_file_unit()
    open(m_js_out,file=TRIM(this%singlet_parameter_file),form='formatted',status='old',iostat=istat)
  endif

  if (NPU > 1) then
    call comlib_bcast(istat,0)
  endif
  if (istat /= 0) goto 100
  
  if (nodeid == 0) then
    read(m_js_out,*)this%run_number
    read(m_js_out,*)this%random_number_seed
    read(m_js_out,'(A)')this%seed_fpath
    read(m_js_out,'(A)')this%seed_fname
    read(m_js_out,*)this%nnoise
    close(m_js_out)
  endif

  if (NPU > 1) then
    call comlib_bcast(this%run_number,0)
    call comlib_bcast(this%random_number_seed,0)
    call comlib_bcast(this%seed_fpath,0)
    call comlib_bcast(this%seed_fname,0)
    call comlib_bcast(this%nnoise,0)
  endif

  return

100 continue
  if (nodeid==0) then
    write(*,'("Error Reading Singlet Parameter file.",A)')TRIM(this%singlet_parameter_file)
  endif
  call comlib_finalize
  stop

end subroutine

subroutine print_main_singlet(this)
  use file_tools_class
  use hmc_status_class, only : fname_cont
  implicit none
  type(main_singlet_obj), intent(inout) :: this
  character(len=CHARLEN) :: fname,char

  call print(this%hadrons)

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'(" Singlet Parameter file :",A)')TRIM(this%singlet_parameter_file)
    write(*,'("               Run # :",I10)')this%run_number
    write(*,'("         Random Seed :",I10)')this%random_number_seed
    write(*,'("      Seed Save path :",A)')TRIM(this%seed_fpath)
    write(*,'("      Seed Save name :",A)')TRIM(this%seed_fname)
    write(*,'(" # of Noise Estimate :",I10)')this%nnoise
    write(*,'(80("="))')
  endif

!=====================================
! Initialize Random Number generator
!=====================================
  fname=fname_cont(this%run_number,this%seed_fpath,this%seed_fname)
  select case (this%run_number)
  case (0)
    call new(m_rand,0,this%random_number_seed,NPU,nodeid)
    if (nodeid == 0) then
      write(*,'(1X,"Random Number Generator Initialized")')
      write(*,'(80("-"))')
    endif
  case (1:)
    call new(m_rand,1,this%random_number_seed,NPU,nodeid)
    m_seed_out = search_free_file_unit()
    open(m_seed_out,file=fname,status='unknown',form='unformatted',action='read')
    call read(m_rand,m_seed_out)
    close(m_seed_out)
    if (nodeid == 0) then
      write(*,'(1X,"Random Number Generator Recovered")')
      write(*,'(80("-"))')
    endif
    write(char,'(1X,A," : read OK ",I3)')TRIM(fname),nodeid
    call print_status(char)
  end select
  this%run_number=this%run_number+1
  m_is_initialized = .TRUE.
  
  return
end subroutine

subroutine save_main_singlet(this)
  use file_tools_class
  use hmc_status_class, only : fname_cont
  implicit none
  type(main_singlet_obj), intent(inout) :: this
  character(len=CHARLEN) :: fname

!=====================================
! Save Random Seed
!=====================================
  fname = fname_cont(this%run_number,this%seed_fpath,this%seed_fname)
  m_seed_out = search_free_file_unit()
  open(m_seed_out,file=fname,status='unknown',form='unformatted',action='write')
  call save(m_rand,m_seed_out)
  close(m_seed_out)

!=====================================
! Save Input Parameters
!=====================================
  if (nodeid == 0) then
    m_js_out = search_free_file_unit()
    open(m_js_out,file=TRIM(this%singlet_parameter_file),form='formatted',status='old')
    write(m_js_out,*)this%run_number
    write(m_js_out,*)this%random_number_seed
    write(m_js_out,'(A)')TRIM(this%seed_fpath)
    write(m_js_out,'(A)')TRIM(this%seed_fname)
    write(m_js_out,*)this%nnoise
    close(m_js_out)
  endif

  return
end subroutine

#include "get_meson_dloop.h90"

subroutine shift(u,itx,ity,itz)
  implicit none
  type(vfield_gluon_wg), intent(inout) :: u
  integer, intent(in) :: itx,ity,itz
  type(sfield_gluon_wg), allocatable :: w
  integer :: idir,ix,iy,iz,mu,Nshift(1:NDIM-1)
  allocate(w)
  call new(w)
  call copy_boundary(u)

  Nshift(1) = itx
  Nshift(2) = ity
  Nshift(3) = itz

  do idir = 1,NDIM-1
  do ix=1,Nshift(idir)-1
    do mu=1,NDIM
      call assign_shift(w%eo(0),u%eo(1)%mu(mu),-idir)
      call assign_shift(w%eo(1),u%eo(0)%mu(mu),-idir)
      call assign(u%eo(0)%mu(mu),w%eo(0))
      call assign(u%eo(1)%mu(mu),w%eo(1))
    enddo
    call copy_boundary(u)
  enddo
  enddo

  deallocate(w)
  return
end subroutine

end module
